;==============================================================================

; file: "_parms.scm"

; Copyright (C) 1994-1996 by Marc Feeley, All Rights Reserved.

(include "fixnum.scm")

;------------------------------------------------------------------------------
;
; Compiler parameters module
; --------------------------

; This module contains definitions that parameterize the behavior of
; the compiler.

;------------------------------------------------------------------------------

; Symbol "canonicalization".

; (string->canonical-symbol str) behaves like 'string->symbol' but all the
; letters in the symbol are set to a particular case.

(define (string->canonical-symbol str)
  (let ((len (string-length str)))
    (let loop ((str str)
               (s (make-string len))
               (i (- len 1)))
      (if (>= i 0)
        (begin
          (string-set! s i (char-downcase (string-ref str i)))
          (loop str s (- i 1)))
        (string->symbol s)))))


; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Specific symbols needed by the compiler.

(define QUOTE-sym              (string->canonical-symbol "QUOTE"))
(define QUASIQUOTE-sym         (string->canonical-symbol "QUASIQUOTE"))
(define UNQUOTE-sym            (string->canonical-symbol "UNQUOTE"))
(define UNQUOTE-SPLICING-sym   (string->canonical-symbol "UNQUOTE-SPLICING"))
(define LAMBDA-sym             (string->canonical-symbol "LAMBDA"))
(define IF-sym                 (string->canonical-symbol "IF"))
(define SET!-sym               (string->canonical-symbol "SET!"))
(define COND-sym               (string->canonical-symbol "COND"))
(define =>-sym                 (string->canonical-symbol "=>"))
(define ELSE-sym               (string->canonical-symbol "ELSE"))
(define AND-sym                (string->canonical-symbol "AND"))
(define OR-sym                 (string->canonical-symbol "OR"))
(define CASE-sym               (string->canonical-symbol "CASE"))
(define LET-sym                (string->canonical-symbol "LET"))
(define LET*-sym               (string->canonical-symbol "LET*"))
(define LETREC-sym             (string->canonical-symbol "LETREC"))
(define BEGIN-sym              (string->canonical-symbol "BEGIN"))
(define DO-sym                 (string->canonical-symbol "DO"))
(define DEFINE-sym             (string->canonical-symbol "DEFINE"))
(define DELAY-sym              (string->canonical-symbol "DELAY"))
(define FUTURE-sym             (string->canonical-symbol "FUTURE"))
(define **DEFINE-MACRO-sym     (string->canonical-symbol "##DEFINE-MACRO"))
(define **DECLARE-sym          (string->canonical-symbol "##DECLARE"))
(define **INCLUDE-sym          (string->canonical-symbol "##INCLUDE"))

(define NOT-sym                (string->canonical-symbol "NOT"))

(define **NOT-sym              (string->canonical-symbol "##NOT"))
(define **EQ?-sym              (string->canonical-symbol "##EQ?"))
(define **EQV?-sym             (string->canonical-symbol "##EQV?"))
(define **QUASI-APPEND-sym     (string->canonical-symbol "##QUASI-APPEND"))
(define **QUASI-LIST-sym       (string->canonical-symbol "##QUASI-LIST"))
(define **QUASI-CONS-sym       (string->canonical-symbol "##QUASI-CONS"))
(define **QUASI-LIST->VECTOR-sym (string->canonical-symbol "##QUASI-LIST->VECTOR"))
(define **MAKE-CELL-sym        (string->canonical-symbol "##MAKE-CELL"))
(define **CELL-REF-sym         (string->canonical-symbol "##CELL-REF"))
(define **CELL-SET!-sym        (string->canonical-symbol "##CELL-SET!"))
(define **MAKE-PLACEHOLDER-sym (string->canonical-symbol "##MAKE-PLACEHOLDER"))

(define IEEE-SCHEME-sym        (string->canonical-symbol "IEEE-SCHEME"))
(define R4RS-SCHEME-sym        (string->canonical-symbol "R4RS-SCHEME"))
(define MULTILISP-sym          (string->canonical-symbol "MULTILISP"))

(define LAMBDA-LIFT-sym        (string->canonical-symbol "LAMBDA-LIFT"))

(define INLINE-sym             (string->canonical-symbol "INLINE"))
(define INLINING-LIMIT-sym     (string->canonical-symbol "INLINING-LIMIT"))

(define BLOCK-sym              (string->canonical-symbol "BLOCK"))
(define SEPARATE-sym           (string->canonical-symbol "SEPARATE"))

(define STANDARD-BINDINGS-sym  (string->canonical-symbol "STANDARD-BINDINGS"))
(define EXTENDED-BINDINGS-sym  (string->canonical-symbol "EXTENDED-BINDINGS"))

(define SAFE-sym               (string->canonical-symbol "SAFE"))

(define INTERRUPTS-ENABLED-sym (string->canonical-symbol "INTERRUPTS-ENABLED"))

(define GENERIC-sym            (string->canonical-symbol "GENERIC"))
(define FIXNUM-sym             (string->canonical-symbol "FIXNUM"))
(define FLONUM-sym             (string->canonical-symbol "FLONUM"))

(define NAMESPACE-sym          (string->canonical-symbol "NAMESPACE"))

(define C-DECLARE-sym          (string->canonical-symbol "C-DECLARE"))
(define C-INITIALIZE-sym       (string->canonical-symbol "C-INITIALIZE"))
(define C-LAMBDA-sym           (string->canonical-symbol "C-LAMBDA"))
(define C-DEFINE-sym           (string->canonical-symbol "C-DEFINE"))

(define VOID-sym               (string->canonical-symbol "VOID"))
(define CHAR-sym               (string->canonical-symbol "CHAR"))
(define SIGNED-CHAR-sym        (string->canonical-symbol "SIGNED-CHAR"))
(define UNSIGNED-CHAR-sym      (string->canonical-symbol "UNSIGNED-CHAR"))
(define LATIN1-sym             (string->canonical-symbol "LATIN1"))
(define UCS4-sym               (string->canonical-symbol "UCS4"))
(define UCS2-sym               (string->canonical-symbol "UCS2"))
(define SHORT-sym              (string->canonical-symbol "SHORT"))
(define UNSIGNED-SHORT-sym     (string->canonical-symbol "UNSIGNED-SHORT"))
(define INT-sym                (string->canonical-symbol "INT"))
(define UNSIGNED-INT-sym       (string->canonical-symbol "UNSIGNED-INT"))
(define LONG-sym               (string->canonical-symbol "LONG"))
(define UNSIGNED-LONG-sym      (string->canonical-symbol "UNSIGNED-LONG"))
(define FLOAT-sym              (string->canonical-symbol "FLOAT"))
(define DOUBLE-sym             (string->canonical-symbol "DOUBLE"))
(define POINTER-sym            (string->canonical-symbol "POINTER"))
(define FUNCTION-sym           (string->canonical-symbol "FUNCTION"))
(define BOOLEAN-sym            (string->canonical-symbol "BOOLEAN"))
(define CHAR-STRING-sym        (string->canonical-symbol "CHAR-STRING"))
(define UCS4-STRING-sym        (string->canonical-symbol "UCS4-STRING"))
(define UCS2-STRING-sym        (string->canonical-symbol "UCS2-STRING"))
(define UTF8-STRING-sym        (string->canonical-symbol "UTF8-STRING"))
(define SCHEME-OBJECT-sym      (string->canonical-symbol "SCHEME-OBJECT"))


; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Identifier prefix for C identifiers generated by the C-interface and
; the C back-end.

(define c-id-prefix "___")


; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Identifier prefix for module names.

(define module-prefix ";")


; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; C types available to the C-interface.

(define scheme-to-c-notation
  (list
    (list VOID-sym           "VOID"       "void")
    (list CHAR-sym           "CHAR"       "char")
    (list SIGNED-CHAR-sym    "SCHAR"      "signed char")
    (list UNSIGNED-CHAR-sym  "UCHAR"      "unsigned char")
    (list LATIN1-sym         "LATIN1"     (string-append c-id-prefix "LATIN1"))
    (list UCS4-sym           "UCS4"       (string-append c-id-prefix "UCS4"))
    (list UCS2-sym           "UCS2"       (string-append c-id-prefix "UCS2"))
    (list SHORT-sym          "SHORT"      "short")
    (list UNSIGNED-SHORT-sym "USHORT"     "unsigned short")
    (list INT-sym            "INT"        "int")
    (list UNSIGNED-INT-sym   "UINT"       "unsigned int")
    (list LONG-sym           "LONG"       "long")
    (list UNSIGNED-LONG-sym  "ULONG"      "unsigned long")
    (list FLOAT-sym          "FLOAT"      "float")
    (list DOUBLE-sym         "DOUBLE"     "double")
    (list BOOLEAN-sym        "BOOLEAN"    "int")
    (list CHAR-STRING-sym    "CHARSTRING" "char*")
    (list UTF8-STRING-sym    "UTF8STRING" (string-append c-id-prefix
                                                         "UTF8STRING"))
    (list UCS4-STRING-sym    "UCS4STRING" (string-append c-id-prefix
                                                         "UCS4STRING"))
    (list UCS2-STRING-sym    "UCS2STRING" (string-append c-id-prefix
                                                         "UCS2STRING"))
    (list SCHEME-OBJECT-sym  "SCMOBJ"     (string-append c-id-prefix
                                                         "WORD"))))


; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Non-standard objects.

(define (string->keyword-object str)
  (string->symbol (string-append str ":")))

(define (keyword-object->string key)
  (let ((str (symbol->string key)))
    (substring str 0 (- (string-length str) 1))))

(define (keyword-object? obj)
  (and #f ; ****************** disable keyword objects
       (symbol? obj)
       (let* ((str (symbol->string obj))
              (len (string-length str)))
         (and (> len 1)
              (char=? (string-ref str (- len 1)) #\:)))))

(define false-object
  (if (eq? '() #f) (string->symbol "#f") #f))

(define (false-object? obj)
  (eq? obj false-object))

(define absent-object
  (string->symbol "#<absent>"))

(define (absent-object? obj)
  (eq? obj absent-object))

(define void-object
  (string->symbol "#<void>"))

(define (void-object? obj)
  (eq? obj void-object))

(define end-of-file-object
  (string->symbol "#!eof"))

(define (end-of-file-object? obj)
  (eq? obj end-of-file-object))

(define optional-object
  (string->symbol "#!optional"))

(define (optional-object? obj)
  (eq? obj optional-object))

(define rest-object
  (string->symbol "#!rest"))

(define (rest-object? obj)
  (eq? obj rest-object))

(define key-object
  (string->symbol "#!key"))

(define (key-object? obj)
  (eq? obj key-object))

(define script-object
  (string->symbol "#!"))

(define (script-object? obj)
  (eq? obj script-object))

(define (symbol-object? obj)
  (and (symbol? obj)
       (not (keyword-object? obj))
       (not (false-object? obj))
       (not (absent-object? obj))
       (not (void-object? obj))
       (not (end-of-file-object? obj))
       (not (optional-object? obj))
       (not (rest-object? obj))
       (not (key-object? obj))
       (not (script-object? obj))))


; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; For '_source.scm'.

; Filename extensions to try in order to find source files from module name.

(define scm-file-exts '(".scm" "")) ; "" means no extension


; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Version of the Gambit compiler and "gambit.h" header file.

(define compiler-version "2.4")
(define header-version 15)


;==============================================================================
