;==============================================================================

; file: "_eval.scm"

; Copyright (C) 1994-1996 by Marc Feeley, All Rights Reserved.

(##include "header.scm")

;------------------------------------------------------------------------------

(##define-macro (make-global-var id) `(##make-global-var ,id))
(##define-macro (global-var-ref gv) `(##global-var-ref ,gv))
(##define-macro (global-var-set! gv val) `(##global-var-set! ,gv ,val))
(##define-macro (global-var->identifier gv) `(##global-var->identifier ,gv))

(##define-macro (quasi-list->vector x) `(##quasi-list->vector ,x))
(##define-macro (quasi-append x y) `(##quasi-append ,x ,y))
(##define-macro (quasi-cons x y) `(##quasi-cons ,x ,y))

(##define-macro (true? x) x)
(##define-macro (unbound? x) `(##unbound? ,x))

(define ##self-var     (##make-uninterned-symbol "<self>"))
(define ##selector-var (##make-uninterned-symbol "<selector>"))
(define ##do-loop-var  (##make-uninterned-symbol "<do-loop>"))

(##define-macro (self-var)     '##self-var)
(##define-macro (selector-var) '##selector-var)
(##define-macro (do-loop-var)  '##do-loop-var)

(define (##internal-var? var)
  (or (##eq? var (self-var))
      (##eq? var (selector-var))
      (##eq? var (do-loop-var))))

(##define-macro (rt-error-unbound-global-var code rte)
  `(##signal '##SIGNAL.GLOBAL-UNBOUND ,code ,rte))

(##define-macro (rt-error-non-procedure-send code rte)
  `(##signal '##SIGNAL.NON-PROCEDURE-SEND ,code ,rte))

(##define-macro (rt-error-non-procedure-oper code rte)
  `(##signal '##SIGNAL.NON-PROCEDURE-OPERATOR ,code ,rte))

(##define-macro (rt-error-too-few-args proc args)
  `(##signal '##SIGNAL.WRONG-NB-ARG ,proc ,args))

(##define-macro (rt-error-too-many-args proc args)
  `(##signal '##SIGNAL.WRONG-NB-ARG ,proc ,args))

(##define-macro (ct-error-global-env-overflow var)
  `(##signal '##SIGNAL.GLOBAL-ENV-OVERFLOW ,var))

(##define-macro (ct-error-syntax msg . args)
  `(##signal '##SIGNAL.SYNTAX-ERROR src ,msg ,@args))

;------------------------------------------------------------------------------

; Macro to create a node of executable code

(##define-macro (mk-code code-prc cte subcodes . lst)
  `(let (($code (##vector #f ,code-prc ,cte ,@subcodes ,@lst)))
     ,@(let loop ((l subcodes) (i 3) (r '()))
         (if (pair? l)
           (loop (cdr l)
                 (+ i 1)
                 (cons `(##vector-set! (##vector-ref $code ,i) 0 $code) r))
           (reverse r)))
     $code))

(##define-macro (code-link c)     `(##vector-ref ,c 0))
(##define-macro (code-cprc c)     `(##vector-ref ,c 1))
(##define-macro (code-cte c)      `(##vector-ref ,c 2))
(##define-macro (code-length c)   `(##fixnum.- (##vector-length ,c) 3))
(##define-macro (code-ref c n)    `(##vector-ref ,c (##fixnum.+ ,n 3)))
(##define-macro (code-set! c n x) `(##vector-set! ,c (##fixnum.+ ,n 3) ,x))
(##define-macro (^ n)             `(##vector-ref $code ,(+ n 3)))

(define (##mk-code* code-prc cte lst n)
  (let ((code (##make-vector (##fixnum.+ (##length lst) (##fixnum.+ n 3)) #f)))
    (##vector-set! code 0 #f)
    (##vector-set! code 1 code-prc)
    (##vector-set! code 2 cte)
    (let loop ((i 0) (l lst))
      (if (##pair? l)
        (let ((child (##car l)))
          (##vector-set! child 0 code)
          (code-set! code i child)
          (loop (##fixnum.+ i 1) (##cdr l)))
        code))))

(##define-macro (code-run c)
  `(let (($$code ,c))
     ((##vector-ref $$code 1) $$code rte)))

; Macro to create the "code procedure" associated with a code node

(##define-macro (mk-cprc . def)
  `(lambda ($code rte) ,@def))

(##define-macro (mk-gen params . def)
  `(lambda (cte src tail? ,@params) ,@def))

(##define-macro (gen proc . args)
  `(,proc cte src tail? ,@args))

;==============================================================================

; Compiler

;------------------------------------------------------------------------------

; Compile time environments

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Representation of local variables (up and over) and global variable.

(##define-macro (mk-loc-access up over) `(##cons ,up ,over))
(##define-macro (loc-access? x) `(##pair? ,x))
(##define-macro (loc-access-up x) `(##car ,x))
(##define-macro (loc-access-over x) `(##cdr ,x))

(##define-macro (mk-glo-access id)
  `(or (make-global-var ,id)
       (ct-error-global-env-overflow ,id)))

(##define-macro (glo-access? x)
  `(##not (##pair? ,x)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Representation of compile time environments

; There are 4 types of structures in a compile time environment:
;
;    top        end of the environment and container for current state
;    frame      binding context for variables
;    macro      binding context for a macro
;    namespace  binding context for a namespace

(define (##cte-top top-cte)
  (##vector top-cte))

(define (##cte-top? cte)
  (##fixnum.= (##vector-length cte) 1))

(define (##cte-top-cte cte)
  (##vector-ref cte 0))

(define (##cte-top-cte-set! cte new-cte)
  (##vector-set! cte 0 new-cte))

(define (##cte-parent-cte cte)
  (##vector-ref cte 0))

(define (##cte-frame parent-cte vars)
  (##vector parent-cte vars))

(define (##cte-frame? cte)
  (##fixnum.= (##vector-length cte) 2))

(define (##cte-frame-vars cte)
  (##vector-ref cte 1))

(define (##cte-macro parent-cte name def)
  (##vector parent-cte name def))

(define (##cte-macro? cte)
  (and (##fixnum.= (##vector-length cte) 3)
       (##not (##string? (##vector-ref cte 1)))))

(define (##cte-macro-name cte)
  (##vector-ref cte 1))

(define (##cte-macro-def cte)
  (##vector-ref cte 2))

(define (##cte-namespace parent-cte prefix vars)
  (##vector parent-cte prefix vars))

(define (##cte-namespace? cte)
  (and (##fixnum.= (##vector-length cte) 3)
       (##string? (##vector-ref cte 1))))

(define (##cte-namespace-prefix cte)
  (##vector-ref cte 1))

(define (##cte-namespace-vars cte)
  (##vector-ref cte 2))

(define (##cte-relink cte new-parent-cte)
  (if new-parent-cte
    (cond ((##cte-frame? cte)
           (##cte-frame new-parent-cte
                        (##cte-frame-vars cte)))
          ((##cte-macro? cte)
           (##cte-macro new-parent-cte
                        (##cte-macro-name cte)
                        (##cte-macro-def cte)))
          ((##cte-namespace? cte)
           (##cte-namespace new-parent-cte
                            (##cte-namespace-prefix cte)
                            (##cte-namespace-vars cte))))
    #f))

(define (##cte-add-macro parent-cte name def)

  (define (replace cte)
    (cond ((##cte-top? cte)
           #f)
          ((and (##cte-macro? cte) (##eq? name (##cte-macro-name cte)))
           (##cte-macro (##cte-parent-cte cte) name def))
          ((##cte-namespace? cte)
           #f) ; don't go beyond a namespace declaration
          (else
           (##cte-relink cte (replace (##cte-parent-cte cte))))))

  (or (replace parent-cte)
      (##cte-macro parent-cte name def)))

(define (##cte-add-namespace parent-cte prefix vars)

  (define (replace cte)
    (cond ((##cte-top? cte)
           #f)
          ((##cte-namespace? cte)
           (if (##null? (##cte-namespace-vars cte))
             (##cte-namespace (##cte-parent-cte cte) prefix vars)
             (replace (##cte-parent-cte cte))))
          (else
           #f))) ; don't go beyond a frame or macro definition

  (if (##null? vars)
    (or (replace parent-cte)
        (##cte-namespace parent-cte prefix vars))
    (##cte-namespace parent-cte prefix vars)))

(define (##check-declarations src decls)

  (define (err) (ct-error-syntax "Ill-formed declaration"))

  (let loop ((decls decls))
    (if (##pair? decls)
      (let ((decl (##car decls)))
        (if (##pair? decl)
          (let ((head (##car decl)))
          (if (##eq? head 'NAMESPACE)
            (let ((rest (##cdr decl)))
              (if (##pair? rest)
                (let ((prefix (##car rest)))
                  (if (and (##string? prefix)
                           (##valid-prefix? prefix))
                    (##for-each (lambda (x)
                                  (if (##not (##symbol? x))
                                    (err)))
                                (##cdr rest))
                    (err)))
                (err))))))
        (loop (##cdr decls))))))

(define (##cte-add-declarations parent-cte src)
  (let ((decls (##cdr src)))
    (##check-declarations src decls)
    (let loop ((cte parent-cte) (decls decls))
      (if (##pair? decls)
        (let ((decl (##car decls)))
          (if (and (##pair? decl)
                   (##eq? (##car decl) 'NAMESPACE))
            (loop (##cte-add-namespace cte (##cadr decl) (##cddr decl))
                  (##cdr decls))
            (loop cte
                  (##cdr decls))))
        cte))))

(define (##cte-get-top-cte cte)
  (if (##cte-top? cte)
    cte
    (##cte-get-top-cte (##cte-parent-cte cte))))

(define (##cte-mutate-top-cte! cte proc)
  (let ((top-cte (##cte-get-top-cte cte)))
    (##cte-top-cte-set! top-cte (proc (##cte-top-cte top-cte)))))

(define (##make-top-cte)
  (let ((top-cte (##cte-top #f)))
    (##cte-top-cte-set! top-cte top-cte)
    top-cte))

(define (##top-cte-add-macro! top-cte name def)
  (let ((global-name (##cte-global-macro-name (##cte-top-cte top-cte) name)))
    (##cte-mutate-top-cte!
      top-cte
      (lambda (cte) (##cte-add-macro cte global-name def)))))

(define (##top-cte-add-declarations! top-cte src)
  (##cte-mutate-top-cte!
    top-cte
    (lambda (cte) (##cte-add-declarations cte src))))

(define (##top-cte-clone top-cte)
  (let ((new-top-cte (##cte-top #f)))

    (define (clone cte)
      (if (##cte-top? cte)
        new-top-cte
        (##cte-relink cte (clone (##cte-parent-cte cte)))))

    (##cte-top-cte-set! new-top-cte (clone (##cte-top-cte top-cte)))
    new-top-cte))

(define (##cte-lookup cte name macro-proc var-proc not-found-proc)
  (let loop ((name name) (full? (##full-name? name)) (cte cte) (up 0))
    (if (##cte-top? cte)
      (not-found-proc name)
      (let ((parent-cte (##cte-parent-cte cte)))
        (cond ((##cte-frame? cte)
               (let* ((vars (##cte-frame-vars cte))
                      (x (##memq name vars)))
                 (if x
                   (var-proc
                     name
                     up
                     (##fixnum.+ (##fixnum.- (##length vars) (##length x)) 1))
                   (loop name full? parent-cte (##fixnum.+ up 1)))))
              ((##cte-macro? cte)
               (if (##eq? name (##cte-macro-name cte))
                 (macro-proc name (##cte-macro-def cte))
                 (loop name full? parent-cte up)))
              ((and (##not full?) (##cte-namespace? cte))
               (let ((vars (##cte-namespace-vars cte)))
                 (if (or (##null? vars) (##memq name vars))
                   (loop (##make-full-name (##cte-namespace-prefix cte) name)
                         #t
                         parent-cte
                         up)
                   (loop name full? parent-cte up))))
              (else
               (loop name full? parent-cte up)))))))

(define (##cte-global-macro-name cte name)
  (if (##full-name? name)
    name
    (let loop ((cte cte))
      (if (##cte-top? cte)
        name
        (let ((parent-cte (##cte-parent-cte cte)))
          (cond ((##cte-namespace? cte)
                 (let ((vars (##cte-namespace-vars cte)))
                   (if (or (##null? vars) (##memq name vars))
                     (##make-full-name (##cte-namespace-prefix cte) name)
                     (loop parent-cte))))
                (else
                 (loop parent-cte))))))))

(define (##full-name? sym) ; full name if it contains a "#"
  (let ((str (##symbol->string sym)))
    (let loop ((i (##fixnum.- (##string-length str) 1)))
      (if (##fixnum.< i 0)
        #f
        (if (##char=? (##string-ref str i) #\#)
          #t
          (loop (##fixnum.- i 1)))))))

(define (##make-full-name prefix sym)
  (if (##fixnum.= (##string-length prefix) 0)
    sym
    (##string->symbol (##string-append prefix (##symbol->string sym)))))

(define (##valid-prefix? str)      ; non-null name followed by a "#" at end is
  (let ((l (##string-length str))) ; valid as is the special prefix ""
    (or (##fixnum.= l 0)
        (and (##not (##fixnum.< l 2))
             (##char=? (##string-ref str (##fixnum.- l 1)) #\#)))))

(define (##var-lookup cte name src)
  (let ((ind
         (##cte-lookup
           cte
           name
           (lambda (name def) #f)
           (lambda (name up over) (mk-loc-access up over))
           (lambda (name) (mk-glo-access name)))))
    (or ind
        (ct-error-syntax "Macro name can't be used as a variable:" name))))

(define ##macro-lookup #f)
(set! ##macro-lookup
  (lambda (cte name)
    (and (##symbol? name)
         (##cte-lookup
           cte
           name
           (lambda (name def) def)
           (lambda (name up over) #f)
           (lambda (name) #f)))))

(define ##macro-expand #f)
(set! ##macro-expand
  (lambda (cte src def)
    (##shape src src (##car def))
    (##apply (##cdr def) (##cdr src))))

(define ##macro-def #f)
(set! ##macro-def
  (lambda (src)
    (let ((src (##force-list src)))
      (if (and (##pair? src) (##eq? (##car src) 'LAMBDA))
        (begin
          (##shape src src -3)
          (##cons (##form-size (##force-list (##cadr src)))
                  (##eval-top src ##interaction-cte)))
        (ct-error-syntax "Macro value must be a lambda expression")))))

(define (##form-size parms)
  (let loop ((l parms) (n 1))
    (if (##pair? l)
      (loop (##cdr l) (##fixnum.+ n 1))
      (if (##null? l) n (##fixnum.- 0 n)))))

(define ##interaction-cte
  (##make-top-cte))

;------------------------------------------------------------------------------

; Utilities

(define (##self-eval? val)
  (force-vars (val)
    (or (##complex? val)
        (##string? val)
        (##char? val)
        (##keyword? val)
        (##memq val '(#f #t #!eof #!optional #!rest #!key #!)))))

(define (##variable src x)
  (if (##not (##symbol? x))
    (ct-error-syntax "Identifier expected:" x))
  (if (##memq x
              '(QUOTE QUASIQUOTE UNQUOTE UNQUOTE-SPLICING LAMBDA IF SET!
                COND => ELSE AND OR CASE LET LET* LETREC BEGIN DO DEFINE
                DELAY FUTURE ##DECLARE ##DEFINE-MACRO ##INCLUDE
                C-DECLARE C-INITIALIZE C-LAMBDA C-DEFINE))
    (ct-error-syntax "Variable name can not be a reserved identifier:" x)))

(define (##shape src x size)
  (let ((n (##proper-length x)))
    (if (or (##not n)
            (if (##fixnum.< 0 size)
              (##not (##fixnum.= n size))
              (##fixnum.< n (##fixnum.- 0 size))))
      (ct-error-syntax "Ill-formed special form:" (##car src)))))

(define (##proper-length l)

  (define (len l n)
    (cond ((##pair? l) (len (##cdr l) (##fixnum.+ n 1)))
          ((##null? l) n)
          (else        #f)))

  (len l 0))

(define (##force-list l)
  (if-forces
    (let loop ((l l))
      (force-vars (l)
        (if (##pair? l)
          (##cons (##car l) (loop (##cdr l)))
          l)))
    l))

(define (##read-expressions src filename)
  (if (##string? filename)

    (let ((port (##open-input-file filename)))
      (if port
        (##read-expressions-from-port
          port
          (lambda (script? exprs)
            exprs))
        (ct-error-syntax "File not found")))

    (ct-error-syntax "Filename expected")))

;------------------------------------------------------------------------------

; Compiler's main entry

(define (##compile-top top-cte src)
  (let ((tail? #f))
    (##comp-top top-cte src tail?)))

(define (##compile-inner cte src)
  (let ((tail? #f))
    (gen ##gen-top
      (##comp (##cte-frame cte (##list (self-var))) src tail?))))

;------------------------------------------------------------------------------

(define (##comp-top top-cte src tail?)
  (let ((src (##force-list src))
        (cte (##cte-top-cte top-cte)))
    (if (##pair? src)
      (let* ((first (##car src))
             (def (##macro-lookup cte first)))
        (if def
          (##comp-top top-cte (##macro-expand cte src def) tail?)
          (case first
            ((BEGIN)          (##comp-top-BEGIN top-cte src tail?))
            ((DEFINE)         (##comp-top-DEFINE top-cte src tail?))
            ((##DECLARE)      (##comp-top-DECLARE top-cte src tail?))
            ((##DEFINE-MACRO) (##comp-top-DEFINE-MACRO top-cte src tail?))
            ((##INCLUDE)      (##comp-top-INCLUDE top-cte src tail?))
            (else             (##comp-aux cte src tail? first)))))
      (##comp-simple cte src tail?))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-top-BEGIN top-cte src tail?)
  (##shape src src -1)
  (##comp-top-seq top-cte src tail? (##cdr src)))

(define (##comp-top-seq top-cte src tail? seq)
  (if (##pair? seq)
    (##comp-top-seq-aux top-cte src tail? seq)
    (let ((cte (##cte-top-cte top-cte)))
      (gen ##gen-cst (##void)))))

(define (##comp-top-seq-aux top-cte src tail? seq)
  (let ((rest (##cdr seq)))
    (if (##pair? rest)
      (let ((cte (##cte-top-cte top-cte)))
        (gen ##gen-seq
          (##comp-top top-cte (##car seq) #f)
          (##comp-top-seq-aux top-cte src tail? rest)))
      (##comp-top top-cte (##car seq) tail?))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-top-DEFINE top-cte src tail?)
  (let ((name (##definition-name src)))
    (##variable src name)
    (let* ((cte (##cte-top-cte top-cte))
           (ind (##var-lookup cte name src)))
      (gen ##gen-glo-def
        name
        ind
        (##comp cte (##definition-value src) #f)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-top-DECLARE top-cte src tail?)
  (##shape src src -1)
  (let ((cte (##cte-top-cte top-cte)))
    (##top-cte-add-declarations! top-cte src)
    (gen ##gen-cst (##void))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-top-DEFINE-MACRO top-cte src tail?)
  (let ((cte (##cte-top-cte top-cte))
        (name (##definition-name src)))
    (##top-cte-add-macro! top-cte name (##macro-def (##definition-value src)))
    (gen ##gen-cst (##void))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-top-INCLUDE top-cte src tail?)
  (##shape src src 2)
  (##comp-top-seq top-cte src tail? (##read-expressions src (##cadr src))))

;------------------------------------------------------------------------------

(define (##comp cte src tail?)
  (let ((src (##force-list src)))
    (if (##pair? src)
      (let* ((first (##car src))
             (def (##macro-lookup cte first)))
        (if def
          (##comp cte (##macro-expand cte src def) tail?)
          (case first
            ((BEGIN)          (##comp-BEGIN cte src tail?))
            ((DEFINE)         (ct-error-syntax "Ill-placed 'define'"))
            ((##DECLARE)      (ct-error-syntax "Ill-placed '##declare'"))
            ((##DEFINE-MACRO) (ct-error-syntax "Ill-placed '##define-macro'"))
            ((##INCLUDE)      (ct-error-syntax "Ill-placed '##include'"))
            (else             (##comp-aux cte src tail? first)))))
      (##comp-simple cte src tail?))))

(define (##comp-simple cte src tail?)
  (cond ((##symbol? src)    (##comp-ref cte src tail?))
        ((##self-eval? src) (##comp-cst cte src tail?))
        (else               (ct-error-syntax "Ill-formed expression"))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-BEGIN cte src tail?)
  (##shape src src -2)
  (##comp-seq cte src tail? (##cdr src)))

(define (##comp-seq cte src tail? seq)
  (if (##pair? seq)
    (##comp-seq-aux cte src tail? seq)
    (gen ##gen-cst (##void))))

(define (##comp-seq-aux cte src tail? seq)
  (let ((rest (##cdr seq)))
    (if (##pair? rest)
      (gen ##gen-seq
        (##comp cte (##car seq) #f)
        (##comp-seq-aux cte src tail? rest))
      (##comp cte (##car seq) tail?))))

;------------------------------------------------------------------------------

(define (##comp-aux cte src tail? first)
  (case first
    ((QUOTE)            (##comp-QUOTE cte src tail?))
    ((QUASIQUOTE)       (##comp-QUASIQUOTE cte src tail?))
    ((UNQUOTE)          (ct-error-syntax "Ill-placed 'unquote'"))
    ((UNQUOTE-SPLICING) (ct-error-syntax "Ill-placed 'unquote-splicing'"))
    ((SET!)             (##comp-SET! cte src tail?))
    ((LAMBDA)           (##comp-LAMBDA cte src tail?))
    ((IF)               (##comp-IF cte src tail?))
    ((COND)             (##comp-COND cte src tail?))
    ((AND)              (##comp-AND cte src tail?))
    ((OR)               (##comp-OR cte src tail?))
    ((CASE)             (##comp-CASE cte src tail?))
    ((LET)              (##comp-LET cte src tail?))
    ((LET*)             (##comp-LET* cte src tail?))
    ((LETREC)           (##comp-LETREC cte src tail?))
    ((DO)               (##comp-DO cte src tail?))
    ((DELAY)            (##comp-DELAY cte src tail?))
    ((FUTURE)           (##comp-FUTURE cte src tail?))
    ((C-DECLARE)        (ct-error-syntax "'c-declare' not supported by interpreter"))
    ((C-INITIALIZE)     (ct-error-syntax "'c-initialize' not supported by interpreter"))
    ((C-LAMBDA)         (ct-error-syntax "'c-lambda' not supported by interpreter"))
    ((C-DEFINE)         (ct-error-syntax "'c-define' not supported by interpreter"))
    (else               (##comp-app cte src tail?))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-ref cte src tail?)
  (##variable src src)
  (let ((x (##var-lookup cte src src)))
    (if (loc-access? x)
      (let ((up (loc-access-up x))
            (over (loc-access-over x)))
        (gen ##gen-loc-ref up over))
      (gen ##gen-glo-ref x))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-cst cte src tail?)
  (gen ##gen-cst src))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-QUOTE cte src tail?)
  (##shape src src 2)
  (gen ##gen-cst (##cadr src)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-QUASIQUOTE cte src tail?)
  (##comp-quasi cte src tail? (##force-list (##cadr src)) 1))

(define (##comp-quasi cte src tail? form level)
  (cond ((##eq? level 0)
         (##comp cte form tail?))
        ((##pair? form)
         (let ((x (##car form)))
           (force-vars (x)
             (case x
               ((QUASIQUOTE)
                (##comp-quasi-list cte src tail? form (##fixnum.+ level 1)))
               ((UNQUOTE)
                (if (##eq? level 1)
                  (begin
                    (##shape src form 2)
                    (##comp cte (##cadr form) tail?))
                  (##comp-quasi-list cte src tail? form (##fixnum.- level 1))))
               ((UNQUOTE-SPLICING)
                (if (##eq? level 1)
                  (ct-error-syntax "Ill-placed 'unquote-splicing'"))
                (##comp-quasi-list cte src tail? form (##fixnum.- level 1)))
               (else
                (##comp-quasi-list cte src tail? form level))))))
        ((##vector? form)
         (gen ##gen-quasi-list->vector
           (##comp-quasi-list cte src #f (##vector->list form) level)))
        (else
         (gen ##gen-cst form))))

(define (##comp-quasi-list cte src tail? l level)
  (if (##pair? l)
    (let ((first (##force-list (##car l))))
      (if (and (##eq? level 1) (##unquote-splicing? first))
        (begin
          (##shape src first 2)
          (if (##null? (##cdr l))
            (##comp cte (##cadr first) tail?)
            (gen ##gen-quasi-append
              (##comp cte (##cadr first) #f)
              (##comp-quasi cte src #f (##cdr l) 1))))
        (gen ##gen-quasi-cons
          (##comp-quasi cte src #f first level)
          (##comp-quasi cte src #f (##cdr l) level))))
    (##comp-quasi cte src tail? l level)))

(define (##unquote-splicing? x)
  (and (##pair? x)
       (let ((y (##car x)))
         (force-vars (y)
           (##eq? y 'UNQUOTE-SPLICING)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-SET! cte src tail?)
  (##shape src src 3)
  (let ((var (##cadr src)))
    (force-vars (var)
      (begin
        (##variable src var)
        (let ((x (##var-lookup cte var src)))
          (if (loc-access? x)
            (let ((up (loc-access-up x))
                  (over (loc-access-over x)))
              (gen ##gen-loc-set up over (##comp cte (##caddr src) #f)))
            (gen ##gen-glo-set x (##comp cte (##caddr src) #f))))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-LAMBDA cte src tail?)
  (##shape src src -3)
  (##comp-lambda-aux cte src tail? (##force-list (##cadr src)) (##cddr src)))

(define (##comp-lambda-aux cte src tail? parms body)
  (let ((frame (##parms->frame src parms)))
    (let ((c
           (##comp-body
             (##cte-frame cte (##cons (self-var) frame))
             src
             #t
             body)))
      (if (##rest-param? parms)
        (gen ##gen-prc-rest frame c)
        (gen ##gen-prc frame c)))))

(define (##parms->frame src parms)
  (cond ((##null? parms)
         '())
        ((##pair? parms)
         (let ((x (##car parms)))
           (force-vars (x)
             (let ((rest (##parms->frame src (##cdr parms))))
               (##variable src x)
               (if (##memq x rest)
                 (ct-error-syntax "Duplicate parameter in parameter list"))
               (##cons x rest)))))
        (else
         (##variable src parms)
         (##list parms))))

(define (##rest-param? parms)
  (cond ((##pair? parms)
         (##rest-param? (##cdr parms)))
        ((##null? parms)
         #f)
        (else
         #t)))

(define (##comp-body cte src tail? body)

  (define (letrec-defines cte vars vals body)
    (if (##pair? body)

      (let ((src (##force-list (##car body))))
        (if (##not (##pair? src))
          (letrec-defines* cte vars vals body)
          (let ((first (##car src)))
            (force-vars (first)
              (let ((def (##macro-lookup cte first)))
                (if def
                  (letrec-defines cte
                                  vars
                                  vals
                                  (##cons (##macro-expand cte src def)
                                          (##cdr body)))
                  (case first
                    ((BEGIN)
                     (letrec-defines cte
                                     vars
                                     vals
                                     (##append (##cdr src) (##cdr body))))
                    ((DEFINE)
                     (let ((x (##definition-name src)))
                       (##variable src x)
                       (if (##memq x vars)
                         (ct-error-syntax "Duplicate definition of a variable"))
                       (letrec-defines cte
                                       (##cons x vars)
                                       (##cons (##definition-value src) vals)
                                       (##cdr body))))
                    ((##DECLARE)
                     (##shape src src -1)
                     (letrec-defines (##cte-add-declarations cte src)
                                     vars
                                     vals
                                     (##cdr body)))
                    ((##DEFINE-MACRO)
                     (let ((x (##definition-name src)))
                       (letrec-defines (##cte-macro
                                         cte
                                         x
                                         (##macro-def (##definition-value src)))
                                       vars
                                       vals
                                       (##cdr body))))
                    ((##INCLUDE)
                     (##shape src src 2)
                     (letrec-defines cte
                                     vars
                                     vals
                                     (##append (##read-expressions src (##cadr src))
                                               (##cdr body))))
                    (else
                     (letrec-defines* cte vars vals body)))))))))

      (ct-error-syntax "Body must contain at least one evaluable expression")))

  (define (letrec-defines* cte vars vals body)
    (if (##null? vars)
      (##comp-seq cte src tail? body)
      (##comp-letrec-aux cte src tail? vars vals body)))

  (letrec-defines cte '() '() body))

(define (##definition-name src)
  (##shape src src -3)
  (let ((pattern (##cadr src)))
    (force-vars (pattern)
      (let ((name (if (##pair? pattern)
                    (let ((name (##car pattern)))
                      (force-vars (name)
                        name))
                    (begin
                      (##shape src src 3)
                      pattern))))
        (if (##not (##symbol? name))
          (ct-error-syntax "Defined variable must be an identifier"))
        name))))

(define (##definition-value src)
  (let ((pattern (##cadr src)))
    (force-vars (pattern)
      (if (##pair? pattern)
        (##cons 'LAMBDA (##cons (##cdr pattern) (##cddr src)))
        (##caddr src)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-IF cte src tail?)
  (##shape src src -3)
  (if (##pair? (##cdddr src))
    (begin
      (##shape src src 4)
      (gen ##gen-if3
        (##comp cte (##cadr src) #f)
        (##comp cte (##caddr src) tail?)
        (##comp cte (##cadddr src) tail?)))
    (begin
      (##shape src src 3)
      (gen ##gen-if2
        (##comp cte (##cadr src) #f)
        (##comp cte (##caddr src) tail?)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-COND cte src tail?)
  (##shape src src -2)
  (##comp-cond-aux cte src tail? (##cdr src)))

(define (##comp-cond-aux cte src tail? clauses)
  (if (##pair? clauses)
    (let ((clause (##force-list (##car clauses))))
      (##shape src clause -1)
      (let ((x (##car clause)))
        (force-vars (x)
          (cond ((##eq? x 'ELSE)
                 (##shape src clause -2)
                 (if (##not (##null? (##cdr clauses)))
                   (ct-error-syntax "ELSE clause must be last"))
                 (##comp-seq cte src tail? (##cdr clause)))
                ((##not (##pair? (##cdr clause)))
                 (gen ##gen-cond-or
                    (##comp cte (##car clause) #f)
                    (##comp-cond-aux cte src tail? (##cdr clauses))))
                (else
                 (let ((y (##cadr clause)))
                   (force-vars (y)
                     (if (##eq? y '=>)
                       (begin
                         (##shape src clause -3)
                         (gen ##gen-cond-send
                           (##comp cte (##car clause) #f)
                           (##comp cte (##caddr clause) #f)
                           (##comp-cond-aux cte src tail? (##cdr clauses))))
                       (gen ##gen-cond-if
                         (##comp cte (##car clause) #f)
                         (##comp-seq cte src tail? (##cdr clause))
                         (##comp-cond-aux cte src tail? (##cdr clauses)))))))))))
    (gen ##gen-cst (##void))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-AND cte src tail?)
  (let ((rest (##cdr src)))
    (if (##pair? rest)
      (##comp-and-aux cte src tail? rest)
      (gen ##gen-cst #t))))

(define (##comp-and-aux cte src tail? l)
  (let ((rest (##cdr l)))
    (if (##pair? rest)
      (gen ##gen-and
        (##comp cte (##car l) #f)
        (##comp-and-aux cte src tail? rest))
      (##comp cte (##car l) tail?))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-OR cte src tail?)
  (let ((rest (##cdr src)))
    (if (##pair? rest)
      (##comp-or-aux cte src tail? rest)
      (gen ##gen-cst #f))))

(define (##comp-or-aux cte src tail? l)
  (let ((rest (##cdr l)))
    (if (##pair? rest)
      (gen ##gen-or
        (##comp cte (##car l) #f)
        (##comp-or-aux cte src tail? rest))
      (##comp cte (##car l) tail?))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-CASE cte src tail?)
  (##shape src src -3)
  (gen ##gen-case
    (##comp cte (##cadr src) #f)
    (let ((cte (##cte-frame cte (##list (selector-var)))))
      (##comp-case-aux cte src tail? (##cddr src)))))

(define (##comp-case-aux cte src tail? clauses)
  (if (##pair? clauses)
    (let ((clause (##force-list (##car clauses))))
      (##shape src clause -2)
      (let ((first (##force-list (##car clause))))
        (if (##eq? first 'ELSE)
          (begin
            (if (##not (##null? (##cdr clauses)))
              (ct-error-syntax "ELSE clause must be last"))
            (gen ##gen-case-else
              (##comp-seq cte src tail? (##cdr clause))))
          (gen ##gen-case-clause
            first
            (##comp-seq cte src tail? (##cdr clause))
            (##comp-case-aux cte src tail? (##cdr clauses))))))
    (gen ##gen-case-else
      (gen ##gen-cst (##void)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-LET cte src tail?)
  (##shape src src -3)
  (let ((x (##force-list (##cadr src))))
    (cond ((##symbol? x)
           (##shape src src -4)
           (let ((bindings (##force-list (##caddr src))))
             (let* ((vars (##bindings->vars src bindings #t))
                    (vals (##bindings->vals bindings)))
               (gen ##gen-app
                 (let ((inner-cte (##cte-frame cte (##list x))))
                   (gen ##gen-letrec
                     (##list x)
                     (let ((cte inner-cte)
                           (tail? #f))
                       (##list (gen ##gen-prc
                               vars
                               (##comp-body (##cte-frame
                                              cte
                                              (##cons (self-var) vars))
                                            src
                                            #t
                                            (##cdddr src)))))
                     (let ((cte inner-cte)
                           (tail? #f))
                       (gen ##gen-loc-ref 0 1)))) ; fetch loop variable
                 (##comp-vals cte vals)))))
          ((##null? x)
           (##comp-body cte src tail? (##cddr src)))
          (else
           (let* ((bindings x)
                  (vars (##bindings->vars src bindings #t))
                  (vals (##bindings->vals bindings)))
             (let ((c
                    (##comp-body
                      (##cte-frame cte vars)
                      src
                      tail?
                      (##cddr src))))
               (gen ##gen-let
                 vars
                 (##comp-vals cte vals)
                 c)))))))

(define (##comp-vals cte l)
  (if (##pair? l)
    (##cons (##comp cte (##car l) #f) (##comp-vals cte (##cdr l)))
    '()))

(define (##bindings->vars src bindings check-duplicates?)
  (if (##pair? bindings)
    (let ((binding (##force-list (##car bindings))))
      (##shape src binding 2)
      (let ((x (##car binding)))
        (force-vars (x)
          (let ((rest (##bindings->vars src (##cdr bindings) check-duplicates?)))
            (##variable src x)
            (if (and check-duplicates? (##memq x rest))
              (ct-error-syntax "Duplicate variable in bindings"))
            (##cons x rest)))))
    (if (##null? bindings)
      '()
      (ct-error-syntax "Ill-terminated bindings"))))

(define (##bindings->vals bindings)
  (if (##pair? bindings)
    (let ((binding (##force-list (##car bindings))))
      (##cons (##cadr binding) (##bindings->vals (##cdr bindings))))
    '()))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-LET* cte src tail?)
  (##shape src src -3)
  (let ((bindings (##cadr src)))
    (force-vars (bindings)
      (let* ((vars (##bindings->vars src bindings #f))
             (vals (##bindings->vals bindings)))
        (##comp-let*-aux cte src tail? vars vals (##cddr src))))))

(define (##comp-let*-aux cte src tail? vars vals body)
  (if (##pair? vars)
    (let ((frame (##list (##car vars))))
      (let ((inner-cte (##cte-frame cte frame)))
        (gen ##gen-let
          frame
          (##list (##comp cte (##car vals) #f))
          (##comp-let*-aux inner-cte src tail? (##cdr vars) (##cdr vals) body))))
    (##comp-body cte src tail? body)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-LETREC cte src tail?)
  (##shape src src -3)
  (let ((bindings (##force-list (##cadr src))))
    (if (##null? bindings)
      (##comp-body cte src tail? (##cddr src))
      (let* ((vars (##bindings->vars src bindings #t))
             (vals (##bindings->vals bindings)))
        (##comp-letrec-aux cte src tail? vars vals (##cddr src))))))

(define (##comp-letrec-aux cte src tail? vars vals body)
  (if (##pair? vars)
    (let ((inner-cte (##cte-frame cte vars)))
      (gen ##gen-letrec
        vars
        (##comp-vals inner-cte vals)
        (##comp-body inner-cte src tail? body)))
    (##comp-body cte src tail? body)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-DO cte src tail?)
  (##shape src src -3)
  (let ((bindings (##force-list (##cadr src)))
        (exit (##force-list (##caddr src))))
    (##shape src exit -1)
    (let* ((vars (##bindings->vars* src bindings))
           (do-loop-vars (##list (do-loop-var)))
           (inner-cte (##cte-frame cte do-loop-vars)))
      (gen ##gen-letrec
        do-loop-vars
        (##list
          (let ((cte inner-cte)
                (tail? #f))
            (gen ##gen-prc
              vars
              (let ((cte (##cte-frame cte (##cons (self-var) vars)))
                    (tail? #t))
                (gen ##gen-if3
                  (##comp cte (##car exit) #f)
                  (##comp-seq cte src tail? (##cdr exit))
                  (let ((call
                          (gen ##gen-app
                            (let ((tail? #f))
                              (gen ##gen-loc-ref 1 1)) ; fetch do-loop-var
                            (##comp-vals cte (##bindings->steps bindings)))))
                    (if (##null? (##cdddr src))
                      call
                      (gen ##gen-seq
                        (##comp-seq cte src #f (##cdddr src))
                        call))))))))
        (let ((cte inner-cte))
          (gen ##gen-app
            (let ((tail? #f))
              (gen ##gen-loc-ref 0 1)) ; fetch do-loop-var
            (##comp-vals cte (##bindings->vals bindings))))))))

(define (##bindings->vars* src bindings)
  (if (##pair? bindings)
    (let ((binding (##force-list (##car bindings))))
      (##shape src binding -2)
      (if (##pair? (##cddr binding)) (##shape src binding 3))
      (let ((x (##car binding)))
        (force-vars (x)
          (let ((rest (##bindings->vars* src (##cdr bindings))))
            (##variable src x)
            (if (##memq x rest)
              (ct-error-syntax "Duplicate variable in bindings"))
            (##cons x rest)))))
    (if (##null? bindings)
      '()
      (ct-error-syntax "Ill-terminated bindings"))))

(define (##bindings->steps bindings)
  (if (##pair? bindings)
    (let ((binding (##force-list (##car bindings))))
      (##cons (if (##pair? (##cddr binding)) (##caddr binding) (##car binding))
              (##bindings->steps (##cdr bindings))))
    '()))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-app cte src tail?)
  (let ((n (##proper-length src)))
    (if n
      (gen ##gen-app
        (##comp cte (##car src) #f)
        (##comp-vals cte (##cdr src)))
      (ct-error-syntax "Ill-formed procedure application"))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-DELAY cte src tail?)
  (##shape src src 2)
  (gen ##gen-delay (##comp cte (##cadr src) #t)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-FUTURE cte src tail?)
  (##shape src src 2)
  (gen ##gen-future (##comp cte (##cadr src) #t)))

;==============================================================================

; Code generation procedures

;------------------------------------------------------------------------------

; Macros to manipulate the runtime environment

(##define-macro (mk-rte rte . lst)
  `(##vector ,rte ,@lst))

(##define-macro (mk-rte* rte n)
  `(let (($rte (##make-vector (##fixnum.+ ,n 1) (##void))))
     (##vector-set! $rte 0 ,rte)
     $rte))

(##define-macro (rte-up rte)         `(##vector-ref ,rte 0))
(##define-macro (rte-ref rte i)      `(##vector-ref ,rte ,i))
(##define-macro (rte-set! rte i val) `(##vector-set! ,rte ,i ,val))

;------------------------------------------------------------------------------

(define ##cprc-top
  (mk-cprc
    (let ((rte (mk-rte rte #f)))
      (code-run (^ 0)))))

(define ##gen-top
  (mk-gen (val)
    (mk-code ##cprc-top cte (val))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-cst-null  (mk-cprc '()))
(define ##cprc-cst-true  (mk-cprc #t))
(define ##cprc-cst-false (mk-cprc #f))
(define ##cprc-cst--2    (mk-cprc -2))
(define ##cprc-cst--1    (mk-cprc -1))
(define ##cprc-cst-0     (mk-cprc 0))
(define ##cprc-cst-1     (mk-cprc 1))
(define ##cprc-cst-2     (mk-cprc 2))
(define ##cprc-cst       (mk-cprc (^ 0)))

(define ##gen-cst
  (mk-gen (val)
    (case val
      ((()) (mk-code ##cprc-cst-null  cte ()))
      ((#t) (mk-code ##cprc-cst-true  cte ()))
      ((#f) (mk-code ##cprc-cst-false cte ()))
      ((-2) (mk-code ##cprc-cst--2    cte ()))
      ((-1) (mk-code ##cprc-cst--1    cte ()))
      ((0)  (mk-code ##cprc-cst-0     cte ()))
      ((1)  (mk-code ##cprc-cst-1     cte ()))
      ((2)  (mk-code ##cprc-cst-2     cte ()))
      (else (mk-code ##cprc-cst       cte () val)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-loc-ref-0-1 (mk-cprc (rte-ref rte 1)))
(define ##cprc-loc-ref-0-2 (mk-cprc (rte-ref rte 2)))
(define ##cprc-loc-ref-0-3 (mk-cprc (rte-ref rte 3)))

(define ##cprc-loc-ref-1-1 (mk-cprc (rte-ref (rte-up rte) 1)))
(define ##cprc-loc-ref-1-2 (mk-cprc (rte-ref (rte-up rte) 2)))
(define ##cprc-loc-ref-1-3 (mk-cprc (rte-ref (rte-up rte) 3)))

(define ##cprc-loc-ref-2-1 (mk-cprc (rte-ref (rte-up (rte-up rte)) 1)))
(define ##cprc-loc-ref-2-2 (mk-cprc (rte-ref (rte-up (rte-up rte)) 2)))
(define ##cprc-loc-ref-2-3 (mk-cprc (rte-ref (rte-up (rte-up rte)) 3)))

(define ##cprc-loc-ref
  (mk-cprc
    (let loop ((e rte) (i (^ 0)))
      (if (##fixnum.< 0 i)
        (loop (rte-up e) (##fixnum.- i 1))
        (rte-ref e (^ 1))))))

(define ##gen-loc-ref
  (mk-gen (up over)
    (case up
      ((0)
       (case over
         ((1)  (mk-code ##cprc-loc-ref-0-1 cte ()))
         ((2)  (mk-code ##cprc-loc-ref-0-2 cte ()))
         ((3)  (mk-code ##cprc-loc-ref-0-3 cte ()))
         (else (mk-code ##cprc-loc-ref     cte () up over))))
      ((1)
       (case over
         ((1)  (mk-code ##cprc-loc-ref-1-1 cte ()))
         ((2)  (mk-code ##cprc-loc-ref-1-2 cte ()))
         ((3)  (mk-code ##cprc-loc-ref-1-3 cte ()))
         (else (mk-code ##cprc-loc-ref     cte () up over))))
      ((2)
       (case over
         ((1)  (mk-code ##cprc-loc-ref-2-1 cte ()))
         ((2)  (mk-code ##cprc-loc-ref-2-2 cte ()))
         ((3)  (mk-code ##cprc-loc-ref-2-3 cte ()))
         (else (mk-code ##cprc-loc-ref     cte () up over))))
     (else
       (mk-code ##cprc-loc-ref cte () up over)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-glo-ref
  (mk-cprc
    (let loop ((val (global-var-ref (^ 0))))
      (if (unbound? val)
        (loop (rt-error-unbound-global-var $code rte))
        val))))

(define ##gen-glo-ref
  (mk-gen (ind)
    (mk-code ##cprc-glo-ref cte () ind)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-loc-set
  (mk-cprc
    (let ((val (code-run (^ 0))))
      (let loop ((e rte) (i (^ 1)))
        (if (##fixnum.< 0 i)
          (loop (rte-up e) (##fixnum.- i 1))
          (begin
            (rte-set! e (^ 2) val)
            (##void)))))))

(define ##gen-loc-set
  (mk-gen (up over val)
    (mk-code ##cprc-loc-set cte (val) up over)))

(define ##cprc-glo-set
  (mk-cprc
    (let ((val (code-run (^ 0))))
      (global-var-set! (^ 1) val)
      (##void))))

(define ##gen-glo-set
  (mk-gen (ind val)
    (mk-code ##cprc-glo-set cte (val) ind)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-glo-def
  (mk-cprc
    (let ((rte #f))
      (global-var-set! (^ 1) (code-run (^ 0)))
      (##void))))

(define ##gen-glo-def
  (mk-gen (name ind val)
    (mk-code ##cprc-glo-def cte (val) ind name)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-if2
  (mk-cprc
    (let ((pred (code-run (^ 0))))
      (force-vars (pred)
        (if (true? pred)
          (code-run (^ 1))
          (##void))))))

(define ##gen-if2
  (mk-gen (pre con)
    (mk-code ##cprc-if2 cte (pre con))))

(define ##cprc-if3
  (mk-cprc
    (let ((pred (code-run (^ 0))))
      (force-vars (pred)
        (if (true? pred)
          (code-run (^ 1))
          (code-run (^ 2)))))))

(define ##gen-if3
  (mk-gen (pre con alt)
    (mk-code ##cprc-if3 cte (pre con alt))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-seq
  (mk-cprc
    (code-run (^ 0))
    (code-run (^ 1))))

(define ##gen-seq
  (mk-gen (val1 val2)
    (mk-code ##cprc-seq cte (val1 val2))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-quasi-list->vector
  (mk-cprc
    (quasi-list->vector (code-run (^ 0)))))

(define ##gen-quasi-list->vector
  (mk-gen (val)
    (mk-code ##cprc-quasi-list->vector cte (val))))

(define ##cprc-quasi-append
  (mk-cprc
    (quasi-append (code-run (^ 0)) (code-run (^ 1)))))

(define ##gen-quasi-append
  (mk-gen (val1 val2)
    (mk-code ##cprc-quasi-append cte (val1 val2))))

(define ##cprc-quasi-cons
  (mk-cprc
    (quasi-cons (code-run (^ 0)) (code-run (^ 1)))))

(define ##gen-quasi-cons
  (mk-gen (val1 val2)
    (mk-code ##cprc-quasi-cons cte (val1 val2))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-cond-if
  (mk-cprc
    (let ((pred (code-run (^ 0))))
      (force-vars (pred)
        (if (true? pred)
          (code-run (^ 1))
          (code-run (^ 2)))))))

(define ##gen-cond-if
  (mk-gen (val1 val2 val3)
    (mk-code ##cprc-cond-if cte (val1 val2 val3))))

(define ##cprc-cond-or
  (mk-cprc
    (let ((pred (code-run (^ 0))))
      (force-vars (pred)
        (if (true? pred)
          pred
          (code-run (^ 1)))))))

(define ##gen-cond-or
  (mk-gen (val1 val2)
    (mk-code ##cprc-cond-or cte (val1 val2))))

(define ##cprc-cond-send-red
  (mk-cprc
    (let ((pred (code-run (^ 0))))
      (force-vars (pred)
        (if (true? pred)
          (let loop ((proc (code-run (^ 1))))
            (force-vars (proc)
              (if (##not (##procedure? proc))
                (loop (rt-error-non-procedure-send $code rte))
                (##reduction-apply1 $code rte proc pred))))
          (code-run (^ 2)))))))

(define ##cprc-cond-send-sub
  (mk-cprc
    (let ((pred (code-run (^ 0))))
      (force-vars (pred)
        (if (true? pred)
          (let loop ((proc (code-run (^ 1))))
            (force-vars (proc)
              (if (##not (##procedure? proc))
                (loop (rt-error-non-procedure-send $code rte))
                (##subproblem-apply1 $code rte proc pred))))
          (code-run (^ 2)))))))

(define ##gen-cond-send
  (mk-gen (val1 val2 val3)
    (mk-code (if tail? ##cprc-cond-send-red ##cprc-cond-send-sub)
             cte
             (val1 val2 val3))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-or
  (mk-cprc
    (let ((pred (code-run (^ 0))))
      (force-vars (pred)
        (if (true? pred)
          pred
          (code-run (^ 1)))))))

(define ##gen-or
  (mk-gen (val1 val2)
    (mk-code ##cprc-or cte (val1 val2))))

(define ##cprc-and
  (mk-cprc
    (let ((pred (code-run (^ 0))))
      (force-vars (pred)
        (if (##not (true? pred))
          pred
          (code-run (^ 1)))))))

(define ##gen-and
  (mk-gen (val1 val2)
    (mk-code ##cprc-and cte (val1 val2))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-case
  (mk-cprc
    (let ((selector (code-run (^ 0))))
      (force-vars (selector)
        (let ((rte (mk-rte rte selector)))
          (code-run (^ 1)))))))

(define ##gen-case
  (mk-gen (val1 val2)
    (mk-code ##cprc-case cte (val1 val2))))

(define ##cprc-case-clause
  (mk-cprc
    (if (##case-memv (rte-ref rte 1) (^ 2))
      (code-run (^ 0))
      (code-run (^ 1)))))

(define ##gen-case-clause
  (mk-gen (cases val1 val2)
    (mk-code ##cprc-case-clause cte (val1 val2) cases)))

(define ##cprc-case-else
  (mk-cprc
    (code-run (^ 0))))

(define ##gen-case-else
  (mk-gen (val)
    (mk-code ##cprc-case-else cte (val))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-let
  (mk-cprc
    (let ((n (##fixnum.- (code-length $code) 2)))
      (let loop1 ((i 1) (args '()))
        (if (##fixnum.< n i)
          (let ((inner-rte (mk-rte* rte n)))
            (let loop2 ((i n) (args args))
              (if (##fixnum.< 0 i)
                (begin
                  (rte-set! inner-rte i (##car args))
                  (loop2 (##fixnum.- i 1) (##cdr args)))
                (let ((rte inner-rte))
                  (code-run (^ 0))))))
          (loop1 (##fixnum.+ i 1) (##cons (code-run (code-ref $code i)) args)))))))

(define ##gen-let
  (mk-gen (vars vals body)
    (let ((c (##mk-code* ##cprc-let cte (##cons body vals) 1)))
      (code-set! c (##fixnum.+ (##length vals) 1) vars)
      c)))

(define ##cprc-letrec
  (mk-cprc
    (let ((n (##fixnum.- (code-length $code) 2)))
      (let ((rte (mk-rte* rte n)))
        (let loop ((i n))
          (if (##fixnum.< 0 i)
            (begin
              (rte-set! rte i (code-run (code-ref $code i)))
              (loop (##fixnum.- i 1)))
            (code-run (^ 0))))))))

(define ##gen-letrec
  (mk-gen (vars vals body)
    (let ((c (##mk-code* ##cprc-letrec cte (##cons body vals) 1)))
      (code-set! c (##fixnum.+ (##length vals) 1) vars)
      c)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-prc0
  (mk-cprc
    (letrec ((proc
               (lambda ()
                 (let ((rte (mk-rte rte proc)))
                   (code-run (^ 0))))))
      proc)))

(define ##cprc-prc1
  (mk-cprc
    (letrec ((proc
               (lambda (arg1)
                 (let ((rte (mk-rte rte proc arg1)))
                   (code-run (^ 0))))))
      proc)))

(define ##cprc-prc2
  (mk-cprc
    (letrec ((proc
               (lambda (arg1 arg2)
                 (let ((rte (mk-rte rte proc arg1 arg2)))
                   (code-run (^ 0))))))
      proc)))

(define ##cprc-prc3
  (mk-cprc
    (letrec ((proc
               (lambda (arg1 arg2 arg3)
                 (let ((rte (mk-rte rte proc arg1 arg2 arg3)))
                   (code-run (^ 0))))))
      proc)))

(define ##cprc-prc
  (mk-cprc
    (letrec ((proc
               (lambda args
                 (let ((n (^ 1)))
                   (let ((inner-rte (mk-rte* rte n)))
                     (rte-set! inner-rte 1 proc)
                     (let loop ((i 2) (l args))
                       (if (##fixnum.< n i)
                         (if (##pair? l)
                           (rt-error-too-many-args proc args)
                           (let ((rte inner-rte))
                             (code-run (^ 0))))
                         (if (##pair? l)
                           (begin
                             (rte-set! inner-rte i (##car l))
                             (loop (##fixnum.+ i 1) (##cdr l)))
                           (rt-error-too-few-args proc args)))))))))
      proc)))

(define ##gen-prc
  (mk-gen (frame body)
    (case (##length frame)
      ((0)  (mk-code ##cprc-prc0 cte (body) frame))
      ((1)  (mk-code ##cprc-prc1 cte (body) frame))
      ((2)  (mk-code ##cprc-prc2 cte (body) frame))
      ((3)  (mk-code ##cprc-prc3 cte (body) frame))
      (else (mk-code ##cprc-prc  cte (body) (##fixnum.+ (##length frame) 1) frame)))))

(define ##cprc-prc-rest
  (mk-cprc
    (letrec ((proc
               (lambda args
                 (let ((n (^ 1)))
                   (let ((inner-rte (mk-rte* rte n)))
                     (rte-set! inner-rte 1 proc)
                     (let loop ((i 2) (l args))
                       (if (##fixnum.< i n)
                         (if (##pair? l)
                           (begin
                             (rte-set! inner-rte i (##car l))
                             (loop (##fixnum.+ i 1) (##cdr l)))
                           (rt-error-too-few-args proc args))
                         (begin
                           (rte-set! inner-rte i l)
                           (let ((rte inner-rte))
                             (code-run (^ 0)))))))))))
      proc)))

(define ##gen-prc-rest
  (mk-gen (frame body)
    (mk-code ##cprc-prc-rest cte (body) (##fixnum.+ (##length frame) 1) frame)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-app0-red
  (mk-cprc
    (let ((proc (code-run (^ 0))))
      (force-vars (proc)
        (if (##not (##procedure? proc))
          (rt-error-non-procedure-oper $code rte)
          (##reduction-apply0 $code rte proc))))))

(define ##cprc-app1-red
  (mk-cprc
    (let ((proc (code-run (^ 0))))
      (force-vars (proc)
        (if (##not (##procedure? proc))
          (rt-error-non-procedure-oper $code rte)
          (let ((arg1 (code-run (^ 1))))
            (##reduction-apply1 $code rte proc arg1)))))))

(define ##cprc-app2-red
  (mk-cprc
    (let ((proc (code-run (^ 0))))
      (force-vars (proc)
        (if (##not (##procedure? proc))
          (rt-error-non-procedure-oper $code rte)
          (let ((arg1 (code-run (^ 1)))
                (arg2 (code-run (^ 2))))
            (##reduction-apply2 $code rte proc arg1 arg2)))))))

(define ##cprc-app3-red
  (mk-cprc
    (let ((proc (code-run (^ 0))))
      (force-vars (proc)
        (if (##not (##procedure? proc))
          (rt-error-non-procedure-oper $code rte)
          (let ((arg1 (code-run (^ 1)))
                (arg2 (code-run (^ 2)))
                (arg3 (code-run (^ 3))))
            (##reduction-apply3 $code rte proc arg1 arg2 arg3)))))))

(define ##cprc-app-red
  (mk-cprc
    (let ((proc (code-run (^ 0))))
      (force-vars (proc)
        (if (##not (##procedure? proc))
          (rt-error-non-procedure-oper $code rte)
          (let loop ((i (##fixnum.- (code-length $code) 1)) (args '()))
            (if (##fixnum.< 0 i)
              (loop (##fixnum.- i 1) (##cons (code-run (code-ref $code i)) args))
              (##reduction-apply $code rte proc args))))))))

(define ##cprc-app0-sub
  (mk-cprc
    (let ((proc (code-run (^ 0))))
      (force-vars (proc)
        (if (##not (##procedure? proc))
          (rt-error-non-procedure-oper $code rte)
          (##subproblem-apply0 $code rte proc))))))

(define ##cprc-app1-sub
  (mk-cprc
    (let ((proc (code-run (^ 0))))
      (force-vars (proc)
        (if (##not (##procedure? proc))
          (rt-error-non-procedure-oper $code rte)
          (let ((arg1 (code-run (^ 1))))
            (##subproblem-apply1 $code rte proc arg1)))))))

(define ##cprc-app2-sub
  (mk-cprc
    (let ((proc (code-run (^ 0))))
      (force-vars (proc)
        (if (##not (##procedure? proc))
          (rt-error-non-procedure-oper $code rte)
          (let ((arg1 (code-run (^ 1)))
                (arg2 (code-run (^ 2))))
            (##subproblem-apply2 $code rte proc arg1 arg2)))))))

(define ##cprc-app3-sub
  (mk-cprc
    (let ((proc (code-run (^ 0))))
      (force-vars (proc)
        (if (##not (##procedure? proc))
          (rt-error-non-procedure-oper $code rte)
          (let ((arg1 (code-run (^ 1)))
                (arg2 (code-run (^ 2)))
                (arg3 (code-run (^ 3))))
            (##subproblem-apply3 $code rte proc arg1 arg2 arg3)))))))

(define ##cprc-app-sub
  (mk-cprc
    (let ((proc (code-run (^ 0))))
      (force-vars (proc)
        (if (##not (##procedure? proc))
          (rt-error-non-procedure-oper $code rte)
          (let loop ((i (##fixnum.- (code-length $code) 1)) (args '()))
            (if (##fixnum.< 0 i)
              (loop (##fixnum.- i 1) (##cons (code-run (code-ref $code i)) args))
              (##subproblem-apply $code rte proc args))))))))

(define ##properly-tail-recursive? #f)
(set! ##properly-tail-recursive? #t)

(define ##gen-app
  (mk-gen (oper args)
    (if (and tail? ##properly-tail-recursive?)
      (case (##length args)
        ((0)
         (mk-code ##cprc-app0-red
                  cte
                  (oper)))
        ((1)
         (mk-code ##cprc-app1-red
                  cte
                  (oper (##car args))))
        ((2)
         (mk-code ##cprc-app2-red
                  cte
                  (oper (##car args) (##cadr args))))
        ((3)
         (mk-code ##cprc-app3-red
                  cte
                  (oper (##car args) (##cadr args) (##caddr args))))
        (else
         (##mk-code* ##cprc-app-red
                     cte
                     (##cons oper args)
                     0)))
      (case (##length args)
        ((0)
         (mk-code ##cprc-app0-sub
                  cte
                  (oper)))
        ((1)
         (mk-code ##cprc-app1-sub
                  cte
                  (oper (##car args))))
        ((2)
         (mk-code ##cprc-app2-sub
                  cte
                  (oper (##car args) (##cadr args))))
        ((3)
         (mk-code ##cprc-app3-sub
                  cte
                  (oper (##car args) (##cadr args) (##caddr args))))
        (else
         (##mk-code* ##cprc-app-sub
                     cte
                     (##cons oper args)
                     0))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##reduction-apply0 $code rte proc)
  (##declare (interrupts-enabled))
  (proc))

(define (##reduction-apply1 $code rte proc arg1)
  (##declare (interrupts-enabled))
  (proc arg1))

(define (##reduction-apply2 $code rte proc arg1 arg2)
  (##declare (interrupts-enabled))
  (proc arg1 arg2))

(define (##reduction-apply3 $code rte proc arg1 arg2 arg3)
  (##declare (interrupts-enabled))
  (proc arg1 arg2 arg3))

(define (##reduction-apply $code rte proc args)
  (##declare (interrupts-enabled))
  (##apply proc args))

(define ##subproblem-apply0
  (let ()
    (##declare (not inline))
    (lambda ($code rte proc)
      (##declare (not interrupts-enabled))
      (##identity (let ()
                    (##declare (interrupts-enabled))
                    (proc))
                  $code
                  rte))))

(define ##subproblem-apply1
  (let ()
    (##declare (not inline))
    (lambda ($code rte proc arg1)
      (##declare (not interrupts-enabled))
      (##identity (let ()
                    (##declare (interrupts-enabled))
                    (proc arg1))
                  $code
                  rte))))

(define ##subproblem-apply2
  (let ()
    (##declare (not inline))
    (lambda ($code rte proc arg1 arg2)
      (##declare (not interrupts-enabled))
      (##identity (let ()
                    (##declare (interrupts-enabled))
                    (proc arg1 arg2))
                  $code
                  rte))))

(define ##subproblem-apply3
  (let ()
    (##declare (not inline))
    (lambda ($code rte proc arg1 arg2 arg3)
      (##declare (not interrupts-enabled))
      (##identity (let ()
                    (##declare (interrupts-enabled))
                    (proc arg1 arg2 arg3))
                  $code
                  rte))))

(define ##subproblem-apply
  (let ()
    (##declare (not inline))
    (lambda ($code rte proc args)
      (##declare (not interrupts-enabled))
      (##identity (let ()
                    (##declare (interrupts-enabled))
                    (##apply proc args))
                  $code
                  rte))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-delay
  (mk-cprc
    (delay (code-run (^ 0)))))

(define ##gen-delay
  (mk-gen (val)
    (mk-code ##cprc-delay cte (val))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-future
  (mk-cprc
    (future (code-run (^ 0)))))

(define ##gen-future
  (mk-gen (val)
    (mk-code ##cprc-future cte (val))))

;------------------------------------------------------------------------------

; Access to structure of closures for interpreter procedures.
;
; Layout of closed variables for closures created by ##cprc-prcXXX:
;
;   slot 1: $code
;   slot 2: proc
;   slot 3: rte

(define ##int-proc-code-pointers
  (##list (##closure-code (##cprc-prc0     #f #f))
          (##closure-code (##cprc-prc1     #f #f))
          (##closure-code (##cprc-prc2     #f #f))
          (##closure-code (##cprc-prc3     #f #f))
          (##closure-code (##cprc-prc      #f #f))
          (##closure-code (##cprc-prc-rest #f #f))))

(define (##int-proc? x)
  (and (##procedure? x)
       (##closure? x)
       (##memq (##closure-code x) ##int-proc-code-pointers)))

(define (##int-proc-code x) ; return "$code" closed variable of x
  (##closure-ref x 1))

(define (##int-proc-rte x) ; return "rte" closed variable of x
  (##closure-ref x 3))

;==============================================================================

; Eval

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Evaluation in a top environment with the current dynamic environment.

(define ##eval-top #f)
(set! ##eval-top
  (lambda (expr top-cte)
    (##eval-top-with-dyn-env expr top-cte (##dynamic-env-ref))))

; Evaluation in a top environment with a specific dynamic environment.

(define ##eval-top-with-dyn-env #f)
(set! ##eval-top-with-dyn-env
  (lambda (expr top-cte dyn-env)
    (let ((c (##compile-top top-cte expr)))
      (##dynamic-env-bind
        dyn-env
        (lambda () (let ((rte #f)) (code-run c)))))))

; Evaluation inside a specific environment ('cte' describes the runtime
; environment 'rte').

(define ##eval-inner #f)
(set! ##eval-inner
  (lambda (expr cte rte dyn-env)
    (let ((c (##compile-inner cte expr)))
      (##dynamic-env-bind
        dyn-env
        (lambda () (let ((rte rte)) (code-run c)))))))

;==============================================================================

; Decompilation of a piece of code

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(##define-macro (mk-degen params . def)
  `(lambda ($code ,@params) ,@def))

(##define-macro (degen proc . args)
  `(,proc $code ,@args))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##extract-proc $code rte)
  (let loop ((c (code-cte $code)) (r rte))
    (cond ((##cte-top? c)
           #f)
          ((##cte-frame? c)
           (let ((vars (##cte-frame-vars c)))
             (if (and (##pair? vars) (##eq? (##car vars) (self-var)))
               (rte-ref r 1)
               (loop (##cte-parent-cte c) (rte-up r)))))
          (else
           (loop (##cte-parent-cte c) r)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##BEGIN? x) (and (##pair? x) (##eq? (##car x) 'BEGIN)))
(define (##COND? x)  (and (##pair? x) (##eq? (##car x) 'COND)))
(define (##AND? x)   (and (##pair? x) (##eq? (##car x) 'AND)))
(define (##OR? x)    (and (##pair? x) (##eq? (##car x) 'OR)))
(define (##void-constant? x)
  (and (##pair? x)
       (##eq? (##car x) 'QUOTE)
       (##eq? (##cadr x) (##void))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##degen-top
  (mk-degen ()
    (##decomp (^ 0))))

(define ##degen-cst-x
  (mk-degen (val)
    (if (##self-eval? val) val (##list 'QUOTE val))))

(define ##degen-cst
  (mk-degen ()
    (degen ##degen-cst-x (^ 0))))

(define ##degen-loc-ref-x-y
  (mk-degen (up over)
    (degen ##degen-up-over up over)))

(define ##degen-up-over
  (mk-degen (up over)
    (let loop1 ((c (code-cte $code)) (up up))
      (cond ((##cte-frame? c)
             (if (##fixnum.= up 0)
               (let loop2 ((vars (##cte-frame-vars c)) (i over))
                 (if (##fixnum.< i 2)
                   (##car vars)
                   (loop2 (##cdr vars) (##fixnum.- i 1))))
               (loop1 (##cte-parent-cte c) (##fixnum.- up 1))))
            (else
             (loop1 (##cte-parent-cte c) up))))))

(define ##degen-loc-ref
  (mk-degen ()
    (degen ##degen-loc-ref-x-y (^ 0) (^ 1))))

(define ##degen-glo-ref
  (mk-degen ()
    (global-var->identifier (^ 0))))

(define ##degen-loc-set
  (mk-degen ()
    (##list 'SET! (degen ##degen-up-over (^ 1) (^ 2))
                  (##decomp (^ 0)))))

(define ##degen-glo-set
  (mk-degen ()
    (##list 'SET! (global-var->identifier (^ 1))
                  (##decomp (^ 0)))))

(define ##degen-glo-def
  (mk-degen ()
    (##list 'DEFINE (global-var->identifier (^ 1))
                    (##decomp (^ 0)))))

(define ##degen-if2
  (mk-degen ()
    (##list 'IF (##decomp (^ 0))
                (##decomp (^ 1)))))

(define ##degen-if3
  (mk-degen ()
    (##list 'IF (##decomp (^ 0))
                (##decomp (^ 1))
                (##decomp (^ 2)))))

(define ##degen-seq
  (mk-degen ()
    (let ((val1 (##decomp (^ 0)))
          (val2 (##decomp (^ 1))))
      (if (##BEGIN? val2)
        (##cons 'BEGIN (##cons val1 (##cdr val2)))
        (##list 'BEGIN val1 val2)))))

(define ##degen-quasi-list->vector
  (mk-degen ()
    (##list 'QUASIQUOTE (##vector (##list 'UNQUOTE-SPLICING (##decomp (^ 0)))))))

(define ##degen-quasi-append
  (mk-degen ()
    (##list 'QUASIQUOTE (##list (##list 'UNQUOTE-SPLICING (##decomp (^ 0)))
                                (##list 'UNQUOTE-SPLICING (##decomp (^ 1)))))))

(define ##degen-quasi-cons
  (mk-degen ()
    (##list 'QUASIQUOTE (##list (##list 'UNQUOTE (##decomp (^ 0)))
                                (##list 'UNQUOTE-SPLICING (##decomp (^ 1)))))))

(define ##degen-cond-if
  (mk-degen ()
    (let ((val1 (##decomp (^ 0)))
          (val2 (##decomp (^ 1)))
          (val3 (##decomp (^ 2))))
      (##build-cond
        (if (##BEGIN? val2) (##cons val1 (##cdr val2)) (##list val1 val2))
        val3))))

(define ##degen-cond-or
  (mk-degen ()
    (let ((val1 (##decomp (^ 0)))
          (val2 (##decomp (^ 1))))
      (##build-cond (##list val1) val2))))

(define ##degen-cond-send
  (mk-degen ()
    (let ((val1 (##decomp (^ 0)))
          (val2 (##decomp (^ 1)))
          (val3 (##decomp (^ 2))))
      (##build-cond (##list val1 '=> val2) val3))))

(define (##build-cond clause rest)
  (cond ((##COND? rest)
         (##cons 'COND (##cons clause (##cdr rest))))
        ((##BEGIN? rest)
         (##cons 'COND (##list clause (##cons 'ELSE (##cdr rest)))))
        ((##void-constant? rest)
         (##list 'COND clause))
        (else
         (##list 'COND clause (##list 'ELSE rest)))))

(define ##degen-or
  (mk-degen ()
    (let ((val1 (##decomp (^ 0)))
          (val2 (##decomp (^ 1))))
      (if (##OR? val2)
        (##cons 'OR (##cons val1 (##cdr val2)))
        (##list 'OR val1 val2)))))

(define ##degen-and
  (mk-degen ()
    (let ((val1 (##decomp (^ 0)))
          (val2 (##decomp (^ 1))))
      (if (##AND? val2)
        (##cons 'AND (##cons val1 (##cdr val2)))
        (##list 'AND val1 val2)))))

(define ##degen-case
  (mk-degen ()
    (let ((val1 (##decomp (^ 0)))
          (val2 (##decomp (^ 1))))
      (##cons 'CASE (##cons val1 val2)))))

(define ##degen-case-clause
  (mk-degen ()
    (let ((val1 (##decomp (^ 0)))
          (val2 (##decomp (^ 1))))
      (##cons (if (##BEGIN? val1)
                (##cons (^ 2) (##cdr val1))
                (##list (^ 2) val1))
              val2))))

(define ##degen-case-else
  (mk-degen ()
    (let ((val (##decomp (^ 0))))
      (if (##void-constant? val)
        '()
        (##list (if (##BEGIN? val)
                  (##cons 'ELSE (##cdr val))
                  (##list 'ELSE val)))))))

(define ##degen-let
  (mk-degen ()
    (let ((n (code-length $code)))
      (let loop ((i (##fixnum.- n 2)) (vals '()))
        (if (##fixnum.< 0 i)
          (loop (##fixnum.- i 1)
                (##cons (##decomp (code-ref $code i)) vals))
          (let ((body (##decomp (^ 0)))
                (bindings (##make-bindings (code-ref $code (##fixnum.- n 1)) vals)))
            (if (##BEGIN? body)
              (##cons 'LET (##cons bindings (##cdr body)))
              (##list 'LET bindings body))))))))

(define (##make-bindings l1 l2)
  (if (##pair? l1)
    (##cons (##list (##car l1) (##car l2))
            (##make-bindings (##cdr l1) (##cdr l2)))
    '()))

(define ##degen-letrec
  (mk-degen ()
    (let ((n (code-length $code)))
      (let loop ((i (##fixnum.- n 2)) (vals '()))
        (if (##fixnum.< 0 i)
          (loop (##fixnum.- i 1)
                (##cons (##decomp (code-ref $code i)) vals))
          (let ((body (##decomp (^ 0)))
                (bindings (##make-bindings (code-ref $code (##fixnum.- n 1)) vals)))
            (if (##BEGIN? body)
              (##cons 'LETREC (##cons bindings (##cdr body)))
              (##list 'LETREC bindings body))))))))

(define ##degen-prc
  (mk-degen ()
    (let ((body (##decomp (^ 0)))
          (params (code-ref $code (##fixnum.- (code-length $code) 1))))
      (if (##BEGIN? body)
        (##cons 'LAMBDA (##cons params (##cdr body)))
        (##list 'LAMBDA params body)))))

(define ##degen-prc-rest
  (mk-degen ()
    (let ((body (##decomp (^ 0)))
          (params (##make-rest-params (^ 2))))
      (if (##BEGIN? body)
        (##cons 'LAMBDA (##cons params (##cdr body)))
        (##list 'LAMBDA params body)))))

(define (##make-rest-params l)
  (if (##null? (##cdr l))
    (##car l)
    (##cons (##car l) (##make-rest-params (##cdr l)))))

(define ##degen-app0
  (mk-degen ()
    (##list (##decomp (^ 0)))))

(define ##degen-app1
  (mk-degen ()
    (##list (##decomp (^ 0))
            (##decomp (^ 1)))))

(define ##degen-app2
  (mk-degen ()
    (##list (##decomp (^ 0))
            (##decomp (^ 1))
            (##decomp (^ 2)))))

(define ##degen-app3
  (mk-degen ()
    (##list (##decomp (^ 0))
            (##decomp (^ 1))
            (##decomp (^ 2))
            (##decomp (^ 3)))))

(define ##degen-app
  (mk-degen ()
    (let ((n (code-length $code)))
      (let loop ((i (##fixnum.- n 1)) (vals '()))
        (if (##not (##fixnum.< i 0))
          (loop (##fixnum.- i 1)
                (##cons (##decomp (code-ref $code i)) vals))
          vals)))))

(define ##degen-delay
  (mk-degen ()
    (##list 'DELAY (##decomp (^ 0)))))

(define ##degen-future
  (mk-degen ()
    (##list 'FUTURE (##decomp (^ 0)))))

;------------------------------------------------------------------------------

(define ##decomp-dispatch-table
  (##list
    (##cons ##cprc-top           ##degen-top)

    (##cons ##cprc-cst-null      (mk-degen () (degen ##degen-cst-x '())))
    (##cons ##cprc-cst-true      (mk-degen () (degen ##degen-cst-x #t)))
    (##cons ##cprc-cst-false     (mk-degen () (degen ##degen-cst-x #f)))
    (##cons ##cprc-cst--2        (mk-degen () (degen ##degen-cst-x -2)))
    (##cons ##cprc-cst--1        (mk-degen () (degen ##degen-cst-x -1)))
    (##cons ##cprc-cst-0         (mk-degen () (degen ##degen-cst-x 0)))
    (##cons ##cprc-cst-1         (mk-degen () (degen ##degen-cst-x 1)))
    (##cons ##cprc-cst-2         (mk-degen () (degen ##degen-cst-x 2)))
    (##cons ##cprc-cst           ##degen-cst)

    (##cons ##cprc-loc-ref-0-1   (mk-degen () (degen ##degen-loc-ref-x-y 0 1)))
    (##cons ##cprc-loc-ref-0-2   (mk-degen () (degen ##degen-loc-ref-x-y 0 2)))
    (##cons ##cprc-loc-ref-0-3   (mk-degen () (degen ##degen-loc-ref-x-y 0 3)))
    (##cons ##cprc-loc-ref-1-1   (mk-degen () (degen ##degen-loc-ref-x-y 1 1)))
    (##cons ##cprc-loc-ref-1-2   (mk-degen () (degen ##degen-loc-ref-x-y 1 2)))
    (##cons ##cprc-loc-ref-1-3   (mk-degen () (degen ##degen-loc-ref-x-y 1 3)))
    (##cons ##cprc-loc-ref-2-1   (mk-degen () (degen ##degen-loc-ref-x-y 2 1)))
    (##cons ##cprc-loc-ref-2-2   (mk-degen () (degen ##degen-loc-ref-x-y 2 2)))
    (##cons ##cprc-loc-ref-2-3   (mk-degen () (degen ##degen-loc-ref-x-y 2 3)))
    (##cons ##cprc-loc-ref       ##degen-loc-ref)
    (##cons ##cprc-glo-ref       ##degen-glo-ref)

    (##cons ##cprc-loc-set       ##degen-loc-set)
    (##cons ##cprc-glo-set       ##degen-glo-set)
    (##cons ##cprc-glo-def       ##degen-glo-def)

    (##cons ##cprc-if2           ##degen-if2)
    (##cons ##cprc-if3           ##degen-if3)
    (##cons ##cprc-seq           ##degen-seq)
    (##cons ##cprc-quasi-list->vector ##degen-quasi-list->vector)
    (##cons ##cprc-quasi-append  ##degen-quasi-append)
    (##cons ##cprc-quasi-cons    ##degen-quasi-cons)
    (##cons ##cprc-cond-if       ##degen-cond-if)
    (##cons ##cprc-cond-or       ##degen-cond-or)
    (##cons ##cprc-cond-send-red ##degen-cond-send)
    (##cons ##cprc-cond-send-sub ##degen-cond-send)

    (##cons ##cprc-or            ##degen-or)
    (##cons ##cprc-and           ##degen-and)

    (##cons ##cprc-case          ##degen-case)
    (##cons ##cprc-case-clause   ##degen-case-clause)
    (##cons ##cprc-case-else     ##degen-case-else)

    (##cons ##cprc-let           ##degen-let)
    (##cons ##cprc-letrec        ##degen-letrec)

    (##cons ##cprc-prc0          ##degen-prc)
    (##cons ##cprc-prc1          ##degen-prc)
    (##cons ##cprc-prc2          ##degen-prc)
    (##cons ##cprc-prc3          ##degen-prc)
    (##cons ##cprc-prc           ##degen-prc)
    (##cons ##cprc-prc-rest      ##degen-prc-rest)

    (##cons ##cprc-app0-red      ##degen-app0)
    (##cons ##cprc-app1-red      ##degen-app1)
    (##cons ##cprc-app2-red      ##degen-app2)
    (##cons ##cprc-app3-red      ##degen-app3)
    (##cons ##cprc-app-red       ##degen-app)
    (##cons ##cprc-app0-sub      ##degen-app0)
    (##cons ##cprc-app1-sub      ##degen-app1)
    (##cons ##cprc-app2-sub      ##degen-app2)
    (##cons ##cprc-app3-sub      ##degen-app3)
    (##cons ##cprc-app-sub       ##degen-app)

    (##cons ##cprc-delay         ##degen-delay)
    (##cons ##cprc-future        ##degen-future)
))

;------------------------------------------------------------------------------

(define (##decomp $code)
  (let ((cprc (code-cprc $code)))
    (let ((x (##assq cprc ##decomp-dispatch-table)))
      (if x
        (degen (##cdr x))
        '?))))

(define (##decompile proc)

  (define (decomp proc)
    (let ((parent (##subprocedure-parent proc))
          (id (##subprocedure-id proc)))
      (let ((info (##procedure-info parent)))
        (if info
          (let ((v (##vector-ref info 0)))
            (let loop ((i (##fixnum.- (##vector-length v) 1)))
              (if (##fixnum.< i 0)
                proc
                (let ((x (##vector-ref v i)))
                  (if (##fixnum.= id (##vector-ref x 0))
                    (source->expression (##vector-ref x 1))
                    (loop (##fixnum.- i 1)))))))
          proc))))

  (define (source-code x)
    (##vector-ref x 0))

  (define (source->expression source)

    (define (list->expression l)
      (cond ((##pair? l)
             (##cons (source->expression (##car l))
                     (list->expression (##cdr l))))
            ((##null? l)
             '())
            (else
             (source->expression l))))

    (define (vector->expression v)
      (let* ((len (##vector-length v))
             (x (##make-vector len #f)))
        (let loop ((i (##fixnum.- len 1)))
          (if (##not (##fixnum.< i 0))
            (begin
              (##vector-set! x i (source->expression (##vector-ref v i)))
              (loop (##fixnum.- i 1)))))
        x))

    (let ((code (source-code source)))
      (cond ((##pair? code)   (list->expression code))
            ((##vector? code) (vector->expression code))
            (else             code))))

  (cond ((##int-proc? proc)
         (##decomp (##int-proc-code proc)))
        ((##closure? proc)
         (decomp (##closure-code proc)))
        (else
         (decomp proc))))

;==============================================================================

; Utilities

;------------------------------------------------------------------------------

; Access to interpreter continuation frames

(define (##int-frame-non-subproblem? f)
  (let ((parent (##subprocedure-parent (##frame-ret f))))
    (##assq parent ##decomp-dispatch-table)))

(define (##int-frame-subproblem? f)
  (let ((parent (##subprocedure-parent (##frame-ret f))))
    (or (##eq? parent ##subproblem-apply0)
        (##eq? parent ##subproblem-apply1)
        (##eq? parent ##subproblem-apply2)
        (##eq? parent ##subproblem-apply3)
        (##eq? parent ##subproblem-apply))))

(define (##int-frame-subproblem-code f)
  (let ((parent (##subprocedure-parent (##frame-ret f))))
    (if (##eq? parent ##subproblem-apply0)
      (##frame-stk-ref f 2)
      (##frame-stk-ref f 1))))

(define (##int-frame-subproblem-rte f)
  (let ((parent (##subprocedure-parent (##frame-ret f))))
    (if (or (##eq? parent ##subproblem-apply2)
            (##eq? parent ##subproblem-apply3))
      (##frame-stk-ref f 2)
      (##frame-stk-ref f 3))))

;------------------------------------------------------------------------------

(define (##uninteresting-frame? f)
  (let ((parent (##subprocedure-parent (##frame-ret f))))
    (or (##assq parent ##decomp-dispatch-table)
        (##eq? parent ##dynamic-env-bind)
        (##eq? parent ##read-eval-print))))

(define (##continuation->subproblems cont)
  (let loop ((f (##continuation->frame cont)) (l '()))
    (if f
      (if (##uninteresting-frame? f)
        (loop (##frame-next f) l)
        (loop (##frame-next f) (##cons f l)))
      (##reverse l))))

(define (##eval-in-frame expr f dyn-bindings)
  (let ((dyn-env (##cons dyn-bindings (##frame-dyn-env f))))
    (if (##int-frame-subproblem? f)
      (let* ((code (##int-frame-subproblem-code f))
             (cte (code-cte code))
             (rte (##int-frame-subproblem-rte f)))
        (##eval-inner expr cte rte dyn-env))
      (##eval-top-with-dyn-env expr ##interaction-cte dyn-env))))

(define (##procedure-name p)
  (or (##object->global-var->identifier p) p))

;==============================================================================

; Read eval print loop

(define ##repl-write #f)
(set! ##repl-write #f)

(define ##repl-read #f)
(set! ##repl-read #f)

(define (##repl
          #!optional (in ##stdin) (out ##stdout) (prompt2 "> ") (prompt1 ""))
  (##call-with-current-continuation
    (lambda (cont) (##read-eval-print in out prompt2 prompt1 cont))))

(define (##read-eval-print in out prompt2 prompt1 cont)

  (define (repl-start subprobs repl-info dyn-bindings)

    (define (repl-read)
      (let ((proc ##repl-read))
        (if (##procedure? proc)
          (proc in)
          (##read in))))

    (define (repl-write val)
      (let ((proc ##repl-write))
        (if (##procedure? proc)
          (proc val out)
          (if (##not (##eq? val (##void)))
            (begin
              (##write val out (if-forces #t #f))
              (##newline out))))))

    (define (repl-n n)
      (let loop ((i 0) (s subprobs))
        (if (and (##fixnum.< n i) (##pair? (##cdr s)))
          (loop (##fixnum.- i 1) (##cdr s))
          (let ((f (##car s)))
            (##display-subproblem i f out)
            (repl i s f)))))

    (define (cmd-d eof?)
      (let ((l (##cdr (##vector-ref repl-info 3))))
        (if eof?
          (##newline out))
        (cond ((##pair? l)
               ((##car l) #f))
              (eof?
               (##write-string "*** EOF again to exit" out)
               (##newline out)
               (if (##eof-object? (##peek-char in))
                 (##exit))))))

    (define (cmd-t)
      (let loop ((l (##vector-ref repl-info 3)))
        (if (##pair? (##cdr l))
          (loop (##cdr l))
          ((##car l) #f))))

    (define (repl pos subprobs* f)

      (##call-with-current-continuation
        (lambda (abort)
          (##set-car! (##vector-ref repl-info 3) abort)))

      (let loop ()

        (##display prompt1 out #f)
        (if (##fixnum.< pos 0) (##display pos out #f))
        (##display prompt2 out #f)

        (let ((expr (repl-read)))
          (if (##eof-object? expr)
            (begin (cmd-d #t) (loop))
            (if (and (##pair? expr)
                     (##pair? (##cdr expr))
                     (##null? (##cddr expr))
                     (##eq? (##car expr) 'UNQUOTE))
              (let ((cmd (##cadr expr)))
                (if (##eof-object? cmd)
                  (begin (cmd-d #t) (loop))
                  (case cmd
                    ((?) (##cmd-? out) (loop))
                    ((-) (repl-n (##fixnum.- pos 1)))
                    ((+) (repl-n (##fixnum.+ pos 1)))
                    ((b) (##cmd-b pos subprobs* out) (loop))
                    ((i) (##cmd-i f out) (loop))
                    ((y) (##cmd-y f out) (loop))
                    ((l) (##cmd-l f out) (loop))
                    ((t) (cmd-t))
                    ((d) (cmd-d #f) (loop))
                    ((r) (##display "Return value: " out #f)
                         (let ((expr (repl-read)))
                           (if (##eof-object? expr)
                             (##void)
                             (##eval-in-frame expr f dyn-bindings))))
                    ((q) (##exit))
                    (else
                     (if (and (##fixnum? cmd) (##fixnum.< cmd 1))
                       (repl-n cmd)
                       (begin
                         (##write-string "Unknown command ," out)
                         (##write cmd out #f)
                         (##newline out)
                         (loop)))))))
            (let ((val (##eval-in-frame expr f dyn-bindings)))
              (repl-write val)
              (loop)))))))

    (repl 0 subprobs (##car subprobs)))

  (let* ((prev-info (##dynamic-ref '##REPL-INFO #f))
         (repl-info
           (##vector
             in
             out
             (if prev-info
               (##fixnum.+ (##vector-ref prev-info 2) 1)
               0)
             (if prev-info
               (let ((conts (##vector-ref prev-info 3)))
                 (##cons (##car conts) conts))
               (##cons (lambda (x) (##exit 1)) '()))))
         (dyn-bindings (##list (##cons '##REPL-INFO repl-info))))
    (##dynamic-let
      dyn-bindings
      (lambda ()
        (repl-start (##continuation->subproblems cont)
                    repl-info
                    dyn-bindings)))))

(define (##repl-out)
  (let ((repl-info (##dynamic-ref '##REPL-INFO #f)))
    (if repl-info
      (##vector-ref repl-info 1)
      ##stderr)))

(define (##debug-repl cont)
  (let ((repl-info (##dynamic-ref '##REPL-INFO #f)))
    (if repl-info
      (##read-eval-print (##vector-ref repl-info 0)
                         (##vector-ref repl-info 1)
                         "> "
                         (##fixnum.+ (##vector-ref repl-info 2) 1)
                         cont)
      (##exit 1))))

(define (##pop-repl)
  (let ((repl-info (##dynamic-ref '##REPL-INFO #f)))
    (if repl-info
      ((##car (##vector-ref repl-info 3)) #f)
      (##exit 1))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##cmd-? out)
  (##write-string ",?        : Summary of commands" out) (##newline out)
  (##write-string ",+ and ,- : Move to next or previous frame of continuation" out) (##newline out)
  (##write-string ",<n>      : Move to particular frame (<n> <= 0)" out) (##newline out)
  (##write-string ",b        : Display frames of continuation (i.e. backtrace)" out) (##newline out)
  (##write-string ",i        : Display procedure attached to current frame" out) (##newline out)
  (##write-string ",y        : Display subproblem of current frame" out) (##newline out)
  (##write-string ",l        : Display list of local variables accessible in current frame" out) (##newline out)
  (##write-string ",t        : Transfer to top-level REP loop" out) (##newline out)
  (##write-string ",d        : Transfer to previous REP loop" out) (##newline out)
  (##write-string ",r        : Return from REP loop" out) (##newline out)
  (##write-string ",q        : Quit" out) (##newline out))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##cmd-b pos subprobs* out)
  (define max-head 10)
  (define max-tail 6)
  (let loop ((i 0) (j (##fixnum.- (##length subprobs*) 1)) (l subprobs*))
    (if (##pair? l)
      (begin
        (cond ((or (##fixnum.< i max-head) (##fixnum.< j max-tail)
                   (and (##fixnum.= i max-head) (##fixnum.= j max-tail)))
               (##display-subproblem (##fixnum.- pos i) (##car l) out))
              ((##fixnum.= i max-head)
               (##write-string "..." out) (##newline out)))
        (loop (##fixnum.+ i 1) (##fixnum.- j 1) (##cdr l))))))

(define (##display-subproblem pos f out)
  (let ((x (##write pos out #f)))
    (##display-spaces (##fixnum.- 4 x) out)
    (##write-string " " out)

    (if (##int-frame-subproblem? f)

      (let ((code (##int-frame-subproblem-code f))
            (rte (##int-frame-subproblem-rte f)))
        (let ((proc (##extract-proc code rte)))
          (let ((x (if proc
                     (##write (##procedure-name proc) out #f)
                     (##display "(interaction)" out #f))))
            (##display-spaces (##fixnum.- 25 x) out)
            (##write-string " " out)
            (##write-string (##object->string (##decomp code) 48 #f) out)
            (##newline out))))

      (let ((parent (##subprocedure-parent (##frame-ret f))))
        (let ((x (##write (##procedure-name parent) out #f)))
          (let ((y (##decompile (##frame-ret f))))
            (if (##not (##eq? y (##frame-ret f)))
              (begin
                (##display-spaces (##fixnum.- 25 x) out)
                (##write-string " " out)
                (##write-string (##object->string y 48 #f) out)))
            (##newline out)))))))

(define (##display-spaces n out)
  (if (##fixnum.< 0 n)
    (let ((m (if (##fixnum.< 40 n) 40 n)))
      (##write-substring "                                        " 0 m out)
      (##display-spaces (##fixnum.- n m) out))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##cmd-l f out)

  (define (display-locals cte rte)
    (let loop1 ((c cte) (r rte))
      (cond ((##cte-top? c))
            ((##cte-frame? c)
             (let loop2 ((vars (##cte-frame-vars c))
                         (vals (##cdr (##vector->list r))))
               (if (##pair? vars)
                 (let ((var (##car vars)))
                   (if (##not (##internal-var? var))
                     (let ((x (##write var out #f)))
                       (##write-string " = " out)
                       (##write-string
                        (##object->string
                          (##car vals)
                          (##fixnum.- (##fixnum.- (##port-width out) 3) x)
                          (if-forces #t #f))
                        out)
                       (##newline out)))
                   (loop2 (##cdr vars) (##cdr vals)))))
             (loop1 (##cte-parent-cte c) (rte-up r)))
            (else
             (loop1 (##cte-parent-cte c) r)))))

  (if (##int-frame-subproblem? f)
    (display-locals (code-cte (##int-frame-subproblem-code f))
                    (##int-frame-subproblem-rte f))
    (begin
      (##write-string "Sorry, can't display compiled code environment" out)
      (##newline out))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##cmd-y f out)
  (if (##int-frame-subproblem? f)
    (##pretty-print (##decomp (##int-frame-subproblem-code f)) out (##port-width out))
    (let ((x (##decompile (##frame-ret f))))
      (if (##eq? x (##frame-ret f))
        (begin
          (##write-string "Sorry, this code was compiled without the DEBUG option" out)
          (##newline out))
        (##pretty-print x out (##port-width out))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##cmd-i f out)
  (if (##int-frame-subproblem? f)

    (let ((code (##int-frame-subproblem-code f))
          (rte (##int-frame-subproblem-rte f)))
      (let ((proc (##extract-proc code rte)))
        (if proc
          (begin
            (##write proc out #f)
            (##write-string " =" out)
            (##newline out)
            (##pretty-print (##decompile proc) out (##port-width out)))
          (begin
            (##write-string "(interaction)" out)
            (##newline out)))))

    (let ((proc (##subprocedure-parent (##frame-ret f))))
      (##write proc out #f)
      (let ((x (##decompile proc)))
        (if (##eq? x proc)
          (##newline out)
          (begin
            (##write-string " =" out)
            (##newline out)
            (##pretty-print x out (##port-width out))))))))

;==============================================================================
