;==============================================================================

; file: "_kernel.scm"

; Copyright (C) 1994-1996 by Marc Feeley, All Rights Reserved.

(##include "header.scm")

(##declare
  (multilisp)
  (standard-bindings)
  (extended-bindings)
  (block)
  (fixnum)
)

(##declare (not interrupts-enabled))

(c-declare "

/*---------------------------------------------------------------------------*/

#include \"os.h\"
#include \"setup.h\"
#include \"mem.h\"
#include \"c_intf.h\"

___NEED_GLO(___G__23__23_exception_2e_stack_2d_overflow)
___NEED_GLO(___G__23__23_exception_2e_heap_2d_overflow)
___NEED_GLO(___G__23__23_interrupt_2d_handler)
___NEED_GLO(___G__23__23_exception_2e_wrong_2d_nb_2d_arg)
___NEED_GLO(___G__23__23_exception_2e_c_2d_conv_2d_error)
___NEED_GLO(___G__23__23_exception_2e_non_2d_proc_2d_jump)
___NEED_GLO(___G__23__23_exception_2e_global_2d_jump)

___WORD ___internal_return;

/*---------------------------------------------------------------------------*/

#ifndef ___CC_PROG1
#define ___CC_PROG1 0
#endif

#ifndef ___CC_PROG2
#define ___CC_PROG2 0
#endif

#ifndef ___CC_PROG3
#define ___CC_PROG3 0
#endif

#ifndef ___LD_PROG1
#define ___LD_PROG1 0
#endif

#ifndef ___LD_PROG2
#define ___LD_PROG2 0
#endif

#ifndef ___LD_PROG3
#define ___LD_PROG3 0
#endif

")

(define ##setup-traps (let () (##declare (not inline)) (lambda ()

(##c-code
"
___ps->handler_break           = ___LBL(1);
___ps->handler_stack_limit     = ___LBL(2);
___ps->handler_heap_limit      = ___LBL(3);
___ps->handler_not_proc        = ___LBL(4);
___ps->handler_not_proc_glo    = ___LBL(5);
___ps->handler_wrong_nargs     = ___LBL(6);
___ps->handler_get_rest        = ___LBL(7);
___ps->handler_get_key         = ___LBL(8);
___ps->handler_get_key_rest    = ___LBL(9);
___ps->handler_force           = ___LBL(10);
___ps->handler_c_conv_error    = ___LBL(11);
___internal_return             = ___LBL(12);
___ps->intr_enabled = 1; /* enable interrupts */
___fp[0] = ___ps->handler_break;
___JUMPEXTPRM(___NOTHING,___R0)
")

(##c-code
"
/*
 * ___LBL(1)
 *
 * This is the break handler.  It is invoked when a function
 * attempts to return to its caller and the caller's stack frame
 * is not on top of the stack because it has been captured.
 *
 * At this point the callee will have cleaned up the stack so
 * that the frame pointer (___fp) points to the break frame.
 * The break frame contains the address in the caller where
 * control will return (the return address) and a pointer to
 * the caller's continuation frame, which can either be in the
 * stack or in the heap.  The two situations are depicted below:
 *
 *              STACK                        STACK              HEAP
 *          |            |               |            |                caller's
 *          +------------+               +------------+                frame
 * ___fp -->|  ret adr   |      ___fp -->|  ret adr   |    +------------+
 *          | call frame ---+            | call frame ---->|    HEAD    |
 *          +------------+  |            +------------+    | next frame ---+
 *          |     .      |  |            |     .      |    |  slot fs   |  |
 *          |     .      |  |            |     .      |    |    ...     |  |
 *          +------------+  |            |            |    |  slot 1    |  |
 * caller's |  slot fs   |<-+            |            |    +------------+  |
 * frame    |    ...     |               |            |                    |
 *          |  slot 1    |               |            |    +------------+  |
 *          +------------+               |            |    |    HEAD    |<-+
 *          |     .      |               |     .      |    | next frame ---+
 *          |     .      |               |     .      |    |    ...     |  |
 *          +------------+               +------------+    +------------+  V
 *                                                                        ...
 *
 * These cases are distinguished by the tag on the pointer to the
 * caller's frame.
 *
 * The break handler puts a copy of the caller's frame on the
 * top of the stack, saves the return address it contains into
 * the break frame and replaces it with the address of the
 * break handler.  The frame pointer in the break frame is made
 * to point to the frame of the caller's caller.  Finally a jump
 * to the return address that was originally in the break frame
 * is performed.  At that point the stack will be in the following
 * state respectively:
 *
 *              STACK                        STACK              HEAP
 *          |            |               |            |
 *          +------------+               +------------+
 * ___fp -->|  slot fs   |      ___fp -->|  slot fs   |
 *          |    ...     |               |    ...     |
 *          |  slot 1    |               |  slot 1    |
 *          +------------+               +------------+
 *          |new ret adr |               |new ret adr |    +------------+
 *          |new call fr.---+            |new call fr.--+  |    HEAD    |
 *          +------------+  |            +------------+ |  | next frame ---+
 *          |     .      |  |            |     .      | |  |  slot fs   |  |
 *          |     .      |  |            |     .      | |  |    ...     |  |
 *          +------------+  |            |            | |  |  slot 1    |  |
 *          |  slot fs   |  |            |            | |  +------------+  |
 *          |    ...     |  |            |            | |                  |
 *          |  slot 1    |  |            |            | |  +------------+  |
 *          +------------+  |            |            | +->|    HEAD    |<-+
 *          |     .      |<-+            |     .      |    | next frame ---+
 *          |     .      | (see note)    |     .      |    |    ...     |  |
 *          +------------+               +------------+    +------------+  V
 *                                                                        ...
 *
 * Note: In the first case, the pointer to the caller's frame
 * is simply advanced to the frame following the caller's frame.
 * However, if the frame following the caller's frame is a break
 * frame, then the content of that break frame is copied to the
 * topmost break frame.  This ensures that break frames never
 * contain pointers to other break frames.
 */

{
  ___WORD ora = ___fp[0]; /* original return address            */
  ___WORD ocf = ___fp[1]; /* original pointer to caller's frame */

  if (___TYP(ocf) == ___tSUBTYPED)
    {
      /* caller's frame is in the heap */

      int fs, link, i;

      ___WORD *fp = ___BODY_AS(ocf,___tSUBTYPED); /* get pointer to frame */

      if (ora == ___internal_return)
        {
          ___WORD fs_link = ((___lbl_struct*)(fp[1]-___tSUBTYPED))->flags;
          fs = ___RETI_INFO_FS(fs_link);
          link = ___RETI_INFO_LINK(fs_link);
        }
      else
        {
          ___WORD fs_link = ((___lbl_struct*)(ora-___tSUBTYPED))->flags;
          fs = ___RETN_INFO_FS(fs_link);
          link = ___RETN_INFO_LINK(fs_link);
        }
      fp += fs+1;

      ___fp[0] = fp[-link-1];
      ___fp[1] = fp[-fs-1];

      for (i=-fs; i<0; i++)
        ___fp[i] = fp[i];

      ___fp[-link-1] = ___ps->handler_break;
      ___fp -= fs;

      ___JUMPEXTPRM(___NOTHING,ora)
    }
  else if (ocf == 0)
    {
      ___JUMPEXTPRM(___NOTHING,___BRK)
    }
  else
    {
      /* continuation's frame is in the stack */

      ___WORD *fp, frame_ra;
      int fs, link, i;

      if (ora == ___internal_return)
        {
          ___WORD fs_link = ((___lbl_struct*)(((___WORD*)ocf)[0]-___tSUBTYPED))->flags;
          fs = ___RETI_INFO_FS(fs_link);
          link = ___RETI_INFO_LINK(fs_link);
        }
      else
        {
          ___WORD fs_link = ((___lbl_struct*)(ora-___tSUBTYPED))->flags;
          fs = ___RETN_INFO_FS(fs_link);
          link = ___RETN_INFO_LINK(fs_link);
        }
      fp = ((___WORD*)ocf)+fs;
      frame_ra = fp[-link-1];

      if (frame_ra == ___ps->handler_break)
        {
          /* first frame of that section */

          ___fp[0] = fp[0];
          ___fp[1] = fp[1];

          for (i=-fs; i<0; i++)
            ___fp[i] = fp[i];

          ___fp -= fs;

          ___JUMPEXTPRM(___NOTHING,ora)
        }
      else
        {
          /* not the first frame of that section */

          ___fp[0] = frame_ra;
          ___fp[1] = (___WORD)fp;

          for (i=-fs; i<0; i++)
            ___fp[i] = fp[i];

          ___fp[-link-1] = ___ps->handler_break;
          ___fp -= fs;

          ___JUMPEXTPRM(___NOTHING,ora)
        }
    }
}" (0))

(##c-code
"
/*
 * ___LBL(2)
 *
 * This is the stack-limit handler.  It is invoked when the stack
 * overflows and when an interrupt is received.
 *
 * This handler checks for which reason it was invoked and dispatches
 * to one of the following procedures:
 *
 *  ##interrupt-handler
 *  ##exception.stack-overflow
 */

{
  ___WORD *stack_trip;

  ___PUSH_REGS                  /* setup internal-return stack frame */
  ___PUSH(___ps->temp1)         /* return address in caller */
  ___SET_R0(___internal_return) /* change return address */

  stack_trip = ___ps->stack_trip; /* get stack trip */

  if (___fp < ___ps->stack_limit) /* stack overflow? */
    {
      int overflow;

      ___fp[-1] = (___WORD)___fp;
      ___fp[-2] = ___R0;
      ___fp -= 2;
      ___ps->stack_break = ___fp;
      ___R0 = ___ps->handler_break;

      ___W_ALL
      overflow = ___gc ();
      ___R_ALL

      if (overflow)
        ___JUMPEXT(___SET_NARGS(0),___G__23__23_exception_2e_stack_2d_overflow.val)
      else
        ___JUMPEXTPRM(___NOTHING,___R0)
    }

  ___ps->stack_trip = ___ps->stack_limit; /* prepare for next intr */

  if (___ps->intr_enabled)
    {
      int i;

      for (i=0; i<___NB_INTRS; i++)
        if (___ps->intr_flag[i])
          break;

      if (i < ___NB_INTRS)
        {
          if (i == ___INTR_TIMER)
            ___periodic_task ();
          ___SET_R1(___FIX(i))
          ___ps->intr_flag[i] = 0; /* lower flag */
          for (i=i+1; i<___NB_INTRS; i++)
            if (___ps->intr_flag[i]) /* make sure other intr is not ignored */
              {
                ___ps->stack_trip = ___ps->stack_base;
                break;
              }
          ___JUMPEXT(___SET_NARGS(1),___G__23__23_interrupt_2d_handler.val)
        }
    }

  ___JUMPEXTPRM(___NOTHING,___ps->temp1)
}" (0))

(##c-code
"
/*
 * ___LBL(3)
 *
 * This is the heap-limit handler.  It is invoked when the heap
 * pointer reaches the end of the current msection.
 *
 * This handler simply calls ##check-heap.
 */

{
  ___PUSH_REGS                  /* setup internal-return stack frame */
  ___PUSH(___ps->temp1)         /* return address in caller */
  ___SET_R0(___internal_return) /* change return address */
  ___JUMPEXT(___SET_NARGS(0),___G__23__23_check_2d_heap.val)
}" (0))

(##c-code
"
/*
 * ___LBL(4)
 *
 * This is the call to non-procedure handler.  It is invoked when
 * there is an attempt to call an object that is not a procedure.
 *
 * This handler simply tail calls ##exception.non-proc-jump (i.e.
 * the continuation will be the same as that of the faulty call).  The
 * arguments are the object that was in operator position followed by
 * the arguments of the faulty call.
 */

{
  int na = ___ps->na;
  int i;

  ___PUSH_ARGS_IN_REGS(na)  /* save all arguments that are in registers */
  ___PUSH(0)                /* make space for operator */

  for (i=0; i<na; i++)
    ___fp[i] = ___fp[i+1];  /* shift arguments up by one */
  ___fp[na] = ___ps->temp1; /* set operator argument */

  ___POP_ARGS_IN_REGS(na+1) /* prepare call to error handler */

  ___JUMPEXT(___SET_NARGS(na+1),___G__23__23_exception_2e_non_2d_proc_2d_jump.val)
}
" (0))

(##c-code
"
/*
 * ___LBL(5)
 *
 * This is the call to global non-procedure handler.  It is invoked
 * when there is an attempt to call an object that is not a procedure
 * that is bound to a global variable.
 *
 * This handler simply tail calls ##exception.global-jump or
 * ##exception.non-proc-jump (i.e. the continuation will be the same as
 * that of the faulty call).  ##exception.global-jump is called
 * when the global variable's name is known (in this case the
 * arguments are the variable's name followed by the arguments of the
 * faulty call).  Otherwise, ##exception.non-proc-jump is called
 * and the arguments are the object that was in operator position
 * followed by the arguments of the faulty call.
 */

{
  int na = ___ps->na;
  int i;
  ___WORD sym, probe, result, fn;

  ___PUSH_ARGS_IN_REGS(na)

  ___PUSH(0)

  for (i=0; i<na; i++)
    ___fp[i] = ___fp[i+1];

  result = ___ps->temp4->val;
  fn = ___G__23__23_exception_2e_non_2d_proc_2d_jump.val;

  i = ___INT(___VECTORLENGTH(___symbol_table)) - 1;

  while (i>=0)
    {
      probe = ___VECTORREF(___symbol_table,___FIX(i));

      while (probe != ___NUL)
        {
          sym = ___CAR(probe);
          if (___ps->temp4 == (___glo_struct*)___VECTORREF(sym,___FIX(2)))
            {
              result = sym;
              fn = ___G__23__23_exception_2e_global_2d_jump.val;
              break;
            }
          probe = ___CDR(probe);
        }
      i--;
    }

  ___fp[na] = result;

  ___POP_ARGS_IN_REGS(na+1)

  ___JUMPEXT(___SET_NARGS(na+1),fn)
}" (0))

(##c-code
"
/*
 * ___LBL(6)
 *
 * This is the wrong number of arguments handler.  It is invoked when
 * a procedure is called with a number of arguments that it does not
 * accept.
 *
 * This handler simply tail calls ##exception.non-proc-jump (i.e.
 * the continuation will be the same as that of the faulty call).  The
 * arguments are the object that was in operator position followed by
 * the arguments of the faulty call.
 */

{
  int na = ___ps->na;
  int i;

  ___PUSH_ARGS_IN_REGS(na)

  ___PUSH(0)

  for (i=0; i<na; i++)
    ___fp[i] = ___fp[i+1];

  if (___RETN_INFO_LINK(((___lbl_struct*)(___ps->temp1-___tSUBTYPED))->flags) == 0)
    ___fp[na] = ___ps->temp1;
  else
    ___fp[na] = ___SELF;

  ___POP_ARGS_IN_REGS(na+1)

  ___JUMPEXT(___SET_NARGS(na+1),___G__23__23_exception_2e_wrong_2d_nb_2d_arg.val)
}" (0))

(##c-code
"
/*
 * ___LBL(7)
 *
 * This is the rest parameter handler.  It is invoked when a non-null
 * rest parameter must be constructed.
 *
 */

{
  int n;
  ___WORD z = ((___lbl_struct*)(___ps->temp1-___tSUBTYPED))->flags;
  int nb_parms = (z>>3)&((1<<14)-1);

  ___PUSH_ARGS_IN_REGS(___ps->na)

  if (___ps->na < nb_parms)
    {
      int na = ___ps->na;
      int i;
      ___PUSH(0)

      for (i=0; i<na; i++)
        ___fp[i] = ___fp[i+1];
      ___fp[na] = ___ps->temp1;

      ___POP_ARGS_IN_REGS(na+1)

      ___JUMPEXT(___SET_NARGS(na+1),___G__23__23_exception_2e_wrong_2d_nb_2d_arg.val)
    }

  ___SET_R1(___NUL)
  n = ___ps->na - nb_parms + 1;
  while (n > 0)
    {
      ___SET_R1(___CONS(___POP,___R1))
      n--;
    }

  ___PUSH(___R1)

  ___POP_ARGS_IN_REGS(nb_parms)

  ___JUMPEXTPRM(___SET_NARGS(-1),___ps->temp1)
}" (0))

(##c-code
"
/*
 * ___LBL(8)
 *
 * This is the keyword parameter handler.  It is invoked when keyword
 * parameters must be processed.
 *
 */

{
  ___WORD z = ((___lbl_struct*)(___ps->temp1-___tSUBTYPED))->flags;
  int nb_parms = (z>>3)&((1<<14)-1);
  ___WORD key_descr = ___ps->temp3;
  int nb_req_opt = ___ps->temp2;
  int nb_key = nb_parms-nb_req_opt;
  int i, j, k;

  ___PUSH_ARGS_IN_REGS(___ps->na)

  if (___ps->na < nb_req_opt ||
      (___ps->na-nb_req_opt) & 1)
    {
      int na = ___ps->na;
      int i;
      ___PUSH(0)

      for (i=0; i<na; i++)
        ___fp[i] = ___fp[i+1];
      ___fp[na] = ___ps->temp1;

      ___POP_ARGS_IN_REGS(na+1)

      ___JUMPEXT(___SET_NARGS(na+1),___G__23__23_exception_2e_wrong_2d_nb_2d_arg.val)
    }

  k = ___ps->na-nb_req_opt;

  for (i=k-1; i>0; i-=2)
    if (!___KEYWORDP(___fp[i]))
      {
        break;
      }

  for (j=nb_key-1; j>=0; j--)
    {
      ___WORD key = ___FIELD(key_descr,j*2);
      ___WORD val = ___FIELD(key_descr,j*2+1);
      for (i=k-1; i>0; i-=2)
        if (___fp[i]==key)
          {
            val = ___fp[i-1];
            break;
          }
      ___fp[-1-j] = val;
    }

  ___fp += k;
  for (i=0; i<nb_key; i++)
    {
      ___fp--;
      *___fp = *(___fp-k);
    }

  ___POP_ARGS_IN_REGS(nb_parms)

  ___JUMPEXTPRM(___SET_NARGS(-1),___ps->temp1)
}" (0))

(##c-code
"
/*
 * ___LBL(9)
 *
 * This is the keyword and rest parameter handler.  It is invoked when
 * keyword parameters must be processed and a rest parameter must be
 * constructed.
 *
 */

{
  exit (1);
}" (0))

(##c-code
"
/*
 * ___LBL(10)
 *
 * This is the force handler.  It is invoked when a promise is forced.
 */

{
  ___WORD x = ___ps->temp2;
  ___WORD y = ___UNTAG_AS(x,___tSUBTYPED)[2];
  if (x != y)
    {
      ___ps->temp2 = y;
      ___JUMPEXTPRM(___NOTHING,___ps->temp1)
    }
  else
    {
      ___WORD z = ___UNTAG_AS(x,___tSUBTYPED)[1];
      ___PUSH_REGS
      ___PUSH(___ps->temp1)
      ___PUSH(x)
      ___PUSH(z)
      ___POP_ARGS_IN_REGS(2)
      ___SET_R0(___internal_return)
      ___JUMPEXT(___SET_NARGS(2),___G__23__23_force_2d_undetermined.val)
    }
}" (0))

(##c-code
"
/*
 * ___LBL(11)
 *
 * This is the C-conversion error handler.  It is invoked when a
 * conversion error is detected by the C-interface.
 */

{
  int na = ___ps->na;
  int i;

  ___SET_R0(___POP)

  ___PUSH(0)

  for (i=0; i<na; i++)
    ___fp[i] = ___fp[i+1];
  ___fp[na] = ___ps->temp1;

  ___POP_ARGS_IN_REGS(na+1)

  ___JUMPEXT(___SET_NARGS(na+1),___G__23__23_exception_2e_c_2d_conv_2d_error.val)
}" (0))

(##c-code
"
/*
 * ___LBL(12)
 *
 * This is the internal-return handler.  It is invoked when an
 * internal return point is returned to.
 */

___ps->temp2 = ___R1; /* for return from ##force-undetermined */
___temp = ___POP;
___POP_REGS
___JUMPEXTPRM(___NOTHING,___temp)
" (0))

)))

(##setup-traps)

(##declare (interrupts-enabled))

;------------------------------------------------------------------------------

; Memory allocation

(define (##make-vector n init)
  (let ((v (##c-code "
{
long i;
long n = ___INT(___ARG1);
long words = n + 1;
if (words > ___MSECTION_BIGGEST)
  {
    ___WORD result;
    ___W_ALL
    result = ___alloc_scmobj (___sVECTOR, n<<___LWS, ___STILL);
    ___R_ALL
    ___RESULT = result;
    if (___RESULT != ___FAL)
      ___still_obj_refcount_dec (___RESULT);
  }
else
  {
    int overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___W_ALL
        overflow = ___heap_limit ();
        ___R_ALL
      }
    else
      ___hp -= words;
    if (overflow)
      ___RESULT = ___FAL;
    else
      {
        ___RESULT = ___TAG(___hp, ___tSUBTYPED);
        ___HEADER(___RESULT) = ___MAKE_HD_WORDS(n, ___sVECTOR);
        ___hp += words;
      }
  }
if (___RESULT != ___FAL)
  {
    for (i=0; i<n; i++)
      ___VECTORSET(___RESULT,___FIX(i),___ARG2)
  }
}
" n init)))
    (or v
        (begin
          (##exception.heap-overflow)
          (##make-vector n init)))))

(define (##make-string n init)
  (let ((s (##c-code "
{
long i;
long n = ___INT(___ARG1);
long words = ___WORDS((n<<___LCS)) + 1;
if (words > ___MSECTION_BIGGEST)
  {
    ___WORD result;
    ___W_ALL
    result = ___alloc_scmobj (___sSTRING, n<<___LCS, ___STILL);
    ___R_ALL
    ___RESULT = result;
    if (___RESULT != ___FAL)
      ___still_obj_refcount_dec (___RESULT);
  }
else
  {
    int overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___W_ALL
        overflow = ___heap_limit ();
        ___R_ALL
      }
    else
      ___hp -= words;
    if (overflow)
      ___RESULT = ___FAL;
    else
      {
        ___RESULT = ___TAG(___hp, ___tSUBTYPED);
        ___HEADER(___RESULT) = ___MAKE_HD_BYTES((n<<___LCS), ___sSTRING);
        ___hp += words;
      }
  }
if (___RESULT != ___FAL)
  {
    for (i=0; i<n; i++)
      ___STRINGSET(___RESULT,___FIX(i),___ARG2);
  }
}
" n init)))
    (or s
        (begin
          (##exception.heap-overflow)
          (##make-string n init)))))

(define (##make-vector8 n init)
  (let ((v (##c-code "
{
long i;
long n = ___INT(___ARG1);
long words = ___WORDS(n) + 1;
if (words > ___MSECTION_BIGGEST)
  {
    ___WORD result;
    ___W_ALL
    result = ___alloc_scmobj (___sVECTOR8, n, ___STILL);
    ___R_ALL
    ___RESULT = result;
    if (___RESULT != ___FAL)
      ___still_obj_refcount_dec (___RESULT);
  }
else
  {
    int overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___W_ALL
        overflow = ___heap_limit ();
        ___R_ALL
      }
    else
      ___hp -= words;
    if (overflow)
      ___RESULT = ___FAL;
    else
      {
        ___RESULT = ___TAG(___hp, ___tSUBTYPED);
        ___HEADER(___RESULT) = ___MAKE_HD_BYTES(n, ___sVECTOR8);
        ___hp += words;
      }
  }
if (___RESULT != ___FAL)
  {
    for (i=0; i<n; i++)
      ___VECTOR8SET(___RESULT,___FIX(i),___ARG2);
  }
}
" n init)))
    (or v
        (begin
          (##exception.heap-overflow)
          (##make-vector8 n init)))))

(define (##make-vector16 n init)
  (let ((v (##c-code "
{
long i;
long n = ___INT(___ARG1);
long words = ___WORDS((n<<1)) + 1;
if (words > ___MSECTION_BIGGEST)
  {
    ___WORD result;
    ___W_ALL
    result = ___alloc_scmobj (___sVECTOR16, n<<1, ___STILL);
    ___R_ALL
    ___RESULT = result;
    if (___RESULT != ___FAL)
      ___still_obj_refcount_dec (___RESULT);
  }
else
  {
    int overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___W_ALL
        overflow = ___heap_limit ();
        ___R_ALL
      }
    else
      ___hp -= words;
    if (overflow)
      ___RESULT = ___FAL;
    else
      {
        ___RESULT = ___TAG(___hp, ___tSUBTYPED);
        ___HEADER(___RESULT) = ___MAKE_HD_BYTES((n<<1), ___sVECTOR16);
        ___hp += words;
      }
  }
if (___RESULT != ___FAL)
  {
    for (i=0; i<n; i++)
      ___VECTOR16SET(___RESULT,___FIX(i),___ARG2);
  }
}
" n init)))
    (or v
        (begin
          (##exception.heap-overflow)
          (##make-vector16 n init)))))

;------------------------------------------------------------------------------

; Procedure application and continuations

(define ##apply
  (let ()
    (##declare (not inline))
    (lambda (proc args)
      (##c-code "
{
  ___WORD proc = ___ARG1; /* procedure to call   */
  ___WORD args = ___ARG2; /* arguments to pass   */
  int n = 0;              /* number of arguments */

  while (___PAIRP(args))
    {
      ___PUSH(___CAR(args))
      args = ___CDR(args);
      n++;
    }

  ___POP_ARGS_IN_REGS(n)

  ___JUMPEXT(___SET_NARGS(n),proc)
}"
        proc
        args))))

(define ##call-with-current-continuation
  (let ()
    (##declare (not inline))
    (lambda (p)
      (let ((v (##c-code "
{
  ___WORD frame, ra;

  if (___R0 == ___ps->handler_break)
    {
      frame = ___fp[1];
      ra    = ___fp[0];
    }
  else
    {
      frame = (___WORD)___fp;
      ra    = ___R0;
      ___fp[-1] = frame;
      ___fp[-2] = ra;
      ___fp -= 2;
      ___ps->stack_break = ___fp;
      ___R0 = ___ps->handler_break;
    }

  {___WORD *___ptr=___hp+1;
  *___hp=___MAKE_HD_WORDS(___CONTINUATION_SIZE,___sCONTINUATION);
  ___hp+=___CONTINUATION_SIZE+1;
  ___ADD_VECTOR_ELEM(frame)
  ___ADD_VECTOR_ELEM(ra)
  ___ADD_VECTOR_ELEM(___ARG1)}
  ___RESULT = ___GET_VECTOR(___CONTINUATION_SIZE);
}" ##dynamic-environment)))
        (p
          (lambda (x)
            (##c-code "
{
  ___WORD v = ___ARG1;
  ___WORD x = ___ARG2;
  ___fp = ___ps->stack_break;
  ___fp[1] = ___VECTORREF(v,___FIX(0));
  ___fp[0] = ___VECTORREF(v,___FIX(1));
  ___G__23__23_dynamic_2d_environment.val = ___VECTORREF(v,___FIX(2));
  ___SET_R1(x)
  ___JUMPEXTPRM(___NOTHING,___ps->handler_break)
}" v x)))))))

(define (##continuation->frame c)
  (##closure-ref c 1))

(define (##frame-ret f)
  (##c-code "
{
  ___WORD f = ___ARG1;
  ___WORD frame = ___VECTORREF(f,___FIX(0));
  ___WORD ra    = ___VECTORREF(f,___FIX(1));

  if (___TYP(frame)==___tSUBTYPED)
    {
      /* continuation's frame is in the heap */

      if (ra == ___internal_return)
        ___RESULT = ___BODY_AS(frame,___tSUBTYPED)[1];
      else
        ___RESULT = ra;
    }
  else
    {
      /* continuation's frame is in the stack */

      if (ra == ___internal_return)
        ___RESULT = ((___WORD*)frame)[0];
      else
        ___RESULT = ra;
    }
}" f))

(define (##frame-dyn-env f)
  (##vector-ref f 2))

(define (##frame-fs f)
  (##c-code "
{
  ___WORD f = ___ARG1;
  ___WORD frame = ___VECTORREF(f,___FIX(0));
  ___WORD ra    = ___VECTORREF(f,___FIX(1));
  int fs;

  if (___TYP(frame)==___tSUBTYPED)
    {
      /* continuation's frame is in the heap */

      if (ra == ___internal_return)
        {
          ___WORD fs_link = ((___lbl_struct*)(___BODY_AS(frame,___tSUBTYPED)[1]-___tSUBTYPED))->flags;
          fs = (fs_link>>3)&((1<<14)-1);
        }
      else
        {
          ___WORD fs_link = ((___lbl_struct*)(ra-___tSUBTYPED))->flags;
          fs = (fs_link>>3)&((1<<14)-1);
        }
    }
  else
    {
      /* continuation's frame is in the stack */

      if (ra == ___internal_return)
        {
          ___WORD fs_link = ((___lbl_struct*)(((___WORD*)frame)[0]-___tSUBTYPED))->flags;
          fs = (fs_link>>3)&((1<<14)-1);
        }
      else
        {
          ___WORD fs_link = ((___lbl_struct*)(ra-___tSUBTYPED))->flags;
          fs = (fs_link>>3)&((1<<14)-1);
        }
    }

  ___RESULT = ___FIX(fs);
}" f))

(define (##frame-stk-ref f i)
  (##c-code "
{
  ___WORD f = ___ARG1;
  int i = ___INT(___ARG2);
  ___WORD frame = ___VECTORREF(f,___FIX(0));
  ___WORD ra    = ___VECTORREF(f,___FIX(1));
  int fs;

  if (___TYP(frame)==___tSUBTYPED)
    {
      /* continuation's frame is in the heap */

      if (ra == ___internal_return)
        {
          ___WORD fs_link = ((___lbl_struct*)(___BODY_AS(frame,___tSUBTYPED)[1]-___tSUBTYPED))->flags;
          fs = (fs_link>>3)&((1<<14)-1);
        }
      else
        {
          ___WORD fs_link = ((___lbl_struct*)(ra-___tSUBTYPED))->flags;
          fs = (fs_link>>3)&((1<<14)-1);
        }
      ___RESULT = ___BODY_AS(frame,___tSUBTYPED)[fs-i+1];  /* what if i==link and frame is first in section???? */
    }
  else
    {
      /* continuation's frame is in the stack */

      if (ra == ___internal_return)
        {
          ___WORD fs_link = ((___lbl_struct*)(((___WORD*)frame)[0]-___tSUBTYPED))->flags;
          fs = (fs_link>>3)&((1<<14)-1);
        }
      else
        {
          ___WORD fs_link = ((___lbl_struct*)(ra-___tSUBTYPED))->flags;
          fs = (fs_link>>3)&((1<<14)-1);
        }
      ___RESULT = ((___WORD*)frame)[fs-i];  /* what if i==link and frame is first in section???? */
    }

}" f i))

(define (##frame-next f)
  (##c-code "
{
  ___WORD f = ___ARG1;
  ___WORD frame = ___VECTORREF(f,___FIX(0));
  ___WORD ra    = ___VECTORREF(f,___FIX(1));
  ___WORD denv  = ___VECTORREF(f,___FIX(2));
  ___WORD *fp, frame_ra, next_ra, next_frame;
  int fs, link;

  if (___TYP(frame)==___tSUBTYPED)
    {
      /* continuation's frame is in the heap */

      fp = ___BODY_AS(frame,___tSUBTYPED);

      if (ra == ___internal_return)
        {
          ___WORD fs_link = ((___lbl_struct*)(fp[1]-___tSUBTYPED))->flags;
          fs = ___RETI_INFO_FS(fs_link);
          link = ___RETI_INFO_LINK(fs_link);
        }
      else
        {
          ___WORD fs_link = ((___lbl_struct*)(ra-___tSUBTYPED))->flags;
          fs = ___RETN_INFO_FS(fs_link);
          link = ___RETN_INFO_LINK(fs_link);
        }
      fp += fs+1;

      next_ra = fp[-link-1];
      next_frame = fp[-fs-1];

      if (next_frame == 0)
        ___RESULT = ___FAL;
      else
        {
          {___WORD *___ptr=___hp+1;
          *___hp=___MAKE_HD_WORDS(___CONTINUATION_SIZE,___sCONTINUATION);
          ___hp+=___CONTINUATION_SIZE+1;
          ___ADD_VECTOR_ELEM(next_frame)
          ___ADD_VECTOR_ELEM(next_ra)
          ___ADD_VECTOR_ELEM(denv)}
          ___RESULT = ___GET_VECTOR(___CONTINUATION_SIZE);
        }
    }
  else
    {
      /* continuation's frame is in the stack */

      if (ra == ___internal_return)
        {
          ___WORD fs_link = ((___lbl_struct*)(((___WORD*)frame)[0]-___tSUBTYPED))->flags;
          fs = ___RETI_INFO_FS(fs_link);
          link = ___RETI_INFO_LINK(fs_link);
        }
      else
        {
          ___WORD fs_link = ((___lbl_struct*)(ra-___tSUBTYPED))->flags;
          fs = ___RETN_INFO_FS(fs_link);
          link = ___RETN_INFO_LINK(fs_link);
        }
      fp = ((___WORD*)frame)+fs;
      frame_ra = fp[-link-1];

      if (frame_ra == ___ps->handler_break)
        {
          /* first frame of that section */

          next_ra    = fp[0];
          next_frame = fp[1];
        }
      else
        {
          /* not the first frame of that section */

          next_ra    = frame_ra;
          next_frame = (___WORD)fp;
        }

      if (next_frame == 0)
        ___RESULT = ___FAL;
      else
        {
          {___WORD *___ptr=___hp+1;
          *___hp=___MAKE_HD_WORDS(___CONTINUATION_SIZE,___sCONTINUATION);
          ___hp+=___CONTINUATION_SIZE+1;
          ___ADD_VECTOR_ELEM(next_frame)
          ___ADD_VECTOR_ELEM(next_ra)
          ___ADD_VECTOR_ELEM(denv)}
          ___RESULT = ___GET_VECTOR(___CONTINUATION_SIZE);
        }
    }
}" f))

(define (##closure? x)
  (##c-code "
{
  if (___TYP(___ARG1)==___tSUBTYPED&&((___lbl_struct*)(___ARG1-___tSUBTYPED))->entry!=___ARG1)
    ___RESULT = ___TRU;
  else
    ___RESULT = ___FAL;
}" x))

(define (##subprocedure? x)
  (##c-code "
{
  if ((___TYP(___ARG1)==___tSUBTYPED) &&
      (((___lbl_struct*)(___ARG1-___tSUBTYPED))->entry==___ARG1) &&
      (((___WORD *)(___ARG1-___tSUBTYPED))[-___LS] != ___MAKE_HD((___INTRO_SIZE<<___LWS),___sVECTOR,___RCN,___PERM)))
    ___RESULT = ___TRU;
  else
    ___RESULT = ___FAL;
}" x))

(define (##subprocedure-id x)
  (##c-code "
{
  if (___TYP(___ARG1)==___tSUBTYPED)
  {
    ___WORD *start = ((___WORD *)(___ARG1-___tSUBTYPED));
    ___WORD *ptr = start;
    while (*ptr != ___MAKE_HD((___INTRO_SIZE<<___LWS),___sVECTOR,___RCN,___PERM)) ptr -= ___LS;
    ptr += ___LS;
    ___RESULT = ___FIX( (start-ptr)/___LS );
  }
  else
    ___RESULT = ___FIX(0);
}" x))

(define (##subprocedure-parent x)
  (##c-code "
{
  if (___TYP(___ARG1)==___tSUBTYPED)
  {
    ___WORD *start = ((___WORD *)(___ARG1-___tSUBTYPED));
    ___WORD *ptr = start;
    while (*ptr != ___MAKE_HD((___INTRO_SIZE<<___LWS),___sVECTOR,___RCN,___PERM)) ptr -= ___LS;
    ptr += ___LS;
   ___RESULT = ___TAG(ptr,___tSUBTYPED);
  }
  else
    ___RESULT = ___FAL;
}" x))

(define (##procedure-info x)
  (##c-code "
{
  if (___TYP(___ARG1)==___tSUBTYPED)
  {
    ___WORD *start = ((___WORD *)(___ARG1-___tSUBTYPED));
    ___WORD *ptr = start;
    while (*ptr != ___MAKE_HD((___INTRO_SIZE<<___LWS),___sVECTOR,___RCN,___PERM)) ptr -= ___LS;
   ___RESULT = ptr[1];
  }
  else
    ___RESULT = ___FAL;
}" x))

(define ##dynamic-environment '())

(define (##dynamic-env-bind dyn-env thunk)
  (let ((env ##dynamic-environment))
    (set! ##dynamic-environment dyn-env)
    (let ((val (thunk)))
      (set! ##dynamic-environment env)
      val)))

(define (##dynamic-env-ref)
  ##dynamic-environment)

;------------------------------------------------------------------------------

; Global variables

(define (##make-global-var id)
  (##c-code "
if (___VECTORREF(___ARG1,___FIX(2)) == 0)
{
___glo_struct *p = ___alloc_global_var ();
if (p == 0)
  ___RESULT = ___FAL;
else
{
  p->val = ___UNB1;
  p->prm = ___FAL;
  p->next = ___ps->glo_list;
  ___ps->glo_list = p;
  ___VECTORSET(___ARG1,___FIX(2),(___WORD)p)
  ___RESULT = ___ARG1;
}
}
else
  ___RESULT = ___ARG1;
" id))

(define (##object->global-var->identifier val)
  (let ((sym-tbl (##symbol-table)))
    (let loop1 ((i (##fixnum.- (##vector-length sym-tbl) 1)) (res #f))
      (if (##fixnum.< i 0)
        res
        (let loop2 ((l (##vector-ref sym-tbl i)) (res res))
          (if (##null? l)
            (loop1 (##fixnum.- i 1) res)
            (loop2 (##cdr l)
                   (##c-code
"
if ((___VECTORREF(___ARG2,___FIX(2)) != 0) &&
    (*(___WORD*)___VECTORREF(___ARG2,___FIX(2)) == ___ARG1) &&
    ((___ARG3 == ___FAL) ||
     (___VECTORREF(___ARG2,___FIX(2)) < ___VECTORREF(___ARG3,___FIX(2)))))
  ___RESULT = ___ARG2;
else
  ___RESULT = ___ARG3;
" val (##car l) res))))))))

(define (##global-var->identifier gv)
  gv)

(define (##make-interned-symbol name)
  (##make-interned-symkey name #t))

(define (##make-interned-keyword name)
  (##make-interned-symkey name #f))

(define (##make-interned-symkey name symbol?)
  (let ((obj ((c-lambda (scheme-object boolean) scheme-object "
{
  unsigned int subtype = ___arg2 ? ___sSYMBOL : ___sKEYWORD;
  ___WORD obj = ___find_symkey_from_schemestring (___arg1, subtype);
  if (obj == ___FAL)
    {
      unsigned long n = ___INT(___STRINGLENGTH(___arg1));
      obj = ___alloc_scmobj (___sSTRING, n<<___LCS, ___PERM);
      if (obj != ___FAL)
        {
          memcpy (___BODY_AS(obj,___tSUBTYPED),
                  ___BODY_AS(___arg1,___tSUBTYPED),
                  n<<___LCS);
          obj = ___new_symkey (obj, subtype);
        }
    }
  ___result = obj;
}
") name symbol?)))
    (or obj
        (begin
          (##exception.heap-overflow)
          (##make-interned-symkey name symbol?)))))

(define ##symbol-hash-counter 0)

(define (##make-uninterned-symbol name) ; name must be a non-mutable string
  (let ((i ##symbol-hash-counter))
    (set! ##symbol-hash-counter
      (##fixnum.modulo (##fixnum.+ ##symbol-hash-counter 1) 65536))
    (##subtype-set! (##vector name i 0) (subtype-symbol))))

;------------------------------------------------------------------------------

; Miscellaneous

(define ##max-unicode
  (##c-code "___RESULT = ___FIX(___MAX_CHR);"))

(define ##symbol-table
  (c-lambda () scheme-object "___result = ___symbol_table;"))

(define ##keyword-table
  (c-lambda () scheme-object "___result = ___keyword_table;"))

(define ##argv
  (c-lambda () scheme-object "___result = ___arguments;"))

(define ##processed-argv '())
(set! ##processed-argv (##argv))

(define (argv) ##processed-argv)

(define (##process-time)
  (##c-code "___RESULT = ___FIX(___process_time ());"))

(define (##check-heap)
  (let ((x (##c-code "
if (___hp > ___ps->heap_limit)
  {
    int overflow;
    ___W_ALL
    overflow = ___heap_limit ();
    ___R_ALL
    if (overflow)
      ___RESULT = ___TRU;
    else
      ___RESULT = ___FAL;
  }
else
  ___RESULT = ___FAL;
")))
    (if x
      (begin
        (##exception.heap-overflow)
        (##check-heap)))))

(define (##force-undetermined ph thunk)
  (let ((val (##force (thunk))))
    (##c-code "
if (___UNTAG_AS(___ARG1,___tSUBTYPED)[2] == ___ARG1)
  ___UNTAG_AS(___ARG1,___tSUBTYPED)[2] = ___ARG2;
___RESULT = ___UNTAG_AS(___ARG1,___tSUBTYPED)[2];
" ph val)))

(define (##exit #!optional (status 0))
  (##c-code "___cleanup (); exit (___INT(___ARG1));" status))

(define (##sequentially thunk)
  (thunk))

(define ##default-ed-file "noname.scm")

(define (ed #!optional (filename ##default-ed-file))
  (if (##string? filename)
    (begin
      (set! ##default-ed-file filename)
      (##shell-command (##string-append "editor " filename))
      filename)
    (##trap-check-string 'ed filename)))

(define ##shell-command
  (c-lambda (char-string) int "___shell_command"))

(define ##path-expand
  (c-lambda (char-string) char-string "
char path[___PATH_MAX_LENGTH+1];
if (___path_expand (___arg1, path, ___PATH_MAX_LENGTH))
  ___result = path;
else
  ___result = 0;
"))

(define ##path-absolute?
  (c-lambda (char-string) boolean "___path_absolute"))

(define ##path-extension
  (c-lambda (char-string) char-string "
char path[___PATH_MAX_LENGTH+1];
if (___path_extension (___arg1, path, ___PATH_MAX_LENGTH))
  ___result = path;
else
  ___result = 0;
"))

(define ##path-strip-extension
  (c-lambda (char-string) char-string "
char path[___PATH_MAX_LENGTH+1];
if (___path_strip_extension (___arg1, path, ___PATH_MAX_LENGTH))
  ___result = path;
else
  ___result = 0;
"))

(define ##path-directory
  (c-lambda (char-string) char-string "
char path[___PATH_MAX_LENGTH+1];
if (___path_directory (___arg1, path, ___PATH_MAX_LENGTH))
  ___result = path;
else
  ___result = 0;
"))

(define ##path-strip-directory
  (c-lambda (char-string) char-string "
char path[___PATH_MAX_LENGTH+1];
if (___path_strip_directory (___arg1, path, ___PATH_MAX_LENGTH))
  ___result = path;
else
  ___result = 0;
"))

(define ##path-drive
  (c-lambda (char-string) char-string "
char path[___PATH_MAX_LENGTH+1];
if (___path_drive (___arg1, path, ___PATH_MAX_LENGTH))
  ___result = path;
else
  ___result = 0;
"))

(define ##path-strip-drive
  (c-lambda (char-string) char-string "
char path[___PATH_MAX_LENGTH+1];
if (___path_strip_drive (___arg1, path, ___PATH_MAX_LENGTH))
  ___result = path;
else
  ___result = 0;
"))

(define (##load-object-file name)
  (let ((vect ((c-lambda (char-string) scheme-object "
{
  char *errmsg;
#if 0
  if (___gc ()) /* to prevent gc while loading */
    ___result = ___FAL;
  else
#endif
    {
      ___WORD result = ___load_object_file (___arg1, &errmsg);
      if (result == ___FAL)
        {
          ___err = ___charstring_to_scmobj (errmsg, &result);
          ___release_scmobj (result);
        }
      ___result = result;
    }
}
") name)))
    (cond ((##vector? vect)
           (lambda ()
             (let ((len (##vector-length vect)))
               (let loop ((i 0))
                 (if (##fixnum.< i len)
                   (if (##fixnum.= i (##fixnum.- len 1))
                     ((##vector-ref vect i))
                     (begin
                       ((##vector-ref vect i))
                       (loop (##fixnum.+ i 1)))))))))
          ((##string? vect)
           vect)
          (else
           (##exception.heap-overflow)
           (##load-object-file name)))))

(define ##cc-prog1
  ((c-lambda () char-string "___result = ___CC_PROG1;")))

(define ##cc-prog2
  ((c-lambda () char-string "___result = ___CC_PROG2;")))

(define ##cc-prog3
  ((c-lambda () char-string "___result = ___CC_PROG3;")))

(define ##ld-prog1
  ((c-lambda () char-string "___result = ___LD_PROG1;")))

(define ##ld-prog2
  ((c-lambda () char-string "___result = ___LD_PROG2;")))

(define ##ld-prog3
  ((c-lambda () char-string "___result = ___LD_PROG3;")))

(define ##c-compile-dynamic #f)

(if ##cc-prog1
  (set! ##c-compile-dynamic
    (lambda (root-filename output-filename)
      (let ((res1 (##shell-command
                    (##string-append
                      ##cc-prog1
                      root-filename
                      ##cc-prog2
                      root-filename
                      ##cc-prog3))))
        (if (##fixnum.= res1 0)
          (let ((res2 (##shell-command
                        (##string-append
                          ##ld-prog1
                          output-filename
                          ##ld-prog2
                          root-filename
                          ##ld-prog3))))
            (##fixnum.= res2 0))
          #f)))))

(##declare (separate))

(define ##gc-report #f)

(define (##gc)
  (let ((result (##c-code "
{
  int overflow;
  ___W_ALL
  overflow = ___gc ();
  ___R_ALL
  ___RESULT = ___BOOL(overflow);
}
")))
    (and result
         (begin
           (##exception.heap-overflow)
           (##gc)))))

(define (##gc-finalization)
  (##declare (not interrupts-enabled))
  (let ((will (##c-code "
{
  ___WORD will = ___ps->executable_wills;
  if (will == ___FAL)
    ___RESULT = ___FAL;
  else
    {
      ___WORD owner = ___BODY_AS(will,___tSUBTYPED)[0];
      ___WORD head;
      ___ps->executable_wills = ___BODY_AS(will,___tSUBTYPED)[2];
      if (___MEM_ALLOCATED(owner) && ___UNREFERENCED((head=___BODY(owner)[-1])))
        {
          ___BODY(owner)[-1] = ___ADD_RC(head, ___RCN);
          ___BODY_AS(will,___tSUBTYPED)[2] = owner;
          ___BODY_AS(will,___tSUBTYPED)[0] = ___FAL;
          ___RESULT = will;
        }
      else
        {
          ___BODY_AS(will,___tSUBTYPED)[2] = ___ps->non_executable_wills;
          ___ps->non_executable_wills = will;
          ___RESULT = ___TRU;
        }
    }
}")))
    (if will
      (begin
        (if (##will? will)
          ((##vector-ref will 1) (##vector-ref will 2)))
        (##gc-finalization)))))

; execute each module in sequence

(define ##initial-continuation
  (let ()
    (##declare (not inline))
    (lambda ()
      (let loop ((i 1))
        (let ((ev (##c-code "___RESULT = ___exec_vector;")))
          (let ((len (##vector-length ev)))
            (if (##fixnum.< i len)
              (begin
                ((##vector-ref ev i))
                (loop (##fixnum.+ i 1))))))))))

(##initial-continuation)
