;==============================================================================

; file: "_num1.scm"

; Copyright (C) 1994-1996 by Marc Feeley, All Rights Reserved.

(##include "header.scm")

;------------------------------------------------------------------------------

; Numerical type predicates

(define (##complex? x)
  (number-dispatch x #f #t #t #t #t #t))

(define (number? x)  (force-vars (x) (##complex? x)))
(define (complex? x) (force-vars (x) (##complex? x)))

(define (##real? x)
  (number-dispatch x #f #t #t #t #t (cpxnum-real? x)))

(define (real? x) (force-vars (x) (##real? x)))

(define (##rational? x)
  (number-dispatch x #f #t #t #t #t (cpxnum-real? x)))

(define (rational? x) (force-vars (x) (##rational? x)))

(define (##integer? x)
  (number-dispatch x #f #t #t #f (flonum-int? x) (cpxnum-int? x)))

(define (integer? x) (force-vars (x) (##integer? x)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Exactness predicates

(define (##exact? x)

  (define (type-error) (##trap-check-number 'exact? x))

  (number-dispatch x (type-error) #t #t #t #f
    (and (##not (##flonum? (cpxnum-real x)))
         (##not (##flonum? (cpxnum-imag x))))))

(define (exact? x)   (force-vars (x) (##exact? x)))
(define (inexact? x) (force-vars (x) (##not (##exact? x))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Numerical comparison predicates

(define (##eqv? x y)
  (number-dispatch x (##eq? x y)
    (if (##fixnum? y) (##fixnum.= x y) #f)
    (if (##bignum? y) (##bignum.= x y) #f)
    (if (##ratnum? y) (##ratnum.= x y) #f)
    (and (##complex? y) (##not (##exact? y)) (##= x y))
    (and (##complex? y) (##eq? (##exact? x) (##exact? y)) (##= x y))))

(define (##= x y)

  (define (type-error) (##trap-check-number '= x y))

  (number-dispatch x (type-error)

    (number-dispatch y (type-error) ; x = fixnum
      (##fixnum.= x y)
      #f
      #f
      (##flonum.= (##flonum.<-fixnum x) y)
      (##cpxnum.= (##cpxnum.<-non-cpxnum x) y))

    (number-dispatch y (type-error) ; x = bignum
      #f
      (##bignum.= x y)
      #f
      (##flonum.= (##flonum.<-bignum x) y)
      (##cpxnum.= (##cpxnum.<-non-cpxnum x) y))

    (number-dispatch y (type-error) ; x = ratnum
      #f
      #f
      (##ratnum.= x y)
      (##ratnum.= x (##flonum.->ratnum y))
      (##cpxnum.= (##cpxnum.<-non-cpxnum x) y))

    (number-dispatch y (type-error) ; x = flonum
      (##ratnum.= (##flonum.->ratnum x) (##ratnum.<-exact-int y))
      (##ratnum.= (##flonum.->ratnum x) (##ratnum.<-exact-int y))
      (##ratnum.= (##flonum.->ratnum x) y)
      (##flonum.= x y)
      (##cpxnum.= (##cpxnum.<-non-cpxnum x) y))

    (number-dispatch y (type-error) ; x = cpxnum
      (##cpxnum.= x (##cpxnum.<-non-cpxnum y))
      (##cpxnum.= x (##cpxnum.<-non-cpxnum y))
      (##cpxnum.= x (##cpxnum.<-non-cpxnum y))
      (##cpxnum.= x (##cpxnum.<-non-cpxnum y))
      (##cpxnum.= x y))))

(define-nary0-bool (= x y) (##= x y) no-check force-vars)

(define (##< x y)

  (define (type-error) (##trap-check-real '< x y))

  (number-dispatch x (type-error)

    (number-dispatch y (type-error) ; x = fixnum
      (##fixnum.< x y)
      (bignum-positive? y)
      (##ratnum.< (##ratnum.<-exact-int x) y)
      (##flonum.< (##flonum.<-fixnum x) y)
      (if (cpxnum-real? y) (##< x (cpxnum-real y)) (type-error)))

    (number-dispatch y (type-error) ; x = bignum
      (bignum-negative? x)
      (##bignum.< x y)
      (##ratnum.< (##ratnum.<-exact-int x) y)
      (##flonum.< (##flonum.<-bignum x) y)
      (if (cpxnum-real? y) (##< x (cpxnum-real y)) (type-error)))

    (number-dispatch y (type-error) ; x = ratnum
      (##ratnum.< x (##ratnum.<-exact-int y))
      (##ratnum.< x (##ratnum.<-exact-int y))
      (##ratnum.< x y)
      (##ratnum.< x (##flonum.->ratnum y))
      (if (cpxnum-real? y) (##< x (cpxnum-real y)) (type-error)))

    (number-dispatch y (type-error) ; x = flonum
      (##ratnum.< (##flonum.->ratnum x) (##ratnum.<-exact-int y))
      (##ratnum.< (##flonum.->ratnum x) (##ratnum.<-exact-int y))
      (##ratnum.< (##flonum.->ratnum x) y)
      (##flonum.< x y)
      (if (cpxnum-real? y) (##< x (cpxnum-real y)) (type-error)))

    (if (cpxnum-real? x) ; x = cpxnum
      (number-dispatch y (type-error)
        (##< (cpxnum-real x) y)
        (##< (cpxnum-real x) y)
        (##< (cpxnum-real x) y)
        (##< (cpxnum-real x) y)
        (if (cpxnum-real? y) (##< (cpxnum-real x) (cpxnum-real y)) (type-error)))
      (type-error))))

(define-nary0-bool (<  x y) (##< x y) no-check force-vars)
(define-nary0-bool (>  x y) (##< y x) no-check force-vars)
(define-nary0-bool (<= x y) (##not (##< y x)) no-check force-vars)
(define-nary0-bool (>= x y) (##not (##< x y)) no-check force-vars)

(define (##zero? x)

  (define (type-error) (##trap-check-number 'zero? x))

  (number-dispatch x (type-error) (##fixnum.zero? x) #f #f (##flonum.zero? x)
    (let ((imag (cpxnum-imag x)))
      (and (flonum-zero? imag)
           (let ((real (cpxnum-real x)))
             (non-cpxnum-zero? real))))))

(define (zero? x) (force-vars (x) (##zero? x)))

(define (##positive? x)

  (define (type-error) (##trap-check-real 'positive? x))

  (number-dispatch x (type-error)
    (##fixnum.positive? x)
    (bignum-positive? x)
    (##positive? (ratnum-numerator x))
    (##flonum.positive? x)
    (if (cpxnum-real? x) (##positive? (cpxnum-real x)) (type-error))))

(define (positive? x) (force-vars (x) (##positive? x)))

(define (##negative? x)

  (define (type-error) (##trap-check-real 'negative? x))

  (number-dispatch x (type-error)
    (##fixnum.negative? x)
    (bignum-negative? x)
    (##negative? (ratnum-numerator x))
    (##flonum.negative? x)
    (if (cpxnum-real? x) (##negative? (cpxnum-real x)) (type-error))))

(define (negative? x) (force-vars (x) (##negative? x)))

(define (##odd? x)

  (define (type-error) (##trap-check-integer 'odd? x))

  (number-dispatch x (type-error)
    (##fixnum.odd? x)
    (bignum-odd? x)
    (type-error)
    (if (flonum-int? x) (##odd? (##flonum.->exact-int x)) (type-error))
    (if (cpxnum-int? x) (##odd? (cpxnum-real x)) (type-error))))

(define (odd? x)  (force-vars (x) (##odd? x)))
(define (even? x) (force-vars (x) (##not (##odd? x))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Max and min

(define (##max x y)

  (define (type-error) (##trap-check-real 'max x y))

  (define (m x y) (if (##< x y) y x))

  (number-dispatch x (type-error)

    (number-dispatch y (type-error) ; x = fixnum
      (m x y)
      (m x y)
      (m x y)
      (if (##< x y) y (##flonum.<-fixnum x))
      (if (cpxnum-real? y) (##max x (cpxnum-real y)) (type-error)))

    (number-dispatch y (type-error) ; x = bignum
      (m x y)
      (m x y)
      (m x y)
      (if (##< x y) y (##flonum.<-bignum x))
      (if (cpxnum-real? y) (##max x (cpxnum-real y)) (type-error)))

    (number-dispatch y (type-error) ; x = ratnum
      (m x y)
      (m x y)
      (m x y)
      (if (##< x y) y (##flonum.<-ratnum x))
      (if (cpxnum-real? y) (##max x (cpxnum-real y)) (type-error)))

    (number-dispatch y (type-error) ; x = flonum
      (if (##< x y) (##flonum.<-fixnum y) x)
      (if (##< x y) (##flonum.<-bignum y) x)
      (if (##< x y) (##flonum.<-ratnum y) x)
      (m x y)
      (if (cpxnum-real? y) (##max x (cpxnum-real y)) (type-error)))

    (if (cpxnum-real? x) ; x = cpxnum
      (number-dispatch y (type-error)
        (##max (cpxnum-real x) y)
        (##max (cpxnum-real x) y)
        (##max (cpxnum-real x) y)
        (##max (cpxnum-real x) y)
        (if (cpxnum-real? y) (##max (cpxnum-real x) (cpxnum-real y)) (type-error)))
      (type-error))))

(define-nary1 (max x y) x (##max x y) force-vars)

(define (##min x y)

  (define (type-error) (##trap-check-real 'min x y))

  (define (m x y) (if (##< x y) x y))

  (number-dispatch x (type-error)

    (number-dispatch y (type-error) ; x = fixnum
      (m x y)
      (m x y)
      (m x y)
      (if (##< x y) (##flonum.<-fixnum x) y)
      (if (cpxnum-real? y) (##min x (cpxnum-real y)) (type-error)))

    (number-dispatch y (type-error) ; x = bignum
      (m x y)
      (m x y)
      (m x y)
      (if (##< x y) (##flonum.<-bignum x) y)
      (if (cpxnum-real? y) (##min x (cpxnum-real y)) (type-error)))

    (number-dispatch y (type-error) ; x = ratnum
      (m x y)
      (m x y)
      (m x y)
      (if (##< x y) (##flonum.<-ratnum x) y)
      (if (cpxnum-real? y) (##min x (cpxnum-real y)) (type-error)))

    (number-dispatch y (type-error) ; x = flonum
      (if (##< x y) x (##flonum.<-fixnum y))
      (if (##< x y) x (##flonum.<-bignum y))
      (if (##< x y) x (##flonum.<-ratnum y))
      (m x y)
      (if (cpxnum-real? y) (##min x (cpxnum-real y)) (type-error)))

    (if (cpxnum-real? x) ; x = cpxnum
      (number-dispatch y (type-error)
        (##min (cpxnum-real x) y)
        (##min (cpxnum-real x) y)
        (##min (cpxnum-real x) y)
        (##min (cpxnum-real x) y)
        (if (cpxnum-real? y) (##min (cpxnum-real x) (cpxnum-real y)) (type-error)))
      (type-error))))

(define-nary1 (min x y) x (##min x y) force-vars)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; +, *, -, /

(define (##+ x y)

  (define (type-error) (##trap-check-number '+ x y))

  (number-dispatch x (type-error)

    (number-dispatch y (type-error) ; x = fixnum
      (##bignum.+/fixnum-fixnum x y)
      (##bignum.+/bignum-fixnum y x)
      (##ratnum.+ (##ratnum.<-exact-int x) y)
      (##flonum.+ (##flonum.<-fixnum x) y)
      (##cpxnum.+ (##cpxnum.<-non-cpxnum x) y))

    (number-dispatch y (type-error) ; x = bignum
      (##bignum.+/bignum-fixnum x y)
      (##bignum.+ x y)
      (##ratnum.+ (##ratnum.<-exact-int x) y)
      (##flonum.+ (##flonum.<-bignum x) y)
      (##cpxnum.+ (##cpxnum.<-non-cpxnum x) y))

    (number-dispatch y (type-error) ; x = ratnum
      (##ratnum.+ x (##ratnum.<-exact-int y))
      (##ratnum.+ x (##ratnum.<-exact-int y))
      (##ratnum.+ x y)
      (##flonum.+ (##flonum.<-ratnum x) y)
      (##cpxnum.+ (##cpxnum.<-non-cpxnum x) y))

    (number-dispatch y (type-error) ; x = flonum
      (##flonum.+ x (##flonum.<-fixnum y))
      (##flonum.+ x (##flonum.<-bignum y))
      (##flonum.+ x (##flonum.<-ratnum y))
      (##flonum.+ x y)
      (##cpxnum.+ (##cpxnum.<-non-cpxnum x) y))

    (number-dispatch y (type-error) ; x = cpxnum
      (##cpxnum.+ x (##cpxnum.<-non-cpxnum y))
      (##cpxnum.+ x (##cpxnum.<-non-cpxnum y))
      (##cpxnum.+ x (##cpxnum.<-non-cpxnum y))
      (##cpxnum.+ x (##cpxnum.<-non-cpxnum y))
      (##cpxnum.+ x y))))

(define-nary0 (+ x y) 0 x (##+ x y) force-vars)

(define (##* x y)

  (define (type-error) (##trap-check-number '* x y))

  (number-dispatch x (type-error)

    (number-dispatch y (type-error) ; x = fixnum
      (##bignum.*/fixnum-fixnum x y)
      (##bignum.*/bignum-fixnum y x)
      (##ratnum.* (##ratnum.<-exact-int x) y)
      (##flonum.* (##flonum.<-fixnum x) y)
      (##cpxnum.* (##cpxnum.<-non-cpxnum x) y))

    (number-dispatch y (type-error) ; x = bignum
      (##bignum.*/bignum-fixnum x y)
      (##bignum.* x y)
      (##ratnum.* (##ratnum.<-exact-int x) y)
      (##flonum.* (##flonum.<-bignum x) y)
      (##cpxnum.* (##cpxnum.<-non-cpxnum x) y))

    (number-dispatch y (type-error) ; x = ratnum
      (##ratnum.* x (##ratnum.<-exact-int y))
      (##ratnum.* x (##ratnum.<-exact-int y))
      (##ratnum.* x y)
      (##flonum.* (##flonum.<-ratnum x) y)
      (##cpxnum.* (##cpxnum.<-non-cpxnum x) y))

    (number-dispatch y (type-error) ; x = flonum
      (##flonum.* x (##flonum.<-fixnum y))
      (##flonum.* x (##flonum.<-bignum y))
      (##flonum.* x (##flonum.<-ratnum y))
      (##flonum.* x y)
      (##cpxnum.* (##cpxnum.<-non-cpxnum x) y))

    (number-dispatch y (type-error) ; x = cpxnum
      (##cpxnum.* x (##cpxnum.<-non-cpxnum y))
      (##cpxnum.* x (##cpxnum.<-non-cpxnum y))
      (##cpxnum.* x (##cpxnum.<-non-cpxnum y))
      (##cpxnum.* x (##cpxnum.<-non-cpxnum y))
      (##cpxnum.* x y))))

(define-nary0 (* x y) 1 x (##* x y) force-vars)

(define (##- x y)

  (define (type-error) (##trap-check-number '- x y))

  (number-dispatch x (type-error)

    (number-dispatch y (type-error) ; x = fixnum
      (##bignum.-/fixnum-fixnum x y)
      (##bignum.-/fixnum-bignum x y)
      (##ratnum.- (##ratnum.<-exact-int x) y)
      (##flonum.- (##flonum.<-fixnum x) y)
      (##cpxnum.- (##cpxnum.<-non-cpxnum x) y))

    (number-dispatch y (type-error) ; x = bignum
      (##bignum.-/bignum-fixnum x y)
      (##bignum.- x y)
      (##ratnum.- (##ratnum.<-exact-int x) y)
      (##flonum.- (##flonum.<-bignum x) y)
      (##cpxnum.- (##cpxnum.<-non-cpxnum x) y))

    (number-dispatch y (type-error) ; x = ratnum
      (##ratnum.- x (##ratnum.<-exact-int y))
      (##ratnum.- x (##ratnum.<-exact-int y))
      (##ratnum.- x y)
      (##flonum.- (##flonum.<-ratnum x) y)
      (##cpxnum.- (##cpxnum.<-non-cpxnum x) y))

    (number-dispatch y (type-error) ; x = flonum
      (##flonum.- x (##flonum.<-fixnum y))
      (##flonum.- x (##flonum.<-bignum y))
      (##flonum.- x (##flonum.<-ratnum y))
      (##flonum.- x y)
      (##cpxnum.- (##cpxnum.<-non-cpxnum x) y))

    (number-dispatch y (type-error) ; x = cpxnum
      (##cpxnum.- x (##cpxnum.<-non-cpxnum y))
      (##cpxnum.- x (##cpxnum.<-non-cpxnum y))
      (##cpxnum.- x (##cpxnum.<-non-cpxnum y))
      (##cpxnum.- x (##cpxnum.<-non-cpxnum y))
      (##cpxnum.- x y))))

(define-nary1 (- x y) (##- 0 x) (##- x y) force-vars)

(define (##/ x y)

  (define (type-error) (##trap-check-number '/ x y))

  (define (divide-by-zero-error) (##trap-divide-by-zero '/ x y))

  (number-dispatch y (type-error)

    (if (##fixnum.zero? y) ; y = fixnum
      (divide-by-zero-error)
      (number-dispatch x (type-error)
        (##ratnum./ (##ratnum.<-exact-int x) (##ratnum.<-exact-int y))
        (##ratnum./ (##ratnum.<-exact-int x) (##ratnum.<-exact-int y))
        (##ratnum./ x (##ratnum.<-exact-int y))
        (##flonum./ x (##flonum.<-fixnum y))
        (##cpxnum./ x (##cpxnum.<-non-cpxnum y))))

    (number-dispatch x (type-error) ; y = bignum
      (##ratnum./ (##ratnum.<-exact-int x) (##ratnum.<-exact-int y))
      (##ratnum./ (##ratnum.<-exact-int x) (##ratnum.<-exact-int y))
      (##ratnum./ x (##ratnum.<-exact-int y))
      (##flonum./ x (##flonum.<-bignum y))
      (##cpxnum./ x (##cpxnum.<-non-cpxnum y)))

    (number-dispatch x (type-error) ; y = ratnum
      (##ratnum./ (##ratnum.<-exact-int x) y)
      (##ratnum./ (##ratnum.<-exact-int x) y)
      (##ratnum./ x y)
      (##flonum./ x (##flonum.<-ratnum y))
      (##cpxnum./ x (##cpxnum.<-non-cpxnum y)))

    (if (##flonum.zero? y) ; y = flonum
      (divide-by-zero-error)
      (number-dispatch x (type-error)
        (##flonum./ (##flonum.<-fixnum x) y)
        (##flonum./ (##flonum.<-bignum x) y)
        (##flonum./ (##flonum.<-ratnum x) y)
        (##flonum./ x y)
        (##cpxnum./ x (##cpxnum.<-non-cpxnum y))))

    (let ((imag (cpxnum-imag y))) ; y = cpxnum
      (if (and (flonum-zero? imag)
               (let ((real (cpxnum-real y)))
                 (non-cpxnum-zero? real)))
        (divide-by-zero-error)
        (number-dispatch x (type-error)
          (##cpxnum./ (##cpxnum.<-non-cpxnum x) y)
          (##cpxnum./ (##cpxnum.<-non-cpxnum x) y)
          (##cpxnum./ (##cpxnum.<-non-cpxnum x) y)
          (##cpxnum./ (##cpxnum.<-non-cpxnum x) y)
          (##cpxnum./ x y))))))

(define-nary1 (/ x y) (##/ 1 x) (##/ x y) force-vars)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; abs

(define (##abs x)

  (define (type-error) (##trap-check-real 'abs x))

  (number-dispatch x (type-error)
    (if (##fixnum.negative? x) (##bignum.-/fixnum-fixnum 0 x) x)
    (if (bignum-negative? x) (##bignum.-/fixnum-bignum 0 x) x)
    (if (##negative? (ratnum-numerator x))
      (ratnum-make (##- 0 (ratnum-numerator x)) (ratnum-denominator x))
      x)
    (##flonum.abs x)
    (if (cpxnum-real? x) (##abs (cpxnum-real x)) (type-error))))

(define (abs x) (force-vars (x) (##abs x)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; quotient, remainder, modulo

(define (##quotient x y)

  (define (type-error) (##trap-check-integer 'quotient x y))

  (define (divide-by-zero-error) (##trap-divide-by-zero 'quotient x y))

  (define (inexact-quotient)
    (##exact->inexact (##quotient (##inexact->exact x) (##inexact->exact y))))

  (number-dispatch y (type-error)

    (if (##fixnum.zero? y) ; y = fixnum
      (divide-by-zero-error)
      (number-dispatch x (type-error)
        (if (##fixnum.= y -1)
          (##bignum.-/fixnum-fixnum 0 x)
          (##fixnum.quotient x y))
        (##bignum.quotient/bignum-fixnum x y)
        (type-error)
        (if (flonum-int? x) (inexact-quotient) (type-error))
        (if (cpxnum-int? x) (inexact-quotient) (type-error))))

    (number-dispatch x (type-error) ; y = bignum
      (##bignum.quotient/fixnum-bignum x y)
      (##bignum.quotient x y)
      (type-error)
      (if (flonum-int? x) (inexact-quotient) (type-error))
      (if (cpxnum-int? x) (inexact-quotient) (type-error)))

    (type-error) ; y = ratnum

    (if (flonum-int? y) ; y = flonum
      (number-dispatch x (type-error)
        (inexact-quotient)
        (inexact-quotient)
        (type-error)
        (if (flonum-int? x) (inexact-quotient) (type-error))
        (if (cpxnum-int? x) (inexact-quotient) (type-error)))
      (type-error))

    (if (cpxnum-int? y) ; y = cpxnum
      (number-dispatch x (type-error)
        (inexact-quotient)
        (inexact-quotient)
        (type-error)
        (if (flonum-int? x) (inexact-quotient) (type-error))
        (if (cpxnum-int? x) (inexact-quotient) (type-error)))
      (type-error))))

(define (quotient x y) (force-vars (x y) (##quotient x y)))

(define (##remainder x y)

  (define (type-error) (##trap-check-integer 'remainder x y))

  (define (divide-by-zero-error) (##trap-divide-by-zero 'remainder x y))

  (define (inexact-remainder)
    (##exact->inexact (##remainder (##inexact->exact x) (##inexact->exact y))))

  (number-dispatch y (type-error)

    (if (##fixnum.zero? y) ; y = fixnum
      (divide-by-zero-error)
      (number-dispatch x (type-error)
        (##fixnum.remainder x y)
        (##bignum.remainder/bignum-fixnum x y)
        (type-error)
        (if (flonum-int? x) (inexact-remainder) (type-error))
        (if (cpxnum-int? x) (inexact-remainder) (type-error))))

    (number-dispatch x (type-error) ; y = bignum
      (##bignum.remainder/fixnum-bignum x y)
      (##bignum.remainder x y)
      (type-error)
      (if (flonum-int? x) (inexact-remainder) (type-error))
      (if (cpxnum-int? x) (inexact-remainder) (type-error)))

    (type-error) ; y = ratnum

    (if (flonum-int? y) ; y = flonum
      (number-dispatch x (type-error)
        (inexact-remainder)
        (inexact-remainder)
        (type-error)
        (if (flonum-int? x) (inexact-remainder) (type-error))
        (if (cpxnum-int? x) (inexact-remainder) (type-error)))
      (type-error))

    (if (cpxnum-int? y) ; y = cpxnum
      (number-dispatch x (type-error)
        (inexact-remainder)
        (inexact-remainder)
        (type-error)
        (if (flonum-int? x) (inexact-remainder) (type-error))
        (if (cpxnum-int? x) (inexact-remainder) (type-error)))
      (type-error))))

(define (remainder x y) (force-vars (x y) (##remainder x y)))

(define (##modulo x y)

  (define (divide-by-zero-error) (##trap-divide-by-zero 'modulo x y))

  (define (type-error) (##trap-check-integer 'modulo x y))

  (define (inexact-modulo)
    (##exact->inexact (##modulo (##inexact->exact x) (##inexact->exact y))))

  (number-dispatch y (type-error)

    (if (##fixnum.zero? y) ; y = fixnum
      (divide-by-zero-error)
      (number-dispatch x (type-error)
        (##fixnum.modulo x y)
        (##bignum.modulo/bignum-fixnum x y)
        (type-error)
        (if (flonum-int? x) (inexact-modulo) (type-error))
        (if (cpxnum-int? x) (inexact-modulo) (type-error))))

    (number-dispatch x (type-error) ; y = bignum
      (##bignum.modulo/fixnum-bignum x y)
      (##bignum.modulo x y)
      (type-error)
      (if (flonum-int? x) (inexact-modulo) (type-error))
      (if (cpxnum-int? x) (inexact-modulo) (type-error)))

    (type-error) ; y = ratnum

    (if (flonum-int? y) ; y = flonum
      (number-dispatch x (type-error)
        (inexact-modulo)
        (inexact-modulo)
        (type-error)
        (if (flonum-int? x) (inexact-modulo) (type-error))
        (if (cpxnum-int? x) (inexact-modulo) (type-error)))
      (type-error))

    (if (cpxnum-int? y) ; y = cpxnum
      (number-dispatch x (type-error)
        (inexact-modulo)
        (inexact-modulo)
        (type-error)
        (if (flonum-int? x) (inexact-modulo) (type-error))
        (if (cpxnum-int? x) (inexact-modulo) (type-error)))
      (type-error))))

(define (modulo x y) (force-vars (x y) (##modulo x y)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; gcd, lcm

(define (##gcd x y)

  (define (type-error) (##trap-check-integer 'gcd x y))

  (define (exact-gcd x y)
    (let loop ((x (##abs x)) (y (##abs y)))
      (if (##eq? y 0) x (loop y (##remainder x y)))))

  (if (and (##integer? x) (##integer? y))
    (if (and (##exact? x) (##exact? y))
      (exact-gcd x y)
      (##exact->inexact (exact-gcd (##inexact->exact x) (##inexact->exact y))))
    (type-error)))

(define-nary0 (gcd x y) 0 x (##gcd x y) force-vars)

(define (##lcm x y)

  (define (type-error) (##trap-check-integer 'lcm x y))

  (define (exact-gcd x y)
    (let loop ((x (##abs x)) (y (##abs y)))
      (if (##eq? y 0) x (loop y (##remainder x y)))))

  (define (exact-lcm x y)
    (if (or (##eq? x 0) (##eq? y 0))
      0
      (##quotient (##abs (##* x y)) (exact-gcd x y))))

  (if (and (##integer? x) (##integer? y))
    (if (and (##exact? x) (##exact? y))
      (exact-lcm x y)
      (##exact->inexact (exact-lcm (##inexact->exact x) (##inexact->exact y))))
    (type-error)))

(define-nary0 (lcm x y) 1 x (##lcm x y) force-vars)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; numerator, denominator

(define (##numerator x)

  (define (type-error) (##trap-check-rational 'numerator x))

  (number-dispatch x (type-error)
    x
    x
    (ratnum-numerator x)
    (##numerator (##flonum.inexact->exact x))
    (if (cpxnum-real? x) (##numerator (cpxnum-real x)) (type-error))))

(define (numerator x) (force-vars (x) (##numerator x)))

(define (##denominator x)

  (define (type-error) (##trap-check-rational 'denominator x))

  (number-dispatch x (type-error)
    1
    1
    (ratnum-denominator x)
    (##denominator (##flonum.inexact->exact x))
    (if (cpxnum-real? x) (##denominator (cpxnum-real x)) (type-error))))

(define (denominator x) (force-vars (x) (##denominator x)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; floor, ceiling, truncate, round

(define (##floor x)

  (define (type-error) (##trap-check-real 'floor x))

  (number-dispatch x (type-error)
    x
    x
    (##ratnum.floor x)
    (##flonum.floor x)
    (if (cpxnum-real? x) (##floor (cpxnum-real x)) (type-error))))

(define (floor x) (force-vars (x) (##floor x)))

(define (##ceiling x)

  (define (type-error) (##trap-check-real 'ceiling x))

  (number-dispatch x (type-error)
    x
    x
    (##ratnum.ceiling x)
    (##flonum.ceiling x)
    (if (cpxnum-real? x) (##ceiling (cpxnum-real x)) (type-error))))

(define (ceiling x) (force-vars (x) (##ceiling x)))

(define (##truncate x)

  (define (type-error) (##trap-check-real 'truncate x))

  (number-dispatch x (type-error)
    x
    x
    (##ratnum.truncate x)
    (##flonum.truncate x)
    (if (cpxnum-real? x) (##truncate (cpxnum-real x)) (type-error))))

(define (truncate x) (force-vars (x) (##truncate x)))

(define (##round x)

  (define (type-error) (##trap-check-real 'round x))

  (number-dispatch x (type-error)
    x
    x
    (##ratnum.round x)
    (##flonum.round x)
    (if (cpxnum-real? x) (##round (cpxnum-real x)) (type-error))))

(define (round x) (force-vars (x) (##round x)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; rationalize

(define (##rationalize x y)

  (define (type-error) (##trap-check-real 'rationalize x y))

  (define (simplest-rational1 x y)
    (cond ((##< y x)
           (simplest-rational2 y x))
          ((##not (##< x y))
           x)
          ((##positive? x)
           (simplest-rational2 x y))
          ((##negative? y)
           (##- 0 (simplest-rational2 (##- 0 y) (##- 0 x))))
          (else
           0)))

  (define (simplest-rational2 x y)
    (let ((fx (##floor x))
          (fy (##floor y)))
      (cond ((##not (##< fx x))
             fx)
            ((##= fx fy)
             (##+ fx
                  (##/ 1
                       (simplest-rational2
                         (##/ 1 (##- y fy))
                         (##/ 1 (##- x fx))))))
            (else
             (##+ fx 1)))))

  (if (and (##real? x) (##real? y))
    (simplest-rational1 (##- x y) (##+ x y))
    (type-error)))

(define (rationalize x y) (force-vars (x y) (##rationalize x y)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; trigonometry and complex numbers

(define (##exp x)

  (define (type-error) (##trap-check-number 'exp x))

  (number-dispatch x (type-error)
    (if (##fixnum.zero? x) 1 (##flonum.exp (##flonum.<-fixnum x)))
    (##flonum.exp (##flonum.<-bignum x))
    (##flonum.exp (##flonum.<-ratnum x))
    (##flonum.exp x)
    (##make-polar (##exp (cpxnum-real x)) (cpxnum-imag x))))

(define (exp x) (force-vars (x) (##exp x)))

(define (##log x)

  (define (type-error) (##trap-check-number 'log x))

  (define (range-error) (##trap-check-range 'log x))

  (define (negative-log x)
    (cpxnum-make (##log (##- 0 x)) (inexact-+pi)))

  (number-dispatch x (type-error)
    (if (##fixnum.positive? x)
      (if (##eq? x 1) 0 (##flonum.log (##flonum.<-fixnum x)))
      (if (##fixnum.zero? x) (range-error) (negative-log x)))
    (if (bignum-positive? x)
      (##flonum.log (##flonum.<-bignum x))
      (negative-log x))
    (if (##positive? (ratnum-numerator x))
      (##flonum.log (##flonum.<-ratnum x))
      (negative-log x))
    (if (##flonum.positive? x)
      (##flonum.log x)
      (if (##flonum.zero? x) (range-error) (negative-log x)))
    (##make-rectangular (##log (##magnitude x)) (##angle x))))

(define (log x) (force-vars (x) (##log x)))

(define (##sin x)

  (define (type-error) (##trap-check-number 'sin x))

  (number-dispatch x (type-error)
    (if (##fixnum.zero? x) 0 (##flonum.sin (##flonum.<-fixnum x)))
    (##flonum.sin (##flonum.<-bignum x))
    (##flonum.sin (##flonum.<-ratnum x))
    (##flonum.sin x)
    (##/ (##- (##exp (cpxnum-make (##- 0 (cpxnum-imag x)) (cpxnum-real x)))
              (##exp (cpxnum-make (cpxnum-imag x) (##- 0 (cpxnum-real x)))))
         (cpxnum-+2i))))

(define (sin x) (force-vars (x) (##sin x)))

(define (##cos x)

  (define (type-error) (##trap-check-number 'cos x))

  (number-dispatch x (type-error)
    (if (##fixnum.zero? x) 1 (##flonum.cos (##flonum.<-fixnum x)))
    (##flonum.cos (##flonum.<-bignum x))
    (##flonum.cos (##flonum.<-ratnum x))
    (##flonum.cos x)
    (##/ (##+ (##exp (cpxnum-make (##- 0 (cpxnum-imag x)) (cpxnum-real x)))
              (##exp (cpxnum-make (cpxnum-imag x) (##- 0 (cpxnum-real x)))))
         2)))

(define (cos x) (force-vars (x) (##cos x)))

(define (##tan x)

  (define (type-error) (##trap-check-number 'tan x))

  (number-dispatch x (type-error)
    (if (##fixnum.zero? x) 0 (##flonum.tan (##flonum.<-fixnum x)))
    (##flonum.tan (##flonum.<-bignum x))
    (##flonum.tan (##flonum.<-ratnum x))
    (##flonum.tan x)
    (let ((a (##exp (cpxnum-make (##- 0 (cpxnum-imag x)) (cpxnum-real x))))
          (b (##exp (cpxnum-make (cpxnum-imag x) (##- 0 (cpxnum-real x))))))
      (let ((c (##/ (##- a b) (##+ a b))))
        (##make-rectangular (##imag-part c) (##- 0 (##real-part c)))))))

(define (tan x) (force-vars (x) (##tan x)))

(define (##asin x)

  (define (type-error) (##trap-check-number 'asin x))

  (define (safe-case x)
    (##* (cpxnum--i)
         (##log (##+ (##* (cpxnum-+i) x)
                     (##sqrt (##- 1 (##* x x)))))))

  (define (unsafe-case x)
    (##- 0 (safe-case (##- 0 x))))

  (define (real-case x)
    (cond ((##< x -1)
           (unsafe-case x))
          ((##< 1 x)
           (safe-case x))
          (else
           (##flonum.asin (##exact->inexact x)))))

  (number-dispatch x (type-error)
    (if (##fixnum.zero? x) 0 (real-case x))
    (real-case x)
    (real-case x)
    (real-case x)
    (let ((imag (cpxnum-imag x)))
      (if (or (##positive? imag)
              (and (flonum-zero? imag) (##negative? (cpxnum-real x))))
        (unsafe-case x)
        (safe-case x)))))

(define (asin x) (force-vars (x) (##asin x)))

(define (##acos x)

  (define (type-error) (##trap-check-number 'acos x))

  (define (complex-case x)
    (##* (cpxnum--i)
         (##log (##+ x
                     (##* (cpxnum-+i) (##sqrt (##- 1 (##* x x))))))))

  (define (real-case x)
    (if (or (##< x -1) (##< 1 x))
      (complex-case x)
      (##flonum.acos (##exact->inexact x))))

  (number-dispatch x (type-error)
    (if (##fixnum.zero? x) (inexact-+pi/2) (real-case x))
    (real-case x)
    (real-case x)
    (real-case x)
    (complex-case x)))

(define (acos x) (force-vars (x) (##acos x)))

(define (##atan x)

  (define (type-error) (##trap-check-number 'atan x))

  (number-dispatch x (type-error)
    (if (##fixnum.zero? x) 0 (##flonum.atan (##flonum.<-fixnum x)))
    (##flonum.atan (##flonum.<-bignum x))
    (##flonum.atan (##flonum.<-ratnum x))
    (##flonum.atan x)
    (let ((a (cpxnum-make (##- 0 (cpxnum-imag x)) (cpxnum-real x))))
      (##/ (##- (##log (##+ a 1)) (##log (##- 1 a)))
           (cpxnum-+2i)))))

(define (##atan2 y x)

  (define (type-error) (##trap-check-real 'atan y x))

  (if (and (##real? x) (##real? y))
    (let ((x (##exact->inexact x)) (y (##exact->inexact y)))
      (cond ((##flonum.positive? x)
             (##flonum.atan (##flonum./ y x)))
            ((##flonum.negative? y)
             (if (##flonum.zero? x)
               (inexact--pi/2)
               (##flonum.+ (##flonum.atan (##flonum./ y x)) (inexact--pi))))
            (else
             (if (##flonum.zero? x)
               (inexact-+pi/2)
               (##flonum.+ (##flonum.atan (##flonum./ y x)) (inexact-+pi))))))
    (type-error)))

(define (atan x #!optional (y (absent-obj)))
  (force-vars (x)
    (if (##eq? y (absent-obj))
      (##atan x)
      (force-vars (y)
        (##atan2 x y)))))

(define (##sqrt x)

  (define (type-error) (##trap-check-number 'sqrt x))

  (define (exact-int-sqrt x)
    (cond ((##eq? x 0)
           0)
          ((##negative? x)
           (cpxnum-make 0 (exact-int-sqrt (##- 0 x))))
          (else
           (let ((y (##exact-int.root x 2)))
             (if (##= x (##* y y))
               y
               (##flonum.sqrt (##exact->inexact x)))))))

  (number-dispatch x (type-error)
    (exact-int-sqrt x)
    (exact-int-sqrt x)
    (##/ (exact-int-sqrt (ratnum-numerator x))
         (exact-int-sqrt (ratnum-denominator x)))
    (if (##flonum.negative? x)
      (cpxnum-make 0 (##flonum.sqrt (##flonum.- (inexact-0) x)))
      (##flonum.sqrt x))
    (##make-polar (##sqrt (##magnitude x)) (##/ (##angle x) 2))))

(define (sqrt x) (force-vars (x) (##sqrt x)))

(define (##expt x y)

  (define (type-error) (##trap-check-number 'expt x y))

  (define (general-expt x y)
    (##exp (##* (##log x) y)))

  (define (exact-int-expt x y)
    (cond ((##eq? y 0)
           1)
          ((or (##zero? x) (##= x 1))
           x)
          (else
           (let loop ((x x) (y y) (result 1))
             (if (##eq? y 1)
               (##* x result)
               (loop (##* x x)
                     (##quotient y 2)
                     (if (##odd? y) (##* x result) result)))))))

  (if (##complex? x)
    (cond ((exact-int? y)
           (if (##negative? y)
             (##/ 1 (exact-int-expt x (##- 0 y)))
             (exact-int-expt x y)))
          ((##complex? y)
           (cond ((##zero? y) (inexact-+1))
                 ((##zero? x) (if (##eq? x 0) 0 (inexact-0)))
                 (else        (general-expt x y))))
          (else
           (type-error)))
    (type-error)))

(define (expt x y) (force-vars (x y) (##expt x y)))

(define (##make-rectangular x y)

  (define (type-error) (##trap-check-real 'make-rectangular x y))

  (if (and (##real? x) (##real? y))
    (if (##eq? y 0)
      x
      (cpxnum-make (##real-part x) (##real-part y)))
    (type-error)))

(define (make-rectangular x y)
  (force-vars (x y) (##make-rectangular x y)))

(define (##make-polar x y)

  (define (type-error) (##trap-check-real 'make-polar x y))

  (if (and (##real? x) (##real? y))
    (let ((x* (##real-part x)) (y* (##real-part y)))
      (##make-rectangular (##* x* (##cos y*)) (##* x* (##sin y*))))
    (type-error)))

(define (make-polar x y) (force-vars (x y) (##make-polar x y)))

(define (##real-part x)

  (define (type-error) (##trap-check-number 'real-part x))

  (number-dispatch x (type-error)
    x x x x (cpxnum-real x)))

(define (real-part x) (force-vars (x) (##real-part x)))

(define (##imag-part x)

  (define (type-error) (##trap-check-number 'imag-part x))

  (number-dispatch x (type-error)
    0 0 0 0 (cpxnum-imag x)))

(define (imag-part x) (force-vars (x) (##imag-part x)))

(define (##magnitude x)

  (define (type-error) (##trap-check-number 'magnitude x))

  (number-dispatch x (type-error)
    (if (##fixnum.negative? x) (##bignum.-/fixnum-fixnum 0 x) x)
    (if (bignum-negative? x) (##bignum.-/fixnum-bignum 0 x) x)
    (if (##negative? (ratnum-numerator x))
      (ratnum-make (##- 0 (ratnum-numerator x)) (ratnum-denominator x))
      x)
    (##flonum.abs x)
    (let ((r (##abs (##real-part x))) (i (##abs (##imag-part x))))
      (define (complex-magn a b)
        (if (##zero? b)
          b
          (let ((c (##/ a b)))
            (##* b (##sqrt (##+ (##* c c) 1))))))
      (if (##< r i) (complex-magn r i) (complex-magn i r)))))

(define (magnitude x) (force-vars (x) (##magnitude x)))

(define (##angle x)

  (define (type-error) (##trap-check-number 'angle x))

  (number-dispatch x (type-error)
    (if (##fixnum.negative? x) (inexact-+pi) 0)
    (if (bignum-negative? x) (inexact-+pi) 0)
    (if (##negative? (ratnum-numerator x)) (inexact-+pi) 0)
    (if (##flonum.negative? x) (inexact-+pi) (inexact-0))
    (if (##zero? x)
      (inexact-0)
      (##atan2 (cpxnum-imag x) (cpxnum-real x)))))

(define (angle x) (force-vars (x) (##angle x)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; exact->inexact, inexact->exact

(define (##exact->inexact x)

  (define (type-error) (##trap-check-number 'exact->inexact x))

  (number-dispatch x (type-error)
    (##flonum.<-fixnum x)
    (##flonum.<-bignum x)
    (##flonum.<-ratnum x)
    x
    (##make-rectangular (##exact->inexact (cpxnum-real x))
                        (##exact->inexact (cpxnum-imag x)))))

(define (exact->inexact x)
  (force-vars (x) (##exact->inexact x)))

(define (##inexact->exact x)

  (define (type-error) (##trap-check-number 'inexact->exact x))

  (number-dispatch x (type-error)
    x
    x
    x
    (##flonum.inexact->exact x)
    (##make-rectangular (##inexact->exact (cpxnum-real x))
                        (##inexact->exact (cpxnum-imag x)))))

(define (inexact->exact x)
  (force-vars (x) (##inexact->exact x)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; number->string, string->number


(define (##number->string x rad)

  (define (range-error) (##trap-check-range 'number->string x rad))

  (define (type-error1) (##trap-check-number 'number->string x rad))

  (define (type-error2) (##trap-check-exact-int 'number->string x rad))

  (if (exact-int? rad)
    (if (or (##eq? rad 2)
            (##eq? rad 8)
            (##eq? rad 10)
            (##eq? rad 16))
      (number-dispatch x (type-error1)
        (##fixnum.number->string x rad)
        (##bignum.number->string x rad)
        (##ratnum.number->string x rad)
        (##flonum.number->string x rad)
        (##cpxnum.number->string x rad))
      (range-error))
    (type-error2)))

(define (##exact-int.number->string x rad)
  (if (##fixnum? x)
    (##fixnum.number->string x rad)
    (##bignum.number->string x rad)))

(define ##digits "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ")

(define (##fixnum.number->string n rad)

  (define (loop k n i)
    (let ((x (##fixnum.quotient n rad)))
      (let ((s (if (##fixnum.zero? x)
                 (##make-string (##fixnum.+ i k) #\space)
                 (loop k x (##fixnum.+ i 1)))))
        (##string-set! s
                       (##fixnum.- (##string-length s) i)
                       (##string-ref ##digits
                                     (##fixnum.- 0 (##fixnum.remainder n rad)))))))

  (if (##fixnum.< n 0)
    (##string-set! (loop 1 n 1) 0 #\-)
    (loop 0 (##fixnum.- 0 n) 1)))

(define (##bignum.number->string n rad)

  (define (bignum->string n rad r r-log-rad radix-log-r-num)
    (let ((len (##fixnum.* (##fixnum.quotient
                             (##fixnum.+
                               (##fixnum.* (##fixnum.- (bignum-length n) 1)
                                           radix-log-r-num)
                               (##fixnum.- (radix-log-den) 1))
                             (radix-log-den))
                           r-log-rad)))
      (let ((n (##bignum.copy n))
            (s (##make-string (##fixnum.+ len 1) #\space)))

        (define (put-digits k i)
          (let loop1 ((k k) (i i) (j r-log-rad) (last-non-zero i))
            (if (##fixnum.< 0 j)
              (let ((d (##fixnum.remainder k rad)))
                (##string-set! s i (##string-ref ##digits d))
                (loop1 (##fixnum.quotient k rad)
                       (##fixnum.- i 1)
                       (##fixnum.- j 1)
                       (if (##fixnum.zero? d) last-non-zero i)))
              last-non-zero)))

        (define (move-digits i j)
          (let loop2 ((i i) (j j))
            (if (##fixnum.< len i)
              (##string-shrink! s j)
              (begin
                (##string-set! s j (##string-ref s i))
                (loop2 (##fixnum.+ i 1) (##fixnum.+ j 1))))))

        (let loop3 ((i len))

          (let ((k
                 ; k = next digit in base `r'
                 ; use simple algo for dividing in place by `r'
                 ; (which is known to be less than or equal to radix)

                 (let loop4 ((j (##fixnum.- (bignum-length n) 1)) (k 0))
                   (if (##fixnum.< 0 j)
                     (let ((x (##fixnum.+ (##fixnum.* k (radix))
                                          (bignum-digit-ref n j))))
                       (bignum-digit-set! n j (##fixnum.quotient x r))
                       (loop4 (##fixnum.- j 1) (##fixnum.remainder x r)))
                     k))))

            (let ((last-non-zero (put-digits k i)))
              (##bignum.remove-leading-0s! n)
              (if (##not (bignum-zero? n))
                (loop3 (##fixnum.- i r-log-rad))
                (if (bignum-negative? n)
                  (begin
                    (##string-set! s 0 #\-)
                    (move-digits last-non-zero 1))
                  (move-digits last-non-zero 0)))))))))

  (cond ((##fixnum.= rad 2)
         (bignum->string n rad (r.2) (r-log-rad.2) (radix-log-r-num.2)))
        ((##fixnum.= rad 8)
         (bignum->string n rad (r.8) (r-log-rad.8) (radix-log-r-num.8)))
        ((##fixnum.= rad 10)
         (bignum->string n rad (r.10) (r-log-rad.10) (radix-log-r-num.10)))
        (else
         (bignum->string n rad (r.16) (r-log-rad.16) (radix-log-r-num.16)))))

(define (##ratnum.number->string x rad)
  (##string-append (##exact-int.number->string (ratnum-numerator x) rad)
                   "/"
                   (##exact-int.number->string (ratnum-denominator x) rad)))

(define (##flonum.number->string x rad)

  (define (num->str x rad)
    (let ((z (##flonum.->exact-exponential-format x)))
      (##flonum.printout (##car z) (##cdr z) rad)))

  (cond ((##flonum.zero? x)
         "0.")
        ((##flonum.negative? x)
         (##string-append "-" (num->str (##flonum.abs x) rad)))
        (else
         (num->str x rad))))

(##define-macro (two) 2)
(##define-macro (ten) 10)
(##define-macro (ten-minus-1) 9)

(define (##flonum.printout m e rad) ; rad parameter is currently ignored

  (define (done h k d)
    (let ((str (##exact-int.number->string d (ten))))
      (cond ((and (##fixnum.< h -1)
                  (or ; (##fixnum.< -5 h)
                      (##fixnum.< (##fixnum.- 0 (flonum-max-digits)) k)))
             (##string-append "."
                              (##make-string (##fixnum.- -1 h) #\0)
                              str))
            ((and (##fixnum.< 0 k)
                  (or ; (##fixnum.< k 3)
                      (##fixnum.< h (flonum-max-digits))))
             (##string-append str
                              (##make-string k #\0)
                              "."))
            ((and (##fixnum.< -2 h) (##fixnum.< k 1))
             (let ((n (##fixnum.+ h 1)))
               (##string-append (##substring str 0 n)
                                "."
                                (##substring str n (##string-length str)))))
            (else
             (##string-append (##substring str 0 1)
                              "."
                              (##substring str 1 (##string-length str))
                              "e"
                              (##exact-int.number->string h (ten)))))))

  (define (fixup-loop1 k r s ceiling-s-div-ten m- m+)
    (if (##< r ceiling-s-div-ten)
      (fixup-loop1 (##fixnum.- k 1)
                   (##* r (ten))
                   s
                   ceiling-s-div-ten
                   (##* m- (ten))
                   (##* m+ (ten)))
      (let fixup-loop2 ((k k) (r r) (s s) (m- m-) (m+ m+))
        (if (##not (##< (##+ (##* r 2) m+) (##* s 2)))
          (fixup-loop2 (##fixnum.+ k 1) r (##* s (ten)) m- m+)
          (let ((h (##fixnum.- k 1)))
            (let ((ur (##exact-int.div (##* r (ten)) s)))
              (let loop ((k (##fixnum.- k 1))
                         (u (##car ur))
                         (r (##cdr ur))
                         (m- (##* m- (ten)))
                         (m+ (##* m+ (ten)))
                         (d 0))
                (let ((r*2 (##* r 2)) (s*2 (##* s 2)))
                  (cond ((##< r*2 m-)
                         (if (##< (##- s*2 m+) r*2)
                           (if (##not (##< s r*2))
                             (done h k (##+ d u))
                             (done h k (##+ d (##fixnum.+ u 1))))
                           (done h k (##+ d u))))
                        ((##< (##- s*2 m+) r*2)
                         (done h k (##+ d (##fixnum.+ u 1))))
                        (else
                         (let ((ur (##exact-int.div (##* r (ten)) s)))
                           (loop (##fixnum.- k 1)
                                 (##car ur)
                                 (##cdr ur)
                                 (##* m- (ten))
                                 (##* m+ (ten))
                                 (##* (##+ d u) (ten))))))))))))))

  (define (fixup r s m-)
    (if (##= m (flonum-+m-min))
      (let ((r* (##* r (two)))
            (s* (##* s (two)))
            (m+ (##* m- (two))))
        (fixup-loop1 0 r* s* (##quotient (##+ s* (ten-minus-1)) (ten)) m- m+))
      (fixup-loop1 0 r s (##quotient (##+ s (ten-minus-1)) (ten)) m- m-)))

  (if (##fixnum.negative? e)
    (fixup m (##expt (two) (##fixnum.- 0 e)) 1)
    (let ((two-to-the-e (##expt (two) e)))
      (fixup (##* m two-to-the-e) 1 two-to-the-e))))

(define (##cpxnum.number->string x rad)
  (let* ((real (cpxnum-real x))
         (real-str (if (##eq? real 0) "" (##number->string real rad))))
    (let ((imag (cpxnum-imag x)))
      (cond ((##eq? imag 1)
             (##string-append real-str "+i"))
            ((##eq? imag -1)
             (##string-append real-str "-i"))
            ((##negative? imag)
             (##string-append real-str (##number->string imag rad) "i"))
            (else
             (##string-append real-str "+" (##number->string imag rad) "i"))))))

(define (number->string n #!optional (r (absent-obj)))
  (force-vars (n r)
    (let ((radix (if (##eq? r (absent-obj)) 10 r)))
      (##number->string n radix))))

(define (##string->number s rad)

  (define (range-error) (##trap-check-range 'string->number s rad))

  (define (type-error) (##trap-check-exact-int 'string->number s rad))

  (define (make-real e n r p)       ; Note: this algorithm does not satisfy the
    (let ((x (##* n (##expt r p)))) ; accuracy required by the IEEE standard
      (if (##eq? e 'E) x (##exact->inexact x))))

  (define (make-rec a b)
    (##make-rectangular a b))

  (define (make-pol a b)
    (##make-polar a b))

  (define (ex e x)
    (if (##eq? e 'I) (##exact->inexact x) x))

  (define (end s i x)
    (if (##eq? i (##string-length s)) x #f))

  (define (radix-prefix s i)
    (if (##fixnum.< (##fixnum.+ i 1) (##string-length s))
      (if (##char=? (##string-ref s i) #\#)
        (let ((c (##string-ref s (##fixnum.+ i 1))))
          (cond ((or (##char=? c #\b) (##char=? c #\B))  2)
                ((or (##char=? c #\o) (##char=? c #\O))  8)
                ((or (##char=? c #\d) (##char=? c #\D)) 10)
                ((or (##char=? c #\x) (##char=? c #\X)) 16)
                (else                                   #f)))
        #f)
      #f))

  (define (exactness-prefix s i)
    (if (##fixnum.< (##fixnum.+ i 1) (##string-length s))
      (if (##char=? (##string-ref s i) #\#)
        (let ((c (##string-ref s (##fixnum.+ i 1))))
          (cond ((or (##char=? c #\i) (##char=? c #\I)) 'I)
                ((or (##char=? c #\e) (##char=? c #\E)) 'E)
                (else                                   #f)))
        #f)
      #f))

  (define (sign s i)
    (if (##fixnum.< i (##string-length s))
      (let ((c (##string-ref s i)))
        (cond ((##char=? c #\+) '+)
              ((##char=? c #\-) '-)
              (else             #f)))
      #f))

  (define (imaginary s i)
    (if (##fixnum.< i (##string-length s))
      (let ((c (##string-ref s i)))
        (or (##char=? c #\i) (##char=? c #\I)))
      #f))

  (define (polar s i)
    (if (##fixnum.< i (##string-length s))
      (##char=? (##string-ref s i) #\@)
      #f))

  (define (ratio s i)
    (if (##fixnum.< i (##string-length s))
      (##char=? (##string-ref s i) #\/)
      #f))

  (define (exponent s i)
    (if (##fixnum.< i (##string-length s))
      (let ((c (##string-ref s i)))
        (cond ((or (##char=? c #\e) (##char=? c #\E)) 'E)
              ((or (##char=? c #\s) (##char=? c #\S)) 'S)
              ((or (##char=? c #\f) (##char=? c #\F)) 'F)
              ((or (##char=? c #\d) (##char=? c #\D)) 'D)
              ((or (##char=? c #\l) (##char=? c #\L)) 'L)
              (else                                   #f)))
      #f))

  (define (digit c r)
    (let ((d (cond ((##not (or (##char<? c #\0) (##char<? #\9 c)))
                    (##fixnum.- (##fixnum.<-char c) 48))
                   ((##not (or (##char<? c #\a) (##char<? #\z c)))
                    (##fixnum.- (##fixnum.<-char c) 87))
                   ((##not (or (##char<? c #\A) (##char<? #\Z c)))
                    (##fixnum.- (##fixnum.<-char c) 55))
                   (else
                    #f))))
      (if (and d (##fixnum.< d r)) d #f)))

  (define (prefix s i r cont)
    (let ((e1 (exactness-prefix s i)))
      (if e1
        (let ((r1 (radix-prefix s (##fixnum.+ i 2))))
          (if r1
            (cont s (##fixnum.+ i 4) r1 e1)
            (cont s (##fixnum.+ i 2) r e1)))
        (let ((r2 (radix-prefix s i)))
          (if r2
            (let ((e2 (exactness-prefix s (##fixnum.+ i 2))))
              (if e2
                (cont s (##fixnum.+ i 4) r2 e2)
                (cont s (##fixnum.+ i 2) r2 #f)))
            (cont s i r #f))))))

  (define (num s i r)
    (prefix s i r complex))

  (define (complex s i r e)
    (let ((+/- (sign s i)))
      (ucomplex s (if +/- (##fixnum.+ i 1) i) r e +/-)))

  (define (ucomplex s i r e +/-)
    (if (and +/- (imaginary s i))
      (end s (##fixnum.+ i 1)
        (make-rec (ex e 0) (ex e (if (##eq? +/- '-) -1 1))))
      (ureal s i r e +/- #f
        (lambda (s i r e +/- dummy x)
          (let ((y (if (##eq? +/- '-) (##- 0 x) x)))
            (cond ((and +/- (imaginary s i))
                   (end s (##fixnum.+ i 1) (make-rec (ex e 0) y)))
                  ((polar s i)
                   (let ((+/-2 (sign s (##fixnum.+ i 1))))
                     (ureal s (##fixnum.+ i (if +/-2 2 1)) r e +/-2 y
                       (lambda (s i r e +/-2 y z)
                         (end s i
                           (make-pol y (if (##eq? +/-2 '-) (##- 0 z) z)))))))
                  (else
                   (let ((+/-2 (sign s i)))
                     (if +/-2
                       (if (imaginary s (##fixnum.+ i 1))
                         (end s (##fixnum.+ i 2)
                           (make-rec y (ex e (if (##eq? +/-2 '-) -1 1))))
                         (ureal s (##fixnum.+ i 1) r e +/-2 y
                           (lambda (s i r e +/-2 y z)
                             (and (imaginary s i)
                                  (end s (##fixnum.+ i 1)
                                    (make-rec y (if (##eq? +/-2 '-) (##- 0 z) z)))))))
                       (end s i y))))))))))

  (define (ureal s i r e +/- x cont)
    (uinteger s i r e +/- x cont (##eq? r 10)
      (lambda (s i r e +/- x cont ex? n p)
        (if p ; decimal point or exponent?
          (cont s i r e +/- x (make-real e n r p))
          (if (ratio s i)
            (uinteger s (##fixnum.+ i 1) r e +/- x cont #f
              (lambda (s i r e +/- x cont ex2? n2 p2)
                (let ((y (##/ n n2)))
                  (cont s i r e +/- x (ex (or e (if (and ex? ex2?) #f 'I)) y)))))
            (cont s i r e +/- x (ex (or e (if ex? #f 'I)) n)))))))

  (define (uinteger s i r a1 a2 a3 a4 decimal? cont)
    (let loop1 ((i i) (state 0) (n 0) (p #f))

      (define (suffix)
        (if (##eq? state 0)
          #f
          (let ((mark (exponent s i)))
            (if (and mark decimal?)
              (let ((+/- (sign s (##fixnum.+ i 1))) (p (or p 0)))
                (let loop2 ((i (##fixnum.+ i (if +/- 2 1))) (j #f))
                  (if (and (##fixnum.< i (##string-length s))
                           (digit (##string-ref s i) 10))
                    (loop2 (##fixnum.+ i 1)
                           (##+ (##* (or j 0) 10)
                                (digit (##string-ref s i) 10)))
                    (and j (cont s i r a1 a2 a3 a4 #f n
                             (##+ p (if (##eq? +/- '-) (##- 0 j) j)))))))
              (cont s i r a1 a2 a3 a4 (##not (or (##eq? state 2) p)) n p)))))

      (if (##fixnum.< i (##string-length s))
        (let ((c (##string-ref s i)))
          (if (and (##char=? c #\.) decimal? (##not p))
            (loop1 (##fixnum.+ i 1) state n 0)
            (if (and (##char=? c #\#) (##fixnum.< 0 state))
              (loop1 (##fixnum.+ i 1) 2 (##* n r) (and p (##fixnum.- p 1)))
              (if (##fixnum.< state 2)
                (let ((d (digit c r)))
                  (if d
                    (loop1 (##fixnum.+ i 1)
                           1
                           (##+ (##* n r) d)
                           (and p (##fixnum.- p 1)))
                    (suffix)))
                (suffix)))))
        (suffix))))

  (if (exact-int? rad)
    (if (or (##eq? rad 2)
            (##eq? rad 8)
            (##eq? rad 10)
            (##eq? rad 16))
      (num s 0 rad)
      (range-error))
    (type-error)))

(define (string->number s #!optional (r (absent-obj)))
  (force-vars (s r)
    (let ((radix (if (##eq? r (absent-obj)) 10 r)))
      (check-string s (string->number s r)
        (##string->number s radix)))))

;------------------------------------------------------------------------------
