/* file: "os.h" */

/* Copyright (C) 1994-1996 by Marc Feeley, All Rights Reserved. */

/*---------------------------------------------------------------------------*/

/* Select OS routines to use. */

#ifdef sparc
#define USE_UNIX
#define USE_dlopen
#define USE_sigaction
#endif

#ifdef sgi
#define USE_UNIX
#define USE_dlopen
#define USE_sigaction
#endif

#ifdef __alpha
#define USE_UNIX
#define USE_dlopen
#define USE_sigaction
#endif

#ifdef hpux
#define USE_UNIX
#define USE_shl_load
#define USE_sigaction
#endif

#ifdef _AIX
#define USE_UNIX
#define USE_sigaction
#endif

#ifdef linux
#ifdef i386
#define USE_UNIX
#define USE_dlopen
#define USE_sigaction
#endif
#endif

#ifdef __FreeBSD__
#ifdef i386
#define USE_UNIX
#define USE_dlopen
#define USE_sigaction
#endif
#endif

#ifdef __DJGPP__
#ifdef __i386__
#define USE_DJGPP
#define USE_signal
#endif
#endif

#ifdef _MSC_VER
#if _MSC_VER < 1000
#ifdef _M_I86
#define USE_WINDOWS
#define USE_signal
#endif
#else
#ifdef _M_IX86
#define USE_WINDOWS
#define USE_signal
#endif
#endif
#endif

#ifdef __WATCOMC__
#ifdef __386__
#define USE_DOS
#define USE_clock
#endif
#endif

#ifdef __MWERKS__
#define USE_MACOS
#define USE_TickCount
#define EINTR _ERRMAX
#define EIO (_ERRMAX+1)
#define fileno(x)x
#define isatty(x)((x==stdin)||(x==stdout))
#endif

#ifdef __MACHTEN__
#define USE_UNIX
#endif

/*---------------------------------------------------------------------------*/

/* Determine which header files to include. */

#ifdef USE_UNIX
#define USE_times
#define INCLUDE_unistd_h
#define INCLUDE_pwd_h
#define INCLUDE_sys_types_h
#define INCLUDE_sys_time_h
#define INCLUDE_sys_times_h
#define INCLUDE_time_h
#endif

#ifdef USE_DJGPP
#define USE_DOS
#define USE_times
#define INCLUDE_sys_time_h
#define INCLUDE_sys_times_h
#endif

#ifdef USE_WINDOWS
#define USE_DOS
#define USE_LoadLibrary
#define USE_GetProcessTime
#define INCLUDE_windows_h
#endif

#ifdef USE_DOS
#ifndef USE_WINDOWS
#define USE_dos_setvect
#define INCLUDE_dos_h
#endif
#define INCLUDE_direct_h
#define INCLUDE_time_h
#endif

#ifdef USE_MACOS
#define INCLUDE_Retrace_h
#define INCLUDE_Files_h
#define INCLUDE_Errors_h
#define INCLUDE_Folders_h
#endif

#ifdef USE_getrusage
#define INCLUDE_sys_resource_h
#endif

#ifdef USE_dlopen
#define INCLUDE_dlfcn_h
#endif

#ifdef USE_shl_load
#define INCLUDE_dl_h
#endif

#ifdef USE_sigaction
#define INCLUDE_signal_h
#endif

#ifdef USE_signal
#define INCLUDE_signal_h
#endif


/*---------------------------------------------------------------------------*/

/* Inclusion of header files. */

#define _STDIO_UNLOCK_CHAR_IO

#include <stdio.h>
#include <errno.h>

#ifndef ESUCCESS
#define ESUCCESS 0
#endif

#ifndef EINTR
#define EINTR -1
#endif

#include <stdlib.h>
#include <limits.h>
#include <string.h>

#ifdef INCLUDE_unistd_h
#include <unistd.h>
#endif

#ifdef INCLUDE_pwd_h
#include <pwd.h>
#endif

#ifdef INCLUDE_dlfcn_h
#include <dlfcn.h>
#endif

#ifdef INCLUDE_dl_h
#include <dl.h>
#endif

#ifdef INCLUDE_sys_types_h
#include <sys/types.h>
#endif

#ifdef INCLUDE_sys_time_h
#include <sys/time.h>
#endif

#ifdef INCLUDE_sys_times_h
#include <sys/times.h>
#endif

#ifdef INCLUDE_time_h
#include <time.h>
#endif

#ifdef INCLUDE_sys_resource_h
#include <sys/resource.h>
#endif

#ifdef INCLUDE_signal_h
#include <signal.h>
#endif

#ifdef INCLUDE_windows_h
#include <windows.h>
#endif

#ifdef INCLUDE_dos_h
#include <dos.h>
#endif

#ifdef INCLUDE_direct_h
#include <direct.h>
#endif

#ifdef INCLUDE_Retrace_h
#include <Retrace.h>
#endif

#ifdef INCLUDE_Files_h
#include <Files.h>
#endif

#ifdef INCLUDE_Errors_h
#include <Errors.h>
#endif

#ifdef INCLUDE_Folders_h
#include <Folders.h>
#endif


/*---------------------------------------------------------------------------*/

/* Time between each timer interrupt, in msecs. */

#define ___DEFAULT_TIMER_INTERVAL 100


/* Max length of a path, not including null. */

#define ___PATH_MAX_LENGTH 1024


/* Does "___alloc_mem" allocate upwards (___ALLOC_MEM_UP) or downwards
 * (___ALLOC_MEM_DOWN).  This setting only affects performance.
 */

#define ___ALLOC_MEM_UP


/* Low-level memory allocation. */

extern void *___alloc_mem (/* bytes */);
extern void ___free_mem (/* ptr */);


/* User and timer signal handling. */

extern void ___set_timer_interval (/* msecs */);


/* Periodic task to run. */

extern void ___periodic_task ();


/* Dynamic loading. */

extern void *___dynamic_load (/* path, init_proc_name, errmsg */);


/* Shell command. */

extern int ___shell_command (/* cmd */);


/* Elapsed process time. */

extern unsigned long ___process_time ();


/* Filename expansion. */

extern int ___path_expand (/* path, new_path, max_length */);
extern int ___path_absolute (/* path */);
extern int ___path_extension (/* path, ext, max_length */);
extern int ___path_strip_extension (/* path, new_path, max_length */);
extern int ___path_directory (/* path, dir, max_length */);
extern int ___path_strip_directory (/* path, new_path, max_length */);
extern int ___path_drive (/* path, drive, max_length */);
extern int ___path_strip_drive (/* path, new_path, max_length */);


/* Low-level I/O. */

typedef struct
  {
    FILE *stdio_stream;
    int io_encoding;
    int force_tty;
    ___WORD lookahead;
  } ___STREAM;

#define ___IO_INPUT        -1
#define ___IO_INPUT_OUTPUT 0
#define ___IO_OUTPUT       1

extern ___STREAM *___io_stdin ();
extern ___STREAM *___io_stdout ();
extern ___STREAM *___io_stderr ();

extern ___STREAM *___io_open_text (/* filename, direction, encoding */);
extern ___WORD ___io_close (/* stream */);

extern ___WORD ___io_getc (/* stream */);
extern ___WORD ___io_peek (/* stream */);
extern ___WORD ___io_ready (/* stream */);
extern ___WORD ___io_putc (/* stream, unicode */);
extern ___WORD ___io_flush (/* stream */);
extern ___WORD ___io_width (/* stream */);
extern ___WORD ___io_isatty (/* stream */);


/* OS initialization/finalization. */

extern void ___os_begin (/* setup_params */);
extern void ___os_end ();


/* Utilities for machine encoding of characters. */

/*
 * For now a LATIN-1 encoding of characters is assumed for 'unsigned char'.
 * This means that the mapping to and from Unicode is the identity function.
 * To support other machine representations, e.g. EBCDIC, these macros
 * would have to actually translate the characters.
 */

#define unicode_to_uchar(u)u
#define uchar_to_unicode(c)c


/*---------------------------------------------------------------------------*/
