/* file: "setup.c" */

/* Copyright (C) 1994-1996 by Marc Feeley, All Rights Reserved. */

/* 
 * This module contains the routines that setup the Scheme program for
 * execution.
 */

#define ___VERSION 15
#include "gambit.h"
___BEGIN_MODULE

#include "os.h"
#include "setup.h"
#include "mem.h"
#include "c_intf.h"

#include <stdio.h>
#include <stdlib.h>

/*---------------------------------------------------------------------------*/

/*
 * Global Scheme variables needed by this module.
 */

___WORD ___symbol_table;   /* Hash table of all interned symbols.  */
___WORD ___keyword_table;  /* Hash table of all interned keywords. */
___WORD ___arguments;      /* List of command line arguments.      */
___WORD ___exec_vector;


/*---------------------------------------------------------------------------*/

/*
 * ___debug_level is non-zero if debugging information is requested.
 */

int ___debug_level = 0;


/*
 * Fatal error handler.
 */


void (*___fatal_error) ();


/* 
 * Length of symbol table and keyword table.
 */

#define SYMKEY_TBL_LENGTH 43


/*---------------------------------------------------------------------------*/

/* 
 * Interrupt handling.
 *
 * '___raise_interrupt (code)' is called when an interrupt has
 * occured.  At some later point in time, the kernel will cause the
 * Scheme procedure ##interrupt-handler to be called with a single
 * integer argument indicating which interrupt has been received.
 * Interrupt codes are defined in "gambit.h".  Currently, the
 * following codes are defined:
 *
 *   ___INTR_USER    user has interrupted the program (e.g. ctrl-C)
 *   ___INTR_TIMER   interval timer has elapsed
 *   ___INTR_GC      a garbage collection has finished
 */

void ___raise_interrupt (code)
int code;
{
  ___processor pstate = ___PSTATE;

  /* 
   * Note: ___raise_interrupt may be called before the processor state
   * is initialized.  As a consequence, the interrupt(s) received
   * before the initialization of the processor state will be ignored.
   */

  pstate->intr_flag[code] = 1;
  if (pstate->intr_enabled)
    pstate->stack_trip = pstate->stack_base;
}

/*---------------------------------------------------------------------------*/

___HIDDEN void fatal_heap_overflow ()
{
  ___fatal_error ("Heap overflow during setup");
}

___HIDDEN ___WORD alloc_perm_vector (n)
unsigned long n;
{
  ___WORD result;
  result = ___alloc_scmobj (___sVECTOR, n<<___LWS, ___PERM);
  if (result == ___FAL)
    fatal_heap_overflow ();
  return result;
}

/* 
 * 'make_perm_string_from_charstring (str)' converts the
 * null-terminated C string 'str' to a Scheme string.  A permanent
 * object is allocated.  The value ___FAL is returned if the object
 * can not be allocated.
 */

___HIDDEN ___WORD make_perm_string_from_charstring (str)
char *str;
{
  ___WORD r;
  unsigned long i, n = 0;
  while (str[n] != 0)
    n++;
  r = ___alloc_scmobj (___sSTRING, n<<___LCS, ___PERM);
  if (r == ___FAL)
    fatal_heap_overflow ();
  for (i=0; i<n; i++)
    /* No error possible because a 'char' always fits in a Scheme char. */
    ___STRINGSET(r,___FIX(i),___CHR((unsigned char)str[i]))
  return r;
}

/* 
 * 'make_perm_string_from_utf8string (str)' converts the UTF-8 encoded
 * C string 'str' to a Scheme string.  A permanent object is
 * allocated.  The value ___FAL is returned if the object can not be
 * allocated.
 */

___HIDDEN ___WORD make_perm_string_from_utf8string (str)
___UTF8STRING str;
{
  ___WORD r;
  unsigned long i, n = 0;
  ___UTF8STRING p = str;
  while (___utf8_get (&p) != 0) /* advance until end or error */
    n++;
  r = ___alloc_scmobj (___sSTRING, n<<___LCS, ___STILL);
  if (r == ___FAL)
    fatal_heap_overflow ();
  p = str;
  for (i=0; i<n; i++)
    {
      ___UTF8STRING start = p;
      ___UCS4 c = ___utf8_get (&p);
      if (p == start || c > ___MAX_CHR)
        ___fatal_error ("Illegal UTF-8 encoding during setup");
      ___STRINGSET(r,___FIX(i),___CHR(c))
    }
  return r;
}

___HIDDEN long hash_utf8string (str)
___UTF8STRING str;
{
  long h = 0;
  ___UTF8STRING p = str;
  ___UCS4 c;

  while (1)
    {
      ___UTF8STRING start = p;
      c = ___utf8_get (&p);
      if (p == start || c > ___MAX_CHR)
        ___fatal_error ("Illegal UTF-8 encoding during setup");
      if (c == 0)
        return h;
      h = (long)((h<<8)+c)%SYMKEY_TBL_LENGTH;
    }
}

___HIDDEN long hash_schemestring (str)
___WORD str;
{
  unsigned long i, n = ___INT(___STRINGLENGTH(str));
  long h = 0;
  for (i=0; i<n; i++)
    h = (long)((h<<8)+___INT(___STRINGREF(str,___FIX(i))))%SYMKEY_TBL_LENGTH;
  return h;
}

___HIDDEN ___WORD symkey_table (subtype)
unsigned int subtype;
{
  switch (subtype)
    {
    case ___sKEYWORD:
      return ___keyword_table;
    default: /* assume ___sSYMBOL */
      return ___symbol_table;
    }
}

___HIDDEN ___WORD find_symkey_from_utf8string (str, subtype)
char *str;
unsigned int subtype;
{
  long h = hash_utf8string (str);
  ___WORD tbl = symkey_table (subtype);
  ___WORD probe = ___FIELD(tbl,h);

  while (probe != ___NUL)
    {
      ___WORD obj = ___CAR(probe);
      ___WORD name = ___FIELD(obj,0);
      unsigned long i;
      unsigned long n = ___INT(___STRINGLENGTH(name));
      ___UTF8STRING p = str;
      for (i=0; i<n; i++)
        if (___utf8_get (&p) != ___INT(___STRINGREF(name,___FIX(i))))
          goto next;
      if (___utf8_get (&p) == 0)
        return obj;
    next:
      probe = ___CDR(probe);
    }

  return ___FAL;
}

___WORD ___find_symkey_from_schemestring (str, subtype)
___WORD str;
unsigned int subtype;
{
  long h = hash_schemestring (str);
  ___WORD tbl = symkey_table (subtype);
  ___WORD probe = ___FIELD(tbl,h);

  while (probe != ___NUL)
    {
      ___WORD obj = ___CAR(probe);
      ___WORD name = ___FIELD(obj,0);
      unsigned long i = 0;
      unsigned long n = ___INT(___STRINGLENGTH(name));
      if (___INT(___STRINGLENGTH(str)) == n)
        {
          for (i=0; i<n; i++)
            if (___STRINGREF(str,___FIX(i)) != ___STRINGREF(name,___FIX(i)))
              goto next;
          return obj;
        }
    next:
      probe = ___CDR(probe);
    }

  return ___FAL;
}

___WORD ___new_symkey (name, subtype) /* name must be a permanent object */
___WORD name;
unsigned int subtype;
{
  long h = hash_schemestring (name);
  ___WORD tbl = symkey_table (subtype);
  ___WORD obj, lst;

  switch (subtype)
    {
    case ___sKEYWORD:
      obj = ___alloc_scmobj (___sKEYWORD, ___KEYWORD_SIZE<<___LWS, ___PERM);
      break;
    default: /* assume ___sSYMBOL */
      obj = ___alloc_scmobj (___sSYMBOL, ___SYMBOL_SIZE<<___LWS, ___PERM);
      break;
    }

  if (obj == ___FAL)
    return ___FAL;

  ___FIELD(obj,0) = name;
  ___FIELD(obj,1) = ___FIX(h);
  if (subtype == ___sSYMBOL)
    ___FIELD(obj,2) = 0;

  lst = ___make_pair (obj, ___FIELD(tbl,h), ___PERM);

  if (lst == ___FAL)
    return ___FAL;

  ___FIELD(tbl,h) = lst;

  return obj;
}

___HIDDEN ___WORD make_symkey (str, subtype)
___UTF8STRING str;
unsigned int subtype;
{
  ___WORD obj = find_symkey_from_utf8string (str, subtype);

  if (obj == ___FAL)
    {
      obj = ___new_symkey (make_perm_string_from_utf8string (str), subtype);
      if (obj == ___FAL)
        fatal_heap_overflow ();
    }

  return obj;
}

___HIDDEN ___glo_struct *make_global (str, supply)
___UTF8STRING str;
int supply;
{
  ___WORD sym = make_symkey (str, ___sSYMBOL);
  ___WORD glo = ___FIELD(sym,2);
  ___glo_struct *p;

  if (glo != 0)
  {
    p = (___glo_struct*)glo;
    if (supply && (p->val == ___UNB1))
      p->val = ___UNB2;
    return p;
  }
  else
    {
      ___processor pstate = ___PSTATE;
      p = ___alloc_global_var ();
      if (p == 0)
        ___fatal_error ("Can't allocate global var");
      p->val = supply?___UNB2:___UNB1;
      p->prm = ___FAL;
      p->next = pstate->glo_list;
      pstate->glo_list = p;
      ___FIELD(sym,2) = (___WORD)p;
      return p;
    }
}

___HIDDEN char hex_digits[] = "0123456789abcdef";

#define c_id_subsequent(unicode) \
(((unicode)>='A'&&(unicode)<='Z') || \
 ((unicode)>='a'&&(unicode)<='z') || \
 ((unicode)>='0'&&(unicode)<='9') || \
 ((unicode)=='_'))

___HIDDEN int scheme_id_to_c_id (scm_id, c_id, max_length)
char *scm_id, *c_id;
long max_length;
{
  long i = 0;
  long j = 0;
  unsigned char c, c2;
  long k, n;

  while ((c = scm_id[i++]) != '\0')
    {
      if (c == '_')
        {
          if (j+2 > max_length)
            return 0;
          c_id[j++] = '_';
          c_id[j++] = '_';
        }
      else if (c_id_subsequent(c))
        {
          if (j+1 > max_length)
            return 0;
          c_id[j++] = c;
        }
      else
        {
          c2 = c;
          n = 1;
          while (c2 > 15)
            {
              c2 = c2 >> 4;
              n++;
            }
          if (j+n+2 > max_length)
            return 0;
          c_id[j++] = '_';
          for (k=n-1; k>=0; k--)
            {
              c_id[j+k] = hex_digits[c & 15];
              c = c >> 4;
            }
          j += n;
          c_id[j++] = '_';
        }
    }
  if (j+1 > max_length)
    return 0;
  c_id[j++] = '\0';
  return 1;
}

/*---------------------------------------------------------------------------*/

/* Alignment of objects */

___HIDDEN ___WORD *align (from, words, flonum)
___WORD *from;
long words;
int flonum;
{
  ___WORD *to;
#if ___WS == 4
  if (flonum)
    to = ___ALIGNUP((from+1), (___FLONUM_SIZE<<___LWS)) - 1;
  else
#endif
    to = ___ALIGNUP(from, ___WS);

  if (from != to)
    {
      /* move object up */
      int i;
      for (i=words-1; i>=0; i--)
        to[i] = from[i];
    }
  return to;
}

___HIDDEN ___WORD align_subtyped( ptr )
___WORD *ptr;
{
  ___WORD head = ptr[0];
  int subtype = ___HD_SUBTYPE(head);
  int words = ___HD_WORDS(head);
  return ___TAG(align (ptr, words+1, subtype==___sFLONUM), ___tSUBTYPED);
}

/*---------------------------------------------------------------------------*/

/* Routines to setup a module for execution */

___HIDDEN ___mod_or_lnk linker_to_mod_or_lnk (linker)
___mod_or_lnk (*linker) ();
{
  ___mod_or_lnk mol = linker (___GSTATE);
  if (mol->module.kind == ___LINKFILE_KIND)
    {
      void **p = mol->linkfile.linker_tbl;
      while (*p != 0)
        {
          *p = linker_to_mod_or_lnk ((___mod_or_lnk (**)())*p);
          p++;
        }
    }
  return mol;
}

___HIDDEN void for_each_module (mol, proc)
___mod_or_lnk mol;
void (*proc) ();
{
  if (mol->module.kind == ___LINKFILE_KIND)
    {
      void **p = mol->linkfile.linker_tbl;
      while (*p != 0)
        for_each_module ((___mod_or_lnk)*p++, proc);
    }
  else
    proc ((struct ___module *)mol);
}

___HIDDEN void fixref (p, sym_tbl, key_tbl, cns_tbl, sub_tbl)
___WORD *p;
___WORD *sym_tbl, *key_tbl, *cns_tbl, *sub_tbl;
{
  ___WORD v = *p;
  switch (___TYP(v))
    {
    case ___tPAIR:
      if (___INT(v)<0)
        *p = sym_tbl[-1-___INT(v)];
      else
        *p = ___TAG(___ALIGNUP(&cns_tbl[(___PAIR_SIZE+1)*___INT(v)],___WS),___tPAIR);
      break;

    case ___tSUBTYPED:
      if (___INT(v)<0)
        *p = key_tbl[-1-___INT(v)];
      else
        *p = sub_tbl[___INT(v)];
      break;
    }
}

___HIDDEN void setup_module_phase1 (module)
struct ___module *module;
{
  int i, j;
  ___WORD *cns;

  char *name       = module->name;
  ___glo_struct **glo_tbl= module->glo_tbl;
  int sup_count    = module->sup_count;
  ___UTF8STRING *glo_names = module->glo_names;
  ___WORD *sym_tbl = (___WORD*)module->sym_tbl;
  int sym_count    = module->sym_count;
  ___UTF8STRING *sym_names = module->sym_names;
  ___WORD *key_tbl = (___WORD*)module->key_tbl;
  int key_count    = module->key_count;
  ___UTF8STRING *key_names = module->key_names;
  ___WORD *lp      = module->lp;
  ___WORD *lbl_tbl = (___WORD*)module->lbl_tbl;
  int lbl_count    = module->lbl_count;
  ___WORD *cns_tbl = module->cns_tbl;
  int cns_count    = module->cns_count;
  ___WORD *sub_tbl = (___WORD*)module->sub_tbl;
  int sub_count    = module->sub_count;

  if (___debug_level)
    fprintf (stderr, "setting up module \"%s\"\n", name);

  if (module->version < ___VERSION)
    ___fatal_error ("Module's version number is older than runtime library");
  else if (module->version > ___VERSION)
    ___fatal_error ("Module's version number is newer than runtime library");

  if (cns_tbl != 0)
    cns = align (cns_tbl, (___PAIR_SIZE+1)*cns_count, 0);

  if (glo_names != 0)
    {
      i = 0;
      while (glo_names[i] != 0)
        {
          glo_tbl[i] = make_global (glo_names[i], i<sup_count);
          i++;
        }
    }

  if (sym_names != 0)
    {
      i = 0;
      while (sym_names[i] != 0)
        {
          sym_tbl[i] = make_symkey (sym_names[i], ___sSYMBOL);
          i++;
        }
    }
  else
    for (i=sym_count-1; i>=0; i--)
      sym_tbl[i] = ___TAG(___ALIGNUP(sym_tbl[i], ___WS), ___tSUBTYPED);

  if (key_names != 0)
    {
      i = 0;
      while (key_names[i] != 0)
        {
          key_tbl[i] = make_symkey (key_names[i], ___sKEYWORD);
          i++;
        }
    }
  else
    for (i=key_count-1; i>=0; i--)
      key_tbl[i] = ___TAG(___ALIGNUP(key_tbl[i], ___WS), ___tSUBTYPED);

  for (i=sub_count-1; i>=0; i--)
    sub_tbl[i] = align_subtyped ((___WORD*)sub_tbl[i]);

  for (i=cns_count-1; i>=0; i--)
  {
    fixref (cns+i*(___PAIR_SIZE+1)+1, sym_tbl, key_tbl, cns_tbl, sub_tbl);
    fixref (cns+i*(___PAIR_SIZE+1)+2, sym_tbl, key_tbl, cns_tbl, sub_tbl);
  }

  for (j=sub_count-1; j>=0; j--)
    {
      ___WORD *p = ___UNTAG_AS(sub_tbl[j],___tSUBTYPED);
      ___WORD head = p[0];
      int subtype = ___HD_SUBTYPE(head);
      int words = ___HD_WORDS(head);
      switch (subtype)
        {
        case ___sSYMBOL:
        case ___sKEYWORD:
        case ___sVECTOR:
        case ___sRATNUM:
        case ___sCPXNUM:
          for (i=1; i<=words; i++)
            fixref (p+i, sym_tbl, key_tbl, cns_tbl, sub_tbl);
        }
    }

  if (lbl_count > 0)
    {
      ___lbl_struct *new;
      new = (___lbl_struct*)align (lbl_tbl, lbl_count*___LS, 0);
      for (i=lbl_count-1; i>=0; i--)
        if (new[i].header == ___MAKE_HD((___INTRO_SIZE<<___LWS),___sVECTOR,___RCN,___PERM))
          fixref (&new[i].entry, sym_tbl, key_tbl, cns_tbl, sub_tbl);
        else
          new[i].entry = ___TAG(&new[i].header,___tSUBTYPED);
      *lp = ___TAG(new,___tSUBTYPED);
    }
}

___HIDDEN void setup_module_phase2 (module)
struct ___module *module;
{
  int i;

  ___UTF8STRING name = module->name;
  ___glo_struct **glo_tbl = module->glo_tbl;
  int glo_count = module->glo_count;
  int sup_count = module->sup_count;
  ___UTF8STRING *glo_names = module->glo_names;

  if (glo_names != 0)
    {
      for (i=sup_count; i<glo_count; i++)
        {
          ___glo_struct *glo = glo_tbl[i];
          if (glo->val == ___UNB1)
            {
              fprintf (stderr,
                       "*** WARNING -- Variable \"%s\" used in module \"%s\" is undefined\n",
                       glo_names[i],
                       name);
            }
        }
    }

  module->init_proc();
}

___HIDDEN void setup_modules (mol)
___mod_or_lnk mol;
{
  for_each_module (mol, setup_module_phase1);
  for_each_module (mol, setup_module_phase2);
}

___HIDDEN int module_count;

___HIDDEN void inc_module_count (module)
struct ___module *module;
{
  if (module->lbl_count > 0)
    module_count++;
}

___HIDDEN int count_modules (mol)
___mod_or_lnk mol;
{
  module_count = 0;
  for_each_module (mol, inc_module_count);
  return module_count;
}

___HIDDEN ___WORD module_vector;

___HIDDEN void store_in_module_vector (module)
struct ___module *module;
{
  if (module->lbl_count > 0)
    {
      ___FIELD(module_vector,module_count) = *module->lp+___LS*___WS;
      module_count++;
    }
}

___HIDDEN void setup_module_vector (mol, vect)
___mod_or_lnk mol;
___WORD vect;
{
  module_count = 0;
  module_vector = vect;
  for_each_module (mol, store_in_module_vector);
}

___HIDDEN char module_prefix[] = ";";

#define module_prefix_length (sizeof(module_prefix)-1)

___HIDDEN char c_id_prefix[] =
#ifdef ___PREPEND_UNDERSCORE
"_"
#endif
"___";

#define c_id_prefix_length (sizeof(c_id_prefix)-1)

___WORD ___load_object_file (path, errmsg)
char *path;
char **errmsg;
{
  int i;
  char temp1[___PATH_MAX_LENGTH+1];
  char temp2[___PATH_MAX_LENGTH+1];
  void *f;
  ___mod_or_lnk mol;
  ___WORD ev;
  for (i=module_prefix_length-1; i>=0; i--)
    temp2[i] = module_prefix[i];
  if (___path_strip_extension (path, temp1, ___PATH_MAX_LENGTH) == 0 ||
      ___path_strip_directory (temp1, temp2+module_prefix_length, ___PATH_MAX_LENGTH-module_prefix_length) == 0 ||
      scheme_id_to_c_id (temp2, temp1+c_id_prefix_length, ___PATH_MAX_LENGTH-c_id_prefix_length) == 0)
    {
      *errmsg = "path is too long";
      return ___FAL;
    }
  for (i=c_id_prefix_length-1; i>=0; i--)
    temp1[i] = c_id_prefix[i];
  f = ___dynamic_load (path, temp1, errmsg);
  if (f == 0)
    return ___FAL;
  mol = linker_to_mod_or_lnk ((___mod_or_lnk(*)())f);
  if (mol->linkfile.version < 0) /* was it already setup? */
    {
      *errmsg = "can't load a given object file more than once";
      return ___FAL;
    }
  setup_modules (mol);
  mol->linkfile.version = -1; /* mark link file as 'setup' */
  ev = ___alloc_scmobj (___sVECTOR, count_modules (mol)<<___LWS, ___STILL);
  if (ev == ___FAL)
    {
      *errmsg = "heap overflow";
      return ___FAL;
    }
  setup_module_vector (mol, ev);
  ___still_obj_refcount_dec (ev);
  return ev;
}

/*---------------------------------------------------------------------------*/

extern double modf ();

double ___round (x)
double x;
{
  double f, i, t;
  if (x < 0.0)
    {
      f = modf (-x, &i);
      if (f > 0.5 || (f == 0.5 && modf (i*0.5, &t) != 0.0))
        return -(i+1.0);
      else
        return -i;
    }
  else
    {
      f = modf (x, &i);
      if (f > 0.5 || (f == 0.5 && modf (i*0.5, &t) != 0.0))
        return i+1.0;
      else
        return i;
    }
}

/*---------------------------------------------------------------------------*/

___HIDDEN void init_sym_glo1 (mol)
___mod_or_lnk mol;
{
  if (mol->module.kind == ___LINKFILE_KIND)
    {
      void **p1 = mol->linkfile.linker_tbl;
      ___FAKEWORD *p2 = mol->linkfile.sym_list;

      while (*p1 != 0)
        init_sym_glo1 ((___mod_or_lnk)*p1++);

      while (p2 != 0)
        {
          ___WORD *sym_ptr;
          ___glo_struct *glo;

          sym_ptr = (___WORD*)p2;

          p2 = (___FAKEWORD*)sym_ptr[0];
          glo = (___glo_struct*)sym_ptr[3];

          sym_ptr[2] = glo->prm;
        }
    }
}

___HIDDEN void init_sym_glo2 (mol)
___mod_or_lnk mol;
{
  if (mol->module.kind == ___LINKFILE_KIND)
    {
      void **p1 = mol->linkfile.linker_tbl;
      ___FAKEWORD *p2 = mol->linkfile.sym_list;

      while (*p1 != 0)
        init_sym_glo2 ((___mod_or_lnk)*p1++);

      while (p2 != 0)
        {
          long h;
          ___WORD sym, lst, str;
          ___WORD *sym_ptr;
          ___glo_struct *glo;
          ___processor pstate = ___PSTATE;

          sym_ptr = (___WORD*)p2;

          p2 = (___FAKEWORD*)sym_ptr[0];
          str = align_subtyped ((___WORD*)sym_ptr[1]);
          glo = (___glo_struct*)sym_ptr[3];

          glo->next = pstate->glo_list;
          pstate->glo_list = glo;

          *sym_ptr = ___MAKE_HD((___SYMBOL_SIZE<<___LWS),___sSYMBOL,___RCN,___PERM);

          sym = align_subtyped (sym_ptr);

          h = hash_schemestring (str);

          ___FIELD(sym,0) = str;
          ___FIELD(sym,1) = ___FIX(h);
          ___FIELD(sym,2) = (___WORD)glo;

          lst = ___make_pair (sym, ___FIELD(___symbol_table,h), ___PERM);
          if (lst == ___FAL)
            fatal_heap_overflow ();

          ___FIELD(___symbol_table,h) = lst;
        }
    }
}


/*---------------------------------------------------------------------------*/

/* Dispatcher */

___HIDDEN int in_dispatcher = 0;

___HIDDEN void dispatcher ()
{
  ___processor pstate = ___PSTATE;
  ___WORD pc = pstate->pc;

  if (in_dispatcher)
    ___fatal_error ("Nested C to Scheme calls are not allowed");

  in_dispatcher = 1;

  while (___TYP(pc) == ___tSUBTYPED &&
         ___SUBTYPE(pc) == ___FIX(___sPROCEDURE))
    {
      pc = ___CAST_FAKEHOST_TO_HOST(((___lbl_struct*)(pc-___tSUBTYPED))->host)(pstate);
      pstate->pc = pc;
    }

  if (pc != ___BRK)
    {
      char msg[100];
      sprintf (msg, "Jump to illegal PC (pc=0x%08lx)", pc);
      ___fatal_error (msg);
    }

  in_dispatcher = 0;
}


/*---------------------------------------------------------------------------*/

/*
 * Setup and run program.
 */

___EXPORT void ___setup_and_run (setup_params)
___setup_params *setup_params;
{
  ___processor pstate;
  ___mod_or_lnk mol;
  int i;

  /*
   * Setup fatal error handler.
   */

  ___fatal_error = setup_params->fatal_error;
  ___debug_level = setup_params->debug_level;

  /* 
   * Initialize the operating system module.
   */

  ___os_begin (setup_params);

  pstate = ___PSTATE;

  /* 
   * Allocate initial stack and heap.
   */

  ___setup_mem (setup_params);

  /* 
   * Setup global state to avoid problems on systems that don't
   * support the dynamic loading of files that import functions.
   */

#ifdef ___NO_DYNAMIC_IMPORTS

  ___GSTATE->fabs  = fabs;
  ___GSTATE->floor = floor;
  ___GSTATE->ceil  = ceil;
  ___GSTATE->exp   = exp;
  ___GSTATE->log   = log;
  ___GSTATE->sin   = sin;
  ___GSTATE->cos   = cos;
  ___GSTATE->tan   = tan;
  ___GSTATE->asin  = asin;
  ___GSTATE->acos  = acos;
  ___GSTATE->atan  = atan;
  ___GSTATE->sqrt  = sqrt;

  ___GSTATE->___scmobj_to_char         = ___scmobj_to_char;
  ___GSTATE->___scmobj_to_schar        = ___scmobj_to_schar;
  ___GSTATE->___scmobj_to_uchar        = ___scmobj_to_uchar;
  ___GSTATE->___scmobj_to_latin1       = ___scmobj_to_latin1;
  ___GSTATE->___scmobj_to_ucs4         = ___scmobj_to_ucs4;
  ___GSTATE->___scmobj_to_ucs2         = ___scmobj_to_ucs2;
  ___GSTATE->___scmobj_to_short        = ___scmobj_to_short;
  ___GSTATE->___scmobj_to_ushort       = ___scmobj_to_ushort;
  ___GSTATE->___scmobj_to_int          = ___scmobj_to_int;
  ___GSTATE->___scmobj_to_uint         = ___scmobj_to_uint;
  ___GSTATE->___scmobj_to_long         = ___scmobj_to_long;
  ___GSTATE->___scmobj_to_ulong        = ___scmobj_to_ulong;
  ___GSTATE->___scmobj_to_float        = ___scmobj_to_float;
  ___GSTATE->___scmobj_to_double       = ___scmobj_to_double;
  ___GSTATE->___scmobj_to_pointer      = ___scmobj_to_pointer;
  ___GSTATE->___scmobj_to_function     = ___scmobj_to_function;
  ___GSTATE->___scmobj_to_boolean      = ___scmobj_to_boolean;
  ___GSTATE->___scmobj_to_charstring   = ___scmobj_to_charstring;
  ___GSTATE->___scmobj_to_latin1string = ___scmobj_to_latin1string;
  ___GSTATE->___scmobj_to_ucs4string   = ___scmobj_to_ucs4string;
  ___GSTATE->___scmobj_to_ucs2string   = ___scmobj_to_ucs2string;
  ___GSTATE->___scmobj_to_utf8string   = ___scmobj_to_utf8string;
  ___GSTATE->___free_function          = ___free_function;
  ___GSTATE->___free_charstring        = ___free_charstring;
  ___GSTATE->___free_latin1string      = ___free_latin1string;
  ___GSTATE->___free_ucs4string        = ___free_ucs4string;
  ___GSTATE->___free_ucs2string        = ___free_ucs2string;
  ___GSTATE->___free_utf8string        = ___free_utf8string;
  ___GSTATE->___char_to_scmobj         = ___char_to_scmobj;
  ___GSTATE->___schar_to_scmobj        = ___schar_to_scmobj;
  ___GSTATE->___uchar_to_scmobj        = ___uchar_to_scmobj;
  ___GSTATE->___latin1_to_scmobj       = ___latin1_to_scmobj;
  ___GSTATE->___ucs4_to_scmobj         = ___ucs4_to_scmobj;
  ___GSTATE->___ucs2_to_scmobj         = ___ucs2_to_scmobj;
  ___GSTATE->___short_to_scmobj        = ___short_to_scmobj;
  ___GSTATE->___ushort_to_scmobj       = ___ushort_to_scmobj;
  ___GSTATE->___int_to_scmobj          = ___int_to_scmobj;
  ___GSTATE->___uint_to_scmobj         = ___uint_to_scmobj;
  ___GSTATE->___long_to_scmobj         = ___long_to_scmobj;
  ___GSTATE->___ulong_to_scmobj        = ___ulong_to_scmobj;
  ___GSTATE->___float_to_scmobj        = ___float_to_scmobj;
  ___GSTATE->___double_to_scmobj       = ___double_to_scmobj;
  ___GSTATE->___pointer_to_scmobj      = ___pointer_to_scmobj;
  ___GSTATE->___function_to_scmobj     = ___function_to_scmobj;
  ___GSTATE->___boolean_to_scmobj      = ___boolean_to_scmobj;
  ___GSTATE->___charstring_to_scmobj   = ___charstring_to_scmobj;
  ___GSTATE->___latin1string_to_scmobj = ___latin1string_to_scmobj;
  ___GSTATE->___ucs4string_to_scmobj   = ___ucs4string_to_scmobj;
  ___GSTATE->___ucs2string_to_scmobj   = ___ucs2string_to_scmobj;
  ___GSTATE->___utf8string_to_scmobj   = ___utf8string_to_scmobj;
  ___GSTATE->___release_scmobj         = ___release_scmobj;

  ___GSTATE->___round = ___round;

  ___GSTATE->___alloc_scmobj           = ___alloc_scmobj;
  ___GSTATE->___make_pair              = ___make_pair;
  ___GSTATE->___make_vector            = ___make_vector;
  ___GSTATE->___make_string            = ___make_string;
  ___GSTATE->___still_obj_refcount_inc = ___still_obj_refcount_inc;
  ___GSTATE->___still_obj_refcount_dec = ___still_obj_refcount_dec;

#endif

  /* 
   * Setup interrupt system.
   */

  pstate->stack_trip = pstate->stack_limit;

  pstate->intr_enabled = 0; /* disable interrupts */
  for (i=0; i<___NB_INTRS; i++)
    pstate->intr_flag[i] = 0;

  /* 
   * Initialize will lists.
   */

  pstate->executable_wills = ___FAL;
  pstate->non_executable_wills = ___FAL;

  /*
   * Setup program's linker structure.
   */

  mol = linker_to_mod_or_lnk (setup_params->linker);

  /* 
   * Create empty symbol table, keyword table, and global
   * variable list.
   */

  ___symbol_table = alloc_perm_vector (SYMKEY_TBL_LENGTH);
  for (i=0; i<SYMKEY_TBL_LENGTH; i++)
    ___FIELD(___symbol_table,i) = ___NUL;

  ___keyword_table = alloc_perm_vector (SYMKEY_TBL_LENGTH);
  for (i=0; i<SYMKEY_TBL_LENGTH; i++)
    ___FIELD(___keyword_table,i) = ___NUL;

  pstate->glo_list = 0;

  /* 
   * Initialize symbol table, keyword table, global variables
   * and primitives.
   */

  init_sym_glo1 (mol);
  init_sym_glo2 (mol);

  /* 
   * Setup each module.
   */

  setup_modules (mol);

  /* 
   * Create execution vector.
   */

  ___exec_vector = alloc_perm_vector (count_modules (mol));
  setup_module_vector (mol, ___exec_vector);

  /* 
   * Create list of command line arguments (accessible through ##argv).
   */

  {
    int argc = setup_params->argc;
    char **argv = setup_params->argv;
    char *dummy_argv[2];
    if (argc < 1)
      {
        dummy_argv[0] = "";
        dummy_argv[1] = 0;
        argc = 1;
        argv = dummy_argv;
      }
    ___arguments = ___NUL;
    for (i=argc-1; i>=0; i--)
      {
        ___WORD arg = make_perm_string_from_charstring (argv[i]);
        ___WORD x = ___make_pair (arg, ___arguments, ___PERM);
        if (x == ___FAL)
          fatal_heap_overflow ();
        ___arguments = x;
      }
  }

  /* 
   * Setup call to kernel to start executing program.
   */

  pstate->fp[-1] = 0; /* create "break frame" of initial top section */
  pstate->fp[-2] = 0;
  pstate->fp -= 2;
  pstate->stack_break = pstate->fp;

  for (i=1; i<___NB_GVM_REGS; i++)
    pstate->r[i] = ___VOID;

  pstate->r[0] = ___VOID; /* the kernel replaces this with the break handler */
  pstate->na = 0;
  pstate->pc = ___FIELD(___exec_vector,0);

  dispatcher (); /* run program */
}

___EXPORT void ___cleanup ()
{
  ___os_end ();
}

/*---------------------------------------------------------------------------*/

/* C to Scheme call handler */

___EXPORT int ___call (nargs, proc)
int nargs;
___WORD proc;
{
  ___processor ___ps = ___PSTATE;
  ___WORD *___fp = ___ps->fp - nargs;
  ___LD_ARG_REGS

  ___POP_ARGS_IN_REGS(nargs)

  ___ST_ARG_REGS

  ___ps->fp = ___fp;
  ___ps->r[0] = ___ps->handler_break;
  ___ps->na = nargs;
  ___ps->pc = proc;

  dispatcher ();

  return 0;
}

___EXPORT void ___propagate_error (err)
int err;
{
  ___fatal_error ("Error propagation not implemented");
}

/*---------------------------------------------------------------------------*/
